/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "engine.h"
#include "game.h"
#include "statview.h"

#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <signal.h>

#include <qmessagebox.h>
#include <qstringlist.h>
#include <qtimer.h>
#include <qsocketnotifier.h>

#include <kdebug.h>
//#include <qprocess.h>
//#include <kprocess.h>

//qprocess worked like a charm, but if the process was running,
//kfiledialog was frozen (blame kfiledialog)
//kprocess sucks shit, with some engines the communication simply
//doesn't work, writeStdin() isn't reliable, i would have to write some wrapper functions.
//so im back to good old pipe() and fork().

Engine::Engine()
{
}

Engine::Engine(QWidget *parent, Game *g, QString fname, QString args):
    canStat(false),
    linebuf(""),
    name(fname),
    game(g),
    canAnalyze(false),
    canPlayother(false),
    canSetboard(false),
    canSan(false),
    canUsermove(false),
    canTime(true),
    canDraw(true),
    canSigTerm(true),
    canReuse(true)
{

    qDebug("StartChildProcess \"%s\" %s\n", fname.latin1(), args.latin1());

    args = "";

    int outpipe[2], inpipe[2];
    //signal(SIGPIPE, SIG_IGN);
    pipe(outpipe);
    pipe(inpipe);

    //prepend path to filename
    FILE *fp;
    char *path;
    size_t len;

    QString cmd = "which " + fname;
    fp = popen(cmd.latin1(), "r");
    path = NULL;
    len = 0;
    if (getline(&path, &len, fp)>1) {
	fname = path;
	fname.setLength(fname.length()-1);
    }
    free(path);
    
    //build argument list
    int argc = (args.isEmpty()) ? 0 : args.contains(' ')+1;
    argv = new (char *)[argc+1];
    int i;
    argv[0] = strdup(fname.latin1());
    nickName = argv[0];

    for (i=0; i<argc; i++) {
        argv[i+1] = strdup(args.section(' ', i, i).latin1());
    }
    argv[i+1] = NULL;

    if ((pid = fork()) == 0) {
        // Child process
        //force stdin to be read from outpipe
        dup2(outpipe[0], 0);
        //force stdout to be written to inpipe
        dup2(inpipe[1], 1);
        execv(fname.latin1(), argv);
	// If we get here, exec failed
        qDebug("Cant't start the chess engine %s. The Operating System returned error %s\n", fname.latin1(), strerror(errno));
    }

    //make input pipe nonblocking
    int oldflags = fcntl (inpipe[0], F_GETFL, 0);
    oldflags |= O_NONBLOCK;
    fcntl (inpipe[0], F_SETFL, oldflags);
    fpIn = fdopen(inpipe[0], "r");
    fpOut = fdopen(outpipe[1], "w");
    QSocketNotifier *sn = new QSocketNotifier( inpipe[0], QSocketNotifier::Read, this );
    connect( sn, SIGNAL(activated(int)), this, SLOT(collect()) );
    
    nickName = fname.section('/', -1);
    color = -1;
    statView = new StatView(parent, game, this);
    write("xboard");
    write("protover 2");
    write("easy");
    write("post");
    write("level 40 5 0");
    t1time = 300;
    t1moves = 40;
    t1inc = 0;
    t2time = 300;
    t2moves = 40;
    t3time = 300;
    t3moves = 40;
    if (game->gameMoves.count() > 0) {
        write("force");
        for (MoveTree *i = &game->gameMoves; i->moves.count(); i = i->moves.first().nextMove) {
            write(i->moves.first().ln);
        }
        color = 0;
    }
}

void Engine::tryTerminate() {
    ::kill(pid, SIGTERM);
}

void Engine::kill() {
    ::kill(pid, SIGKILL);
}

void Engine::collect() {
    int c;
    while ((c = fgetc(fpIn)) != EOF) {
        if ((char)c == '\n') {
            QString buf=linebuf;
	    linebuf.setLength(0);
	    game->receive(this, buf);
        }
        linebuf += (char)c;
    }
}

void Engine::end() {
    write("quit");
//    proc->kill();
    if (canSigTerm) {
        QTimer::singleShot( 1000, this, SLOT( tryTerminate() ) );
    }
    QTimer::singleShot( 5000, this, SLOT( kill() ) );
}

void Engine::destructor() {
    qDebug("process %s exited", name.latin1());
//    delete statView;
//    delete this;
}

void Engine::features(QString arg) {
    QStringList features = QStringList::split(' ', arg);
    for (QStringList::iterator f = features.begin(); f != features.end(); ++f) {
        QString var = (*f).section('=',0,0).lower();
        QString val = (*f).section('=',1,1);
        if (val.contains('"') == 1)
            do
                val += ' ' + (*++f);
            while(!(*f).contains('"'));

        if (val == "1") {
            if (var == "setboard")
                canSetboard = true;
            else if (var == "playother")
                canPlayother = true;
            else if (var == "san")
                canSan = true;
            else if (var == "usermove")
                canUsermove = true;
            else if (var == "time")
                canTime = true;
            else if (var == "draw")
                canDraw = true;
            else if (var == "sigterm")
                canSigTerm = true;
            else if (var == "reuse")
                canReuse = true;
            else if (var == "analyze")
                canAnalyze = true;
            else if (var == "stat")
                canStat = true;
        }
        if (var == "myname")
            nickName = val;
    }
}

void Engine::setColor(int c) {
    if (c == color) return;
    delayedGo = false;
    color = c;
    if (c == 0) {
        if (canAnalyze)
            write("analyze");
        else {
            write("force");
            qDebug("engine %s doesn't support analyzing", name.ascii());
        }
	return;
    }
    if (c == game->color) {
        write("go");
    } else {
        if (canPlayother)
            write("playother");
        else
            delayedGo = true;
    }
}

void Engine::write(QString cmd) { 
    if (cmd != "stat")
	kdDebug() << cmd << endl;
    cmd += '\n';
    fwrite(cmd.latin1(), cmd.length(), 1, fpOut);
    fflush(fpOut);
}

StatView* Engine::getStatView( )  {
    return statView;
}

void Engine::addText(QString arg) {
    statText = arg;
    if (statView)
        statView->setText("<table>" + lineText + "</table>" + statText);
}

void Engine::addLine(QString arg) {
    QTextStream stream(arg, IO_ReadOnly);
    double ply, time, nodes;
    QString eval, bestline;
    stream >> ply >> eval >> time >> nodes;
    bestline = stream.read();

    bool evalok;
    double evaldouble = eval.toDouble(&evalok);
    if (evalok)
        eval.setNum(evaldouble*0.01, 'f', 2);

    lineText = "<tr><td>" + QString::number(ply,'f',0) + "</td><td>"
        + eval + "</td><td>"
        + QString::number(time*0.01,'f',2) + "</td><td>"
        + QString::number(nodes,'f',0) + "</td><td>"
        + bestline + "</td></tr>"
	+ lineText;

    if (statView)
        statView->setText("<table>" + lineText + "</table>" + statText);
}

void Engine::clear() {
    lineText = "";
}
#include "engine.moc"

