/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef ENGINE_H
#define ENGINE_H

#include <cstdio>
#include <qstring.h>
#include <qobject.h>

class Game;
class StatView;
class QWidget;
/**
class for handling chess engines

@author Gunther Piez
*/
class Engine: QObject {
Q_OBJECT
public:
    Engine(QWidget *, Game* game, QString fname, QString argv = "");

    QString read();
    void write(QString);
    void features(QString);

    Game *getGame() { return game; }
    QString getName() { return name; }
    int getColor() { return color; }
    void setColor(int);
    StatView *getStatView();
    void addText(QString);
    void addLine(QString);
    void clear();
    void end();

    bool delayedGo;
    bool canStat;
    QString nickName;

    int t1time;
    int t1moves;
    int t1inc;
    int t2time;
    int t2moves;
    int t3time;
    int t3moves;

private:
    StatView *statView;
    FILE *fpIn;
    FILE *fpOut;
    int pid;
    QString linebuf;
    QString name;
    QString statText, lineText;
    Game *game;
    char **argv;
    int color;
    bool canAnalyze;
    bool canPlayother;
    bool            canSetboard;
    bool            canSan;
    bool            canUsermove;
    bool            canTime;
    bool            canDraw;
    bool            canSigTerm;
    bool            canReuse;

    Engine();
    
private slots:
    void tryTerminate();
    void kill();
    void destructor();
    void collect();

};

#endif
