/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include <math.h>
#include <stdlib.h>
#include "flatview.h"
#include "game.h"
#include "promodialog.h"
//#include "slibo.h"

#include <qtimer.h>
#include <qpushbutton.h>
#include <qpainter.h>
#include <qdatetime.h>

#include <kstandarddirs.h>
#include <kaction.h>

FlatView::FlatView(QWidget * parent, Game *g):
    QWidget(parent)
{
    game = g;
    setCaption(game->gameMoves.tagPairs["White"] + " vs " + game->gameMoves.tagPairs["Black"]);
    QPixmap::setDefaultOptimization( QPixmap::BestOptim );
    setBackgroundMode(Qt::NoBackground);
    chessPieces = QImage(locate("appdata", "set.png"));

    frameTimer = new QTimer(this);
    connect(frameTimer, SIGNAL(timeout()), SLOT(frame()));
    frameTimer->start(20);
}

FlatView::~FlatView() {
}

//convert board to screen coordinates

QPoint FlatView::board2screen(CoorFloat q) {
    int squaresize = (width() <? height())/8;
    return QPoint((int)(q.x*squaresize), (int)((8.0-q.y)*squaresize));
}

//convert screen to board coordinates.

CoorFloat FlatView::screen2board(QPoint p1) {
    float squaresize = (width() <? height())/8.0;
    return CoorFloat(p1.x() / squaresize, p1.y() / squaresize);
}

int FlatView::selection(QPoint p) {
    QValueList<ViewPiece>::iterator i;
    int ii;
    CoorFloat q = screen2board(p);
    for (i = game->visiblePieces.begin(), ii=0; i != game->visiblePieces.end(); ++i, ++ii)
        if (round((*i).curr.x) == floor(q.x) && round((*i).curr.y) == floor(q.y))
            return ii;
    return -1;
}

void FlatView::resizeEvent(QResizeEvent *e) {
    int squaresize = (e->size().width() <? e->size().height())/8;
    int nsize = squaresize * 8;
    if (nsize != width() || nsize != height()) {
        resize(nsize, nsize);
        return;
    }
    chessPiecesPM = QPixmap(chessPieces.smoothScale(squaresize*6, squaresize*2));
    bufferPM = QPixmap(e->size());
    boardPM = QPixmap(e->size());
    boardWithNonMovingPM = QPixmap(e->size());
    bufferPM.detach();
    boardPM.detach();
    boardWithNonMovingPM.detach();
    QPainter painter(&boardPM);
    painter.setPen(NoPen);
    for (int x=0; x<8; x++)
        for (int y=0; y<8; y++) {
            if ((x + y) & 1)
                painter.setBrush(QColor(100, 90, 70));
            else
                painter.setBrush(QColor(230, 220, 200));
            painter.drawRect(x*squaresize, y*squaresize, squaresize, squaresize);
        }
    game->movingChanged = true;
}

void FlatView::paintEvent(QPaintEvent*) {
    int squaresize = (width() <? height())/8;
    QTime t = QTime::currentTime();

    QValueList<ViewPiece>::iterator i;

    if (game->movingChanged) {
        bitBlt(&boardWithNonMovingPM, 0, 0, &boardPM);
        for (i = game->visiblePieces.begin(); i != game->visiblePieces.end(); i++)
            if (!(*i).moving && !(*i).picked)
                bitBlt(&boardWithNonMovingPM, (int)((*i).curr.x * squaresize), (int)((7-(*i).curr.y) * squaresize), &chessPiecesPM,
                    (abs((*i).type)-1)*squaresize, (*i).type<0 ? squaresize : 0, squaresize, squaresize);
        game->movingChanged = false;
    }

    bitBlt(&bufferPM, 0, 0, &boardWithNonMovingPM);

    for (i = game->visiblePieces.begin(); i != game->visiblePieces.end(); i++)
        if ((*i).moving || (*i).picked)
            bitBlt(&bufferPM, (int)((*i).curr.x * squaresize), (int)((7-(*i).curr.y) * squaresize), &chessPiecesPM,
                (abs((*i).type)-1)*squaresize, (*i).type<0 ? squaresize : 0, squaresize, squaresize);

    bitBlt(this, 0, 0, &bufferPM);
 //   qDebug("%d", t.msecsTo(QTime::currentTime()));
}

void FlatView::frame() {
    if (repaintNeeded) {
        repaint();
        repaintNeeded = false;
    }
}

/*!
    \fn FlatView::mousePressEvent(QMouseEvent *e)
 */
void FlatView::mousePressEvent(QMouseEvent *e)
{
    oldPos = e->pos();
    if (e->button() == Qt::LeftButton) {
        setMouseTracking(leftClick(e->pos()));
    }
}

void FlatView::mouseMoveEvent(QMouseEvent * e) {
    if (picking >= 0) {
        pickPos = oldpickPos + e->pos() - oldPos;
        game->visiblePieces[picking].curr = screen2board(pickPos);
        game->repaintAllViews();
    }
}

#include "flatview.moc"

