/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

// application specific includes
#include "slibo.h"
#include "game.h"
#include "pgn.h"
#include "glview.h"
#include "flatview.h"
#include "listview.h"
#include "statview.h"
#include "engine.h"

#include <pwd.h>
#include <unistd.h>
#include <sys/types.h>
#include <math.h>
// include files for Qt
#include <qdir.h>
#include <qwidget.h>
#include <qstring.h>
#include <qtimer.h>
#include <qregexp.h>
#include <qdeepcopy.h>
// include files for KDE
#include <klocale.h>
#include <kmessagebox.h>
#include <kio/job.h>
#include <kio/netaccess.h>
#include <ktempfile.h>

Game::Game(QWidget * parent, const char *name):
QObject(parent, name) {
    glView = new GLView(parent, this);
    flatView = new FlatView(parent, this);
    listView = new ListView(parent, this);
    hasUrl = false;
    speed = 0.15;
    
    newGame();

    QTimer *frameTimer = new QTimer(this);
    connect(frameTimer, SIGNAL(timeout()), SLOT(updatePieces()));
    frameTimer->start(20);
    
    connect (listView->lb, SIGNAL(clicked(int, int)), this, SLOT(slotGoto(int, int))); 
}

Game::~Game() {}

void Game::receive(Engine* i, QString fromengine) {
    while(fromengine[0].latin1() == '\t' || fromengine[0].latin1() == ' ' || fromengine[0].latin1() == '\n')
	fromengine = fromengine.mid(1);
    QString cmd = fromengine.section(" ", 0, 0).lower();
    if (cmd.isEmpty())
	return;
    //	    qDebug("%d", fromengine[0].latin1());
    QString arg = fromengine.section(" ", 1);
    /*            if (cmd != "") {
			    qDebug("from engine %p:", i);
			    qDebug("%s %s", cmd.latin1(), arg.latin1());
			}*/
    if (cmd == "feature")
	cmd_feature(i, arg);
    else if (cmd == "illegal")
	cmd_illegal(i, arg);
    else if (cmd == "error")
	cmd_error(i, arg);
    else if (cmd == "move")
	cmd_move(i, arg);
    else if (cmd == "my")
	cmd_mymove(i, arg);
    else if (cmd == "result")
	cmd_result(i, arg);
    else if (cmd == "resign")
	cmd_resign(i, arg);
    else if (cmd == "offer")
	cmd_offer(i, arg);
    else if (cmd == "telluser")
	cmd_telluser(i, arg);
    else if (cmd == "tellusererror")
	cmd_tellusererror(i, arg);
    else if (cmd == "askuser")
	cmd_askuser(i, arg);
    else if (cmd == "tellics")
	cmd_tellics(i, arg);
    else if (cmd == "stat")
	cmd_stat(i, arg);
    else if (cmd[0].isDigit())
	cmd_line(i, cmd+" "+arg);
    else
	cmd_default(i, cmd+" "+arg);
}

void Game:: cmd_line(Engine *engine, QString arg) {
    engine->addLine(arg);
}

void Game:: cmd_feature(Engine *engine, QString arg) {
    engine->features(arg);
}
void Game:: cmd_illegal(Engine */*engine*/, QString /*arg*/) { 
    KMessageBox::error(0,"Illegal Move", "Error");
}
void Game:: cmd_error(Engine */*engine*/, QString /*arg*/) { }
void Game:: cmd_mymove(Engine *engine, QString arg) {
    if (arg.left(9) == "move is: ") {
	cmd_move(engine, arg.mid(9));
    }
}
void Game:: cmd_move(Engine *engine, QString arg) {
    if (color == engine->getColor()) {
        VMove move;
        int match = strToVMove(arg, move);
	addSan(move);
        if (match == 0) {
            qDebug("illegal move %s from engine %s", arg.latin1(), engine->getName().latin1());
            return;
        }
        if (match>1) {
            qDebug("ambigous move %s from engine %s", arg.latin1(), engine->getName().latin1());
            return;
        }
        inputMove(move, engine);
    } else
        qDebug("out of turn move %s from %s", arg.latin1(), engine->getName().latin1());
}
void Game:: cmd_result(Engine */*engine*/, QString /*arg*/) { }
void Game:: cmd_resign(Engine */*engine*/, QString /*arg*/) { }
void Game:: cmd_offer(Engine */*engine*/, QString /*arg*/) { }
void Game:: cmd_telluser(Engine */*engine*/, QString /*arg*/) { }
void Game:: cmd_tellusererror(Engine */*engine*/, QString /*arg*/) { }
void Game:: cmd_askuser(Engine */*engine*/, QString /*arg*/) { }
void Game:: cmd_tellics(Engine */*engine*/, QString /*arg*/) { }
void Game:: cmd_stat(Engine *engine, QString arg) {
    engine->addText(arg);
}

void Game:: cmd_default(Engine *engine, QString arg) {
    qDebug("unknown command \"%s\" from engine %s", arg.ascii(), engine->getName().ascii());
}

/* compares Game::visibleBoard to the pieces displayed on the screen,
 if there are differences, set up moving/fading pieces to make
 the connected views look like Game::visibleBoard soon. */
void Game::updateVisibleBoard() {
    //first build a new list of Coors for each type of piece
    QValueList<CoorInt> coordsNew[13];
    int p;
    for (int x=0; x<8; x++)
        for (int y=0; y<8; y++)
            if ((p = visibleBoard[x][y]))
                coordsNew[p+6].append(CoorInt( x, y ));

    QValueList<CoorInt>::iterator j,k;

    for (int i=0; i<13; i++) {
	//for each type of piece, look if we find its new coor in the cuurent list
	//of coor. if we do, the piece is already at the correct position
	//remove it from both lists, so both lists contain only the differences
        for (j=coordsNew[i].begin(); j != coordsNew[i].end(); j++)
            if ((k = coords[i].find(*j)) != coords[i].end()) {
                coords[i].remove(k);
                j = coordsNew[i].remove(j);
                j--;
            }
	
	//if there are more pieces in the new list than in the current, we have to
	//add some pieces. set them up to fade in
        while (coordsNew[i].count() > coords[i].count()) {
            fadein(coordsNew[i].last(), i);
            coordsNew[i].remove(coordsNew[i].last());
        }
	
	//if there are move pieces in the current list, fade them out
        while (coordsNew[i].count() < coords[i].count()) {
            fadeout(coords[i].last(), i);
            coords[i].remove(coords[i].last());
        }
	
	//the reminder of pieces is of the correct type, but at the wrong position
	//set them up to slide to the new position.
        while (coords[i].count() > 0) {
            slide(coords[i].last(), coordsNew[i].last(), i);
            coordsNew[i].remove(coordsNew[i].last());
            coords[i].remove(coords[i].last());
        }
    }

    //build a current list of coords of pieces from scratch.
    for (int x=0; x<8; x++)
        for (int y=0; y<8; y++)
            if ((p = visibleBoard[x][y]))
                coords[p+6].append(CoorInt( x, y));

    repaintAllViews();
}

//update positions of moving pieces and opacity of fading pieces
//returns true, if something visible has changed.
//gets called some 50 times per second, but returns immediatly
//if nothing moves. to initiate a movement, changing has to be set to true.
void Game::updatePieces() {
    if (!changing)
        return;
    QValueList<ViewPiece>::iterator i;
    changing = false;
    for (i = visiblePieces.begin(); i != visiblePieces.end(); i++) {
        (*i).stime += speed;
        if ((*i).moving > 0) {
            (*i).curr = (*i).src + (*i).vec * ((*i).stime*(*i).stime / (*i).d);
            if ((*i).stime*(*i).stime > 0.5*(*i).d) {
                (*i).stime = (*i).stime - sqrt(2.0*(*i).d);
                (*i).moving = -1;
            }
            changing = true;
        }
        if ((*i).moving < 0) {
            (*i).curr = (*i).dst + (*i).vec * -((*i).stime*(*i).stime / (*i).d);
            if ((*i).stime >= 0.0) {
                movingChanged = true;
                (*i).moving = 0;
		(*i).curr = (*i).dst;
            }
            changing = true;
        }
        if ((*i).fading > 0) {
            (*i).alpha -= speed*0.5;
            if ((*i).alpha <= 0.0) {
                (*i).alpha = 0.0;
                (*i).fading = 0;
                movingChanged = true;
            }
            changing = true;
        }
        if ((*i).fading < 0) {
            (*i).alpha += speed*0.5;
            if ((*i).alpha >= 1.0) {
                i = visiblePieces.remove(i);
                i--;
                movingChanged = true;
            }
            changing = true;
        }
    }
    repaintAllViews();
}

// set up piece on square x1 to visibly move to x2
// does not change the internal board representation
void Game::slide(CoorInt src, CoorInt dst, int type) {
    QValueList<ViewPiece>::iterator i;
    for (i = visiblePieces.begin(); i != visiblePieces.end(); i++)
        if ((*i).pos == src && (*i).type == type-6)
            break;
    //    qDebug("%d %d %d %d", pos.x, pos.y, to.x, to.y);
    (*i).pos = dst;
    (*i).src = (*i).curr;
    (*i).dst.x = dst.x;
    (*i).dst.y = dst.y;
    (*i).vec = (*i).dst + (*i).src * -1.0;
    (*i).d = sqrt((*i).vec.x * (*i).vec.x + (*i).vec.y * (*i).vec.y);
    (*i).stime = 0.0;
    (*i).moving = 1;
    if ((*i).picked) {
	(*i).picked = false;
	glView->picking = -1;
	flatView->picking = -1;
    }
    movingChanged = true;
    changing = true;
}

//set up a piece to fade in
// does not change the internal board representation
void Game::fadein(CoorInt pos, int type) {
    ViewPiece newpiece;
    newpiece.type = type-6;
    newpiece.pos = pos;
    newpiece.curr.x = pos.x;    
    newpiece.curr.y = pos.y;
    newpiece.fading = 1;
    newpiece.moving = 0;
    newpiece.alpha = 1.0;
    visiblePieces.append(newpiece);
    movingChanged = true;
    changing = true;
}

//set up a piece to fade in
// does not change the internal board representation
void Game::fadeout(CoorInt x1, int type) {
    QValueList<ViewPiece>::iterator i;
    for (i = visiblePieces.begin(); i != visiblePieces.end(); i++)
        if ((*i).pos == x1 && (*i).type == type-6 && (*i).fading >= 0)
            break;
    (*i).fading = -1;
    (*i).alpha = 0.0;
    if ((*i).picked) {
	(*i).picked = false;
	glView->picking = -1;
	flatView->picking = -1;
    }
    movingChanged = true;
    changing = true;
}

void Game::clearBoard() {
    for (int x=0; x<8; x++)
        for (int y=0; y<8; y++)
            visibleBoard[x][y] = 0;
}

void Game::nextpos() {
    QTextStream stream(fp);
    QString s = stream.readLine();
    importEPD(s);
}

void Game::repaintAllViews() {
    if (glView)
        glView->repaintNeeded = true;
    if (flatView)
        flatView->repaintNeeded = true;
}

QString Game::squareStr(int x, int y) {    //returns the name of the square x,y in algebraic notation
    return QString(QChar(x + 'a')) + QString(QChar(y + '1'));
}

void Game::strSquare(QString sq, int &x, int &y) {
    x = sq[0].latin1() - 'A';
    if (x>7)
        x = x - ('a' - 'A');
    y = sq[1].latin1() - '1';
}

bool Game::isLegal(const VMove &m) {
    return legalMoves.contains(m);
}

VMoveList Game::generateLegalMoves() {
    VMoveList l;

    CoorInt v[] = {  CoorInt(1, 1), CoorInt(-1, 1), CoorInt(-1, -1), CoorInt(1, -1), CoorInt(1, 0), CoorInt(0, 1), CoorInt(-1, 0), CoorInt(0, -1)};
    CoorInt v2[] = {  CoorInt(2, 1), CoorInt(-2, 1), CoorInt(-2, -1), CoorInt(2, -1), CoorInt(1, 2), CoorInt(-1, 2), CoorInt(-1, -2), CoorInt(1, -2)};

    int direction, i;

    int lastrow, firstrow, secondrow, forward;

    for (int x = 0; x<8; x++)
        for (int y = 0; y<8; y++) {
            CoorInt src(x,y);
            CoorInt dst;
            if (color == WHITE) {
                lastrow = 6;
                firstrow = 0;
                secondrow = 1;
                forward = 1;
            } else {
                lastrow = 1;
                firstrow = 7;
                secondrow = 6;
                forward = -1;
            }

            switch (visibleBoard[x][y] * color) {
                case PAWN:
                    dst = CoorInt(x,y+forward);
                    if (!visibleBoard[dst])
                        if (y == lastrow) {
                            l << VMove(PAWN, src, dst, PROMOQ, 0);
                            l << VMove(PAWN, src, dst, PROMON, 0);
                            l << VMove(PAWN, src, dst, PROMOR, 0);
                            l << VMove(PAWN, src, dst, PROMOB, 0);
                        } else {
                            l << VMove(PAWN, src, dst, 0, 0);
                            dst = CoorInt(x, y+2*forward);
                            if ( y == secondrow && !visibleBoard[dst] )
                                l << VMove(PAWN, src, dst, PAWNDBL);
                        }

                    dst = CoorInt(x+1, y+forward);
                    if (x < 7 && visibleBoard[dst]*color < 0 )
                        if (y == lastrow) {
                            l << VMove(PAWN, src, dst, PROMOQ, visibleBoard[dst]);
                            l << VMove(PAWN, src, dst, PROMON, visibleBoard[dst]);
                            l << VMove(PAWN, src, dst, PROMOR, visibleBoard[dst]);
                            l << VMove(PAWN, src, dst, PROMOB, visibleBoard[dst]);
                        } else {
                            l << VMove(PAWN, src, dst, 0, visibleBoard[dst]);
                        };

                    if (x < 7 && y == lastrow-2*forward && visibleBoard[x+1][y]*color == -PAWN && visibleBoard[dst] == 0)
                        l << VMove(PAWN, src, dst, PAWNEP);

                    dst = CoorInt(x-1, y+forward);
                    if (x > 0 && visibleBoard[dst]*color < 0 )
                        if (y == lastrow) {
                            l << VMove(PAWN, src, dst, PROMOQ, visibleBoard[dst]);
                            l << VMove(PAWN, src, dst, PROMON, visibleBoard[dst]);
                            l << VMove(PAWN, src, dst, PROMOR, visibleBoard[dst]);
                            l << VMove(PAWN, src, dst, PROMOB, visibleBoard[dst]);
                        } else {
                            l << VMove(PAWN, src, dst, 0, visibleBoard[dst]);
                        };

                    if (x > 0 && y == lastrow-2*forward && visibleBoard[x-1][y]*color == -PAWN && visibleBoard[dst] == 0)
                        l << VMove(PAWN, src, dst, PAWNEP, -PAWN*color);

                    break;
                case BISHOP:
                    for (direction = 0; direction < 4; direction++) {
                        dst = src;
                        for (i = 1; i < 8; i++) {
                            dst = dst + v[direction];
                            if (dst.x > 7 || dst.x < 0 || dst.y > 7 || dst.y < 0 || visibleBoard[dst]*color > 0)
                                break;
                            l << VMove(BISHOP, src, dst, 0, visibleBoard[dst]);
                        }
                    }
                    break;
                case ROOK:
                    for (direction = 4; direction < 8; direction++) {
                        dst = src;
                        for (i = 1; i < 8; i++) {
                            dst = dst + v[direction];
                            if (dst.x > 7 || dst.x < 0 || dst.y > 7 || dst.y < 0 || visibleBoard[dst]*color > 0)
                                break;
                            l << VMove(ROOK, src, dst, 0, visibleBoard[dst]);
                        }
                    }
                    break;
                case QUEEN:
                    for (direction = 0; direction < 8; direction++) {
                        dst = src;
                        for (i = 1; i < 8; i++) {
                            dst = dst + v[direction];
                            if (dst.x > 7 || dst.x < 0 || dst.y > 7 || dst.y < 0 || visibleBoard[dst]*color > 0)
                                break;
                            l << VMove(QUEEN, src, dst, 0, visibleBoard[dst]);
                        }
                    }
                    break;
                case KING:
                    for (direction = 0; direction < 8; direction++) {
                        dst = src + v[direction];
                        if (dst.x > 7 || dst.x < 0 || dst.y > 7 || dst.y < 0 || visibleBoard[dst]*color > 0)
                            continue;
                        l << VMove(KING, src, dst, 0, visibleBoard[dst]);
                    }
                    if (x == 4 && y == firstrow && visibleBoard[7][firstrow]*color == ROOK && visibleBoard[6][firstrow]*color == 0 && visibleBoard[5][firstrow]*color == 0)
                        l << VMove(KING, src, src + CoorInt(2,0), CASTLEK);
                    if (x == 4 && y == firstrow && visibleBoard[0][firstrow]*color == ROOK && visibleBoard[1][firstrow]*color == 0 && visibleBoard[2][firstrow]*color == 0 && visibleBoard[3][firstrow]*color == 0)
                        l << VMove(KING, src, src + CoorInt(-2, 0), CASTLEQ);
                    break;
                case KNIGHT:
                    for (direction = 0; direction < 8; direction++) {
                        dst = src + v2[direction];
                        if (dst.x > 7 || dst.x < 0 || dst.y > 7 || dst.y < 0 || visibleBoard[dst]*color > 0)
                            continue;
                        l << VMove(KNIGHT, src, dst, 0, visibleBoard[dst]);
                    }
                    break;
            }
        }
    for (VMoveList::iterator i = l.begin(); i != l.end(); ++i) {
        (*i).ln.sprintf("%c%d%c%d", "abcdefgh"[(*i).src.x], (*i).src.y+1, "abcdefgh"[(*i).dst.x], (*i).dst.y+1);
        switch((*i).special) {
            case PROMOQ:
                (*i).ln += "q";
                break;
            case PROMOR:
                (*i).ln += "r";
                break;
            case PROMOB:
                (*i).ln += "b";
                break;
            case PROMON:
                (*i).ln += "n";
                break;
        }
    }
    return l;
};

void Game::inputMove(VMove &move, Engine *fromEngine) {
    addSan(move);
    writeToAllEngines(move.ln, fromEngine);
    for (Engine *i = engines.first(); i; i = engines.next())
	if (i != fromEngine)
	    i->clear();

    doMove(&move);
    gameMoves.insert(move);
    if (listView)
        listView->setMoveList(&gameMoves);
    updateVisibleBoard();
}

void Game::doMove(const VMove *move) {
    visibleBoard[move->dst] = visibleBoard[move->src];
    visibleBoard[move->src] = 0;
    switch(move->special) {
        case PROMOQ:
            visibleBoard[move->dst] = QUEEN * color;
            break;
        case PROMOR:
            visibleBoard[move->dst] = ROOK * color;
            break;
        case PROMOB:
            visibleBoard[move->dst] = BISHOP * color;
            break;
        case PROMON:
            visibleBoard[move->dst] = PROMON * color;
            break;
        case PAWNEP:
            visibleBoard[move->dst.x][move->dst.y + (color>0 ? -1 : 1)] = 0;
            break;
        case CASTLEK:
            if (color>0) {
                visibleBoard[7][0] = 0;
                visibleBoard[5][0] = ROOK;
            } else {
                visibleBoard[7][7] = 0;
                visibleBoard[5][7] = -ROOK;
            }
            break;
        case CASTLEQ:
            if (color>0) {
                visibleBoard[0][0] = 0;
                visibleBoard[3][0] = ROOK;
            } else {
                visibleBoard[0][7] = 0;
                visibleBoard[3][7] = -ROOK;
            }
            break;
    }
    color = -color;
    legalMoves = generateLegalMoves();
}

void Game::undoMove(const VMove *move) {
    color = -color;
    visibleBoard[move->src] = visibleBoard[move->dst];
    visibleBoard[move->dst] = move->capture;
    switch(move->special) {
        case PROMOQ:
        case PROMOR:
        case PROMOB:
        case PROMON:
            visibleBoard[move->src] = PAWN;
            break;
        case PAWNEP:
            visibleBoard[move->dst.x][move->dst.y + (color>0 ? -1 : 1)] = move->capture;
            visibleBoard[move->dst] = 0;
            break;
        case CASTLEK:
            if (color>0) {
                visibleBoard[7][0] = ROOK;
                visibleBoard[5][0] = 0;
            } else {
                visibleBoard[7][7] = -ROOK;
                visibleBoard[5][7] = 0;
            }
            break;
        case CASTLEQ:
            if (color>0) {
                visibleBoard[0][0] = ROOK;
                visibleBoard[3][0] = 0;
            } else {
                visibleBoard[0][7] = -ROOK;
                visibleBoard[3][7] = 0;
            }
            break;
    }
    legalMoves = generateLegalMoves();
}

void Game::undoMove() {
    VMove *move;
    if (gameMoves.prev(&move)) {
        for (Engine *engine = engines.first(); engine; engine = engines.next())
            engine->setColor(0);
        writeToAllEngines("undo");
        undoMove(move);
        listView->setMoveList(&gameMoves);
        updateVisibleBoard();
        repaintAllViews();
    }
}

void Game::redoMove(bool update) {
    VMove *move;
    if (gameMoves.next(&move)) {
        if (move->special == ONLY_SAN) {
            int result = strToVMove(move->san, *move);
            addSan(*move);
            if (result != 1) {
                qDebug("illegal move %s in redoMove, %d matches", move->san.latin1(), result);
                return;
            }
        }

	for (Engine *engine = engines.first(); engine; engine = engines.next()) {
            engine->setColor(0);
	    engine->clear();
	}
        writeToAllEngines(move->ln);
        doMove(move);
	if (update) {
	    listView->setMoveList(&gameMoves);
	    updateVisibleBoard();
	    repaintAllViews();
	}
    }
}

void Game::setURL(const KURL & url) {
    doc_url = url;
}

const KURL & Game::getURL() const {
    return doc_url;
}

bool Game::saveModified() {
    bool completed = true;
    if (modified) {
        QWidget *win = (QWidget *) parent();
        int want_save = KMessageBox::warningYesNoCancel(win,
                        i18n("The current file has been modified.\n" "Do you want to save it?"),
                        i18n("Warning"));
        switch (want_save) {
            case KMessageBox::Yes:
                if (doc_url.fileName() == i18n("Untitled")) {
                    ;
                    //		win->slotGameSaveAs();
                } else {
                    saveGame(doc_url);
                };

                completed = true;
                break;

            case KMessageBox::No:
                setModified(false);
                completed = true;
                break;

            case KMessageBox::Cancel:
                completed = false;
                break;

            default:
                completed = false;
                break;
        }
    }

    return completed;
}

void Game::closeGame() {}

bool Game::newGame() {
    modified = false;
    doc_url.setFileName(i18n("Untitled"));
    clearBoard();
    importEPD("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w QKqk -");
    writeToAllEngines("new");
    gameMoves.clear();
    gameMoves.current = &gameMoves;
    gameMoves.tagPairs["Event"] = "Casual game";
    gameMoves.tagPairs["Date"] = QDate::currentDate().toString("YYYY.MM.DD");
    gameMoves.tagPairs["Result"] = "*";
    gameMoves.tagPairs["Round"] = "?";
    gameMoves.tagPairs["Site"] = "?";
    return true;
}

void Game::setCaption() {
    gameMoves.tagPairs["White"] =  getpwuid(geteuid())->pw_name;
    gameMoves.tagPairs["Black"] = getpwuid(geteuid())->pw_name;
    for (Engine *i = engines.first(); i; i = engines.next()) {
        if (i->getColor() == WHITE)
            gameMoves.tagPairs["White"] = i->nickName;
        if (i->getColor() == BLACK)
            gameMoves.tagPairs["Black"] = i->nickName;
    }
    if (glView) glView->setCaption(gameMoves.tagPairs["White"] + " vs " + gameMoves.tagPairs["Black"]);
    if (flatView) flatView->setCaption(gameMoves.tagPairs["White"] + " vs " + gameMoves.tagPairs["Black"]);
    if (listView) listView->setCaption(gameMoves.tagPairs["White"] + " vs " + gameMoves.tagPairs["Black"]);
}

bool Game::openGame(const KURL & url, const char *format /*=0*/ ) {
    format = format;
    newGame();
    QString tmpfile;
    KIO::NetAccess::download(url, tmpfile);
    QFile file(tmpfile);
    if (file.open(IO_ReadOnly)) {
        if (url.filename().right(3).lower() == "epd") {
            QTextStream stream(&file);
            QString line;
            while (!stream.eof()) {
                line = stream.readLine();
            }
        }
        if (url.filename().right(3).lower() == "pgn")  {
	    gameMoves.importPGN(file);
		
            listView->setMoveList(&gameMoves);
        }
        file.close();
    }

    KIO::NetAccess::removeTempFile(tmpfile);

    modified = false;
    return true;
}

bool Game::saveGame(const KURL & url, const char */*format =0*/ ) {
    KTempFile tmpfile;
    *tmpfile.textStream() << gameMoves.exportPGN();
    tmpfile.close();
    KIO::NetAccess::upload(tmpfile.name(), url);
    modified = false;
    return true;
    tmpfile.unlink();
    return false;
}

/** No descriptions */
void Game::importEPD(QString s) {
    clearBoard();
    s.simplifyWhiteSpace();
    s.stripWhiteSpace();
    QString board = s.section(' ', 0, 0);
    QString activeColor = s.section(' ', 1, 1);
    QString castling = s.section(' ', 2, 2);
    ep = s.section(' ', 3, 3);

    for (int i = 0; i < 8; i++) {
        int x = 0;
        QString r = board.section('/', i, i);
        for (unsigned int j = 0; j < r.length(); j++) {
            switch (r[j].latin1()) {
                case '1'...'8':
                    x += r[j].latin1() - '1';
                    break;
                case 'k':
                    visibleBoard[x][7-i]= -KING;
                    break;
                case 'p':
                    visibleBoard[x][7-i]= -PAWN;
                    break;
                case 'b':
                    visibleBoard[x][7-i]= -BISHOP;
                    break;
                case 'n':
                    visibleBoard[x][7-i]= -KNIGHT;
                    break;
                case 'r':
                    visibleBoard[x][7-i]= -ROOK;
                    break;
                case 'q':
                    visibleBoard[x][7-i]= -QUEEN;
                    break;
                case 'K':
                    visibleBoard[x][7-i]= KING;
                    break;
                case 'P':
                    visibleBoard[x][7-i]= PAWN;
                    break;
                case 'B':
                    visibleBoard[x][7-i]= BISHOP;
                    break;
                case 'N':
                    visibleBoard[x][7-i]= KNIGHT;
                    break;
                case 'R':
                    visibleBoard[x][7-i]= ROOK;
                    break;
                case 'Q':
                    visibleBoard[x][7-i]= QUEEN;
                    break;
                default:
                    qDebug("EDP Import Error");
            }
            x++;
        }
    }


    if (activeColor.contains('W') || activeColor.contains('w'))
        color = 1;
    else
        color = -1;

    if (castling.contains('Q'))
        o_o |= CAST_WQ;
    if (castling.contains('K'))
        o_o |= CAST_WK;
    if (castling.contains('q'))
        o_o |= CAST_BQ;
    if (castling.contains('k'))
        o_o |= CAST_BK;

    QString options, tag, val;

    for (int i = 2; (options = s.section(' ', i * 2, i * 2 + 1)) != ""; i++) {
        options.truncate(options.length() - 1);
        tag = options.section(' ', 0, 0);
        val = options.section(' ', 1, 1);

        if (tag == "bm") {
            //	    setBreakMove(&m);
        }
    }
    updateVisibleBoard();
    legalMoves = generateLegalMoves();
}

/** No descriptions */
QString Game::exportEPD() {
    QString r;
    int x, t;
    char pieceLetters[] = "kpnqbr RBQNPK";

    for (int i = 0; i < 8; i++) {
        x = 0;
        for (int j = 0; j < 8; j++) {
            t = visibleBoard[i][j];
            if (!t)
                x++;
            else {
                if (x)
                    r += char ('0' + x);
                r += pieceLetters[t + 6];
            }
        }
        if (i < 7)
            r += '/';
    }

    r += ' ';
    r += color > 0 ? 'w' : 'b';

    r += ' ';
    if (o_o) {
        if (o_o & CAST_WQ)
            r += 'Q';
        if (o_o & CAST_WK)
            r += 'K';
        if (o_o & CAST_BQ)
            r += 'q';
        if (o_o & CAST_BK)
            r += 'k';
    } else
        r += '-';

    r += ' ';
    if (ep)
        r += ep;
    else
        r += '-';

    return r;
}

void Game::writeToAllEngines(QString txt, Engine *butThis) {
    for (Engine *i = engines.first(); i; i = engines.next()) {
        if (i != butThis)
            i->write(txt);
        if (i->delayedGo) {
            i->write("go");
            i->delayedGo = false;
        }
    }
}

int Game::matchMove(const VMove &m, VMove &matchmove) {
    int matchcount = 0;
    for (VMoveList::iterator i = legalMoves.begin(); i != legalMoves.end(); ++i) {
        if ((m.src.x == -1 || m.src.x == (*i).src.x)
                && (m.src.y == -1 || m.src.y == (*i).src.y)
                && (m.dst.x == -1 || m.dst.x == (*i).dst.x)
                && (m.dst.y == -1 || m.dst.y == (*i).dst.y)
                && (m.piece == -1 || m.piece == (*i).piece)
                && (m.special == -1 || m.special == (*i).special)) {
            matchcount++;
            matchmove = *i;
        }
    }
    return matchcount;
}

int Game::strToVMove(QString s, VMove &m) {
    m = VMove(-1, CoorInt(-1,-1), CoorInt(-1,-1), -1);

    //remove all useless stuff, whitespaces, dashes and so on.
    QRegExp regexp("[^a-h0-8rqnkoBRQNKO]");
    s.remove(regexp);

    //check for castling first, this is simple
    if (s == "00" || s == "oo" || s == "OO") {
        m.special = CASTLEK;
        return matchMove(m, m);
    }
    if (s == "000" || s == "ooo" || s == "OOO") {
        m.special = CASTLEQ;
        return matchMove(m, m);
    }
    //check for promotion and if so, remove trailing char
    char promo = s.right(1)[0].latin1();
    switch (promo) {
        case 'q':
        case 'Q':
            m.special = PROMOQ;
            s.truncate(s.length()-1);
            break;
        case 'b':
        case 'B':
            m.special = PROMOB;
            s.truncate(s.length()-1);
            break;
        case 'n':
        case 'N':
            m.special = PROMON;
            s.truncate(s.length()-1);
            break;
        case 'r':
        case 'R':
            m.special = PROMOR;
            s.truncate(s.length()-1);
            break;
    }

    //rest of move must have at least 2 chars. these 2 chars are always the destination square
    //remove them after converting them into coordinates.
    if (s.length() < 2)
        return 0;
    m.dst.x = s.right(2)[0].latin1() - 'a';
    if (m.dst.x < 0 && m.dst.x > 7)
        return 0;
    m.dst.y = s.right(2)[1].latin1() - '1';
    if (m.dst.y < 0 && m.dst.y > 7)
        return 0;
    s.truncate(s.length()-2);

    VMove piecemove, algebraicmove;
    int piecematch, algebraicmatch;
    switch (s.length()) {
        case 0:
            //2 char move: a move like "e4" or "h8=q", always a pawn move
            m.piece = PAWN;
            return matchMove(m, m);

        case 1:
            //3 char move: first test if this is somthing like fxe4, which is
            //a pawn move. note that a move like bxc4 can also be a bishop move.
            if (s[0].latin1() >= 'a' && s[0].latin1() <= 'h') {
                m.src.x = s[0].latin1()-'a';
                //don't take moves like bb4 as a pawn move, this has to be a bishop move
                if (m.src.x != m.dst.x) {
                    m.piece = PAWN;
                    algebraicmatch = matchMove(m, algebraicmove);
                } else
                    algebraicmatch = 0;
                m.src.x = -1;
            } else
                algebraicmatch = 0;

            //now test for piece moves like Nf3 or Bxc4.
        piecematchtest:
            switch (s[0].latin1()) {
                case 'k':
                case 'K':
                    m.piece = KING;
                    piecematch = matchMove(m, piecemove);
                    break;
                case 'b':
                    //if we already got a pawn-match, we got something like ba4.
                    //take this move and do not try bxa4 as a bishop move,
                    //because we have to assume the input is not a bishop move.
                    //otherwise the correct bxa4 would be rejected due to ambiguity
                    //if a move like Bxa4 is possible at the same time.
		    if (algebraicmatch) {
			piecematch = 0;
                        break;
		    }
                case 'B':
                    m.piece = BISHOP;
                    piecematch = matchMove(m, piecemove);
                    break;
                case 'r':
                case 'R':
                    m.piece = ROOK;
                    piecematch = matchMove(m, piecemove);
                    break;
                case 'N':
                case 'n':
                    m.piece = KNIGHT;
                    piecematch = matchMove(m, piecemove);
                    break;
                case 'Q':
                case 'q':
                    m.piece = QUEEN;
                    piecematch = matchMove(m, piecemove);
                    break;
                case 'a':
                case 'c' ... 'h':
                    piecematch = 0;
                    break;
                default:
                    return 0;
            }
            //if more than exactly one move matches, the move is ambigous.
            //exception: exatcly one algebraic and one piecemove matches and
            //both are the same, this is ok (i. e. b1c2, if there is a bishop on b1).
            //if none matches, it is illegal.
            if (algebraicmatch * piecematch == 1) {
                if (algebraicmove == piecemove)
                    return 1;
                else
                    return 2;
            }
            if (algebraicmatch + piecematch > 1)
                return 2;
            if (algebraicmatch) {
                m = algebraicmove;
                return 1;
            }
            if (piecematch) {
                m = piecemove;
                return 1;
            }
            return 0;

        case 2:
            //4 char moves, like e2e4 (long algebraic), Nfd2 or R1e6
            //first check second char. it is always a coordinate,
            //either a file (letter) or a row (digit).
            if (s[1].latin1() >= '1' && s[1].latin1() <= '8')
                m.src.y = s[1].latin1()-'1';
            else if (s[1].latin1() >= 'a' && s[1].latin1() <= 'h') {
                m.src.x = s[1].latin1()-'a';
            } else
                return 0;

            //if the second char was a row, check for a..h in first char (algebraic notation)
            //reset source file after match so we can match a move like B4c5 (bishop move, improbable)
            //in the following piece match
            if (m.src.x == -1 && s[0].latin1() >= 'a' && s[0].latin1() <= 'h') {
                m.src.x = s[0].latin1()-'a';
                algebraicmatch = matchMove(m, algebraicmove);
                m.src.x = -1;
            } else
                algebraicmatch = 0;

            //in any case check if the first char is a piece letter
            goto piecematchtest;

        case 3:
            //five char moves, Nb1-c3 and like.
            if (s[2].latin1() >= '1' && s[2].latin1() <= '8')
                m.src.y = s[2].latin1()-'1';
            else
                return 0;
            if (s[1].latin1() >= 'a' && s[1].latin1() <= 'h') {
                m.src.x = s[1].latin1()-'a';
            } else
                return 0;
            algebraicmatch = 0;
            goto piecematchtest;

        default:
            return 0;
    }
}

void Game::addSan(VMove &move) {
    if (move.special == CASTLEQ) {
        move.san = "O-O-O";
        return;
    }
    if (move.special == CASTLEK) {
        move.san = "O-O";
        return;
    }

    VMove dummy;

    if (move.piece != PAWN) {
        move.san = " RBQNPK"[move.piece];
    } else {
        if (move.capture)
            move.san = "abcdefgh"[move.src.x];
        else
            move.san = "";
    }
    if (move.capture)
        move.san += "x";
    move.san += "abcdefgh"[move.dst.x];
    move.san += "12345678"[move.dst.y];

    if (strToVMove(move.san, dummy) == 1)
        return;

    if (move.piece != PAWN) {
        move.san = " RBQNPK"[move.piece];
    } else
        move.san = "";
    move.san += "abcdefgh"[move.src.x];
    if (move.capture)
        move.san += "x";
    move.san += "abcdefgh"[move.dst.x];
    move.san += "12345678"[move.dst.y];
    if (strToVMove(move.san, dummy) == 1)
        return;

    if (move.piece != PAWN) {
        move.san = " RBQNPK"[move.piece];
    } else {
        if (move.capture)
            move.san = "abcdefgh"[move.src.x];
        else
            move.san = "";
    }
    move.san += "12345678"[move.src.y];
    if (move.capture)
        move.san += "x";
    move.san += "abcdefgh"[move.dst.x];
    move.san += "12345678"[move.dst.y];
    if (strToVMove(move.san, dummy) == 1)
        return;

    if (move.piece != PAWN) {
        move.san = " RBQNPK"[move.piece];
    } else {
        move.san = "";
    }
    move.san += "abcdefgh"[move.src.x];
    move.san += "12345678"[move.src.y];
    if (move.capture)
        move.san += "x";
    move.san += "abcdefgh"[move.dst.x];
    move.san += "12345678"[move.dst.y];
    return;
}

void Game::slotGoto(int , int c) {
    clearBoard();
    importEPD("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w QKqk -");
    gameMoves.current = &gameMoves;
    writeToAllEngines("new");
    int move = 0;
    while (listView->movePos[move] < c) {
	redoMove(listView->movePos[move+1] >= c);
	move++;
    }
    qDebug("move %d", move);
    
}

void Game::changeSides() {
    Engine *wengine = NULL;
    Engine *bengine = NULL;
    for (Engine*  i=engines.first(); i; i=engines.next()) {
	if (i->getColor() == WHITE) 
	    wengine = i;
	if (i->getColor() == BLACK)
	    bengine = i;
	i->setColor(0);
    }
    if (wengine) wengine->setColor(BLACK);
    if (bengine) bengine->setColor(WHITE);
}
#include "game.moc"

