/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#ifndef Game_H
#define Game_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "view.h"
#include "pgn.h"

// include files for QT
#include <qobject.h>
#include <qstring.h>
#include <qlist.h>
#include <qvaluelist.h>
#include <qfile.h>

// include files for KDE
#include <kurl.h>


#define ROOK 1
#define BISHOP 2
#define QUEEN 3
#define KNIGHT 4
#define PAWN 5
#define KING 6

#define WHITE 1
#define BLACK -1

#define PROMOQ  0x1
#define PROMON  0x2
#define PROMOB  0x3
#define PROMOR  0xb       //PROMOR must be ROOKM | something
#define PAWNEP  0x5
#define CASTLEQ 0x6
#define CASTLEK 0x7
#define PAWNDBL 0x9
#define ROOKM   0x8
#define KINGM   0xa
#define ONLY_SAN 0x10

#define CAST_WQ 0x1
#define CAST_WK 0x2
#define CAST_BQ 0x4
#define CAST_BK 0x8

// forward declaration of the Slibo classes
class Info;
class Test;
class MoveTree;
class slibo;
class ListView;
class GLView;
class FlatView;
class Engine;
class Pgn;

/**	Game provides a document object for a document-view model.
  *
  * The Game class provides a document object that can be used in conjunction with the classes SliboApp and View
  * to create a document-view model for standard KDE applications based on KApplication and KMainWindow. Thereby, the document object
  * is created by the SliboApp instance and contains the document structure with the according methods for manipulation of the document
  * data by View objects. Also, Game contains the methods for serialization of the document data from and to files.
  *
  * @author Source Framework Automatically Generated by KDevelop, (c) The KDevelop Team.
  * @version KDevelop version 1.2 code generation
  */

typedef QValueList<VMove> VMoveList;

class Game:public QObject {
  Q_OBJECT public:
    /** Constructor for the fileclass of the application */
     Game(QWidget * parent, const char *name = 0);
    /** Destructor for the fileclass of the application */
    ~Game();

    /** sets the modified flag for the document after a modifying action on the view connected to the document.*/
    void setModified(bool _m = true) { modified = _m; };
    /** returns if the document is modified or not. Use this to determine if your document needs saving by the user on closing.*/
    bool isModified() { return modified; };
    /** "save modified" - asks the user for saving if the document is modified */
    bool saveModified();
    /** initializes the document generally */
    bool newGame();
    /** closes the acutal document */
    void closeGame();
    /** loads the document by filename and format and emits the updateViews() signal */
    bool openGame(const KURL & url, const char *format = 0);
    /** saves the document under filename and format.*/
    bool saveGame(const KURL & url, const char *format = 0);
    /** returns the KURL of the document */
    const KURL & getURL() const;
    /** sets the URL of the document */
    void setURL(const KURL & url);
  /** No descriptions */
    void importEPD(QString f);
  /** No descriptions */
    QString exportEPD();

    //executes a move, coming from user input or a engine.
    //updates everything, including the current move pointer and the engines
    void        inputMove(VMove &, Engine *fromEngine = 0);
    
    //takes back (if possible) one move. updates everything.
    void        undoMove();
    
    //redo a previously taken back move.  updates everything. 
    void        redoMove(bool update=true);
    
    void        strSquare(QString sq, int &, int &);
    QString     squareStr(int x, int y);
    bool isLegal(const VMove &);
    //adds the san (standard algebraic notation) member 
    void addSan(VMove &);
    
    GLView* getGLView() { return glView; }
    FlatView* getFlatView( )  { return flatView; }
    ListView* getListView( )  { return listView; }
    
    void repaintAllViews();
    VMoveList generateLegalMoves();
    void setCaption();

    void writeToAllEngines(QString txt, Engine *butThis = 0);
    int matchMove(const VMove &m, VMove &matchmove);
    int strToVMove(QString s, VMove &m);
    
    void changeSides();
    
    QValueList<ViewPiece> visiblePieces;
    QPtrList<Engine> engines;

  /**  */
    float speed;
  /**  */
    struct {
        int vb[8][8];
        int& operator [] (const CoorInt p) {
            return vb[p.x][p.y];
        }
        int* operator [] (const int x) {
            return vb[x];
        }
    } visibleBoard;
    int color;
    int moveNr;
    int o_o;
    bool hasUrl;
    QString ep;
    VMoveList legalMoves;
    Pgn gameMoves;

    bool movingChanged;
    
public slots:
    void receive(Engine*, QString);

private:
  //default constructor disabled, to inhibit use as a auto variable
    Game();
    void clearBoard();
    /** the modified flag of the current document */
    bool modified;
    bool changing;
    KURL doc_url;
    QFile *fp;
    GLView *glView;
    FlatView *flatView;
    ListView *listView;
    QValueList<CoorInt> coords[13];
    void updateVisibleBoard();
    void slide(CoorInt, CoorInt, int);
    void fadein(CoorInt, int type);
    void fadeout(CoorInt, int type);
    //internal do/undo routines
    //updates only visibleBoard and color according to move
    void        doMove(const VMove *);
    void        undoMove(const VMove *);
    void cmd_line(Engine *engine, QString arg);
    void cmd_feature(Engine *engine, QString arg);
    void cmd_illegal(Engine *engine, QString arg);
    void cmd_error(Engine *engine, QString arg);
    void cmd_move(Engine *engine, QString arg);
    void cmd_mymove(Engine *engine, QString arg);
    void cmd_result(Engine *engine, QString arg);
    void cmd_resign(Engine *engine, QString arg);
    void cmd_offer(Engine *engine, QString arg);
    void cmd_telluser(Engine *engine, QString arg);
    void cmd_tellusererror(Engine *engine, QString arg);
    void cmd_askuser(Engine *engine, QString arg);
    void cmd_tellics(Engine *engine, QString arg);
    void cmd_stat(Engine *engine, QString arg);
    void cmd_default(Engine *engine, QString arg);
    
  private slots:
    void nextpos();
    void updatePieces();
    void slotGoto(int, int);

};

#endif				// Game_H
