/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#ifndef GLVIEW_H
#define GLVIEW_H

#include "view.h"
#include <qgl.h>
#include <qstringlist.h>

class Game;

class GLView: public QGLWidget, public View {
Q_OBJECT public:
    typedef GLfloat GLcolor[4];

    GLView(QWidget * parent , Game *);
    virtual ~GLView();
    void print(QPrinter * pPrinter);
    void rotate180();
    QPoint viewAngle, viewPos;
    GLfloat viewDistance;

    GLfloat randomtable[64];
    QStringList legalMoves;
    QString *lastLegalMove;
    bool reflections, fading, textures, shadows;
    GLfloat lightPosition[4];

protected slots:
    void frame();

protected:
    CoorFloat screen2board(QPoint);
    QPoint board2screen(CoorFloat);
    int selection(QPoint);
    void mousePressEvent(QMouseEvent * e);

    GLuint initTexture(QString);
    void initializeObjects();
    void mouseMoveEvent(QMouseEvent * e);
    void extrude(GLfloat * x, GLfloat * y, int c, GLfloat d);
    void wheelEvent(QWheelEvent * e);
    void paintGL();
    void initializeGL();
    void resizeGL(int w, int h);
    void revolve(GLfloat * xx, GLfloat * yy, int count);
    void render();
    void rotate();
    void renderPieces(float, int);
    void boardEdge(GLfloat x0, GLfloat y0, GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2, GLfloat x3, GLfloat y3);

    GLfloat aspect;
    GLuint texWwood, texBwood, texWmarble, texBmarble, texCoor[16];
    QPoint oldviewPos;
    QPoint oldviewAngle;
    GLuint selectionBufferSize;
    GLuint *selectBuffer;	/* select buffer */

    GLint dlBoard,		/* display list for chessboard */
    dlPiece[7], dlCoor, dlBoardFrame, dl4edges;
    GLfloat groundplane[4];
    QTimer *frameTimer;
    int rotateSteps;
    
signals:
    void enterEvent(QEvent *);
    void leaveEvent(QEvent *);
};

#endif
