/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@cruncher                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "pgn.h"
#include "game.h"

#include <ctype.h>


Pgn::Pgn() {
    current = this;
}

void Pgn::importPGN(QFile &f) {
    getMoveTree(f, this);
}

QString Pgn::exportPGN() {
    QString s;
    for (QMap<QString,QString>::iterator i = tagPairs.begin(); i != tagPairs.end(); ++i) 
	s += "[" + i.key() + " \"" + i.data() + "\"]\n";
    s += '\n';
    s += moveList();
    return s;
}

QString Pgn::moveList() {
    QString s;
    int m=2;
    for (MoveTree *i = this; i->moves.count(); i = i->moves.first().nextMove, m++) {
	if (!(m&1)) {
            s += QString::number(m/2)+". ";
	}
	s += i->moves.first().san+" ";
    }
    s += "*";
    return s;
}

void Pgn::getMoveTree(QFile &f, MoveTree *mt) {
    while (!f.atEnd()) {
    QString token = getToken(f);
    if (token == "")	
	break;		
    if (token == "[") {
	QString tag = getToken(f);
	QString val = getToken(f);
	tagPairs[tag] = val;
	while ((token = getToken(f)) != "]");
	continue;
    }
    if (token == "{") {
	while ((token = getToken(f)) != "}");
	continue;
    }	
    bool isNum;
    token.toInt(&isNum);
    if (isNum) {
	continue;
    }
    if (token == ".")
	continue;
    if (token == "*" || token == "1-0" || token == "0-1" || token == "1/2-1/2") 
	return;
    
	    

    MoveNode mn;
    if (token == "(") {
	getMoveTree(f, mt);
	continue;
    }
    if (token == ")")
	return;

    mn.special = ONLY_SAN;
    mn.san = token;
    mn.nextMove = new MoveTree;
    mn.nextMove->prevMove = mt;
    mt->moves.prepend(mn);
    getMoveTree(f, mn.nextMove);
    return;
    }
}

MoveTree::~MoveTree() {
    while (moves.count()) {
        delete moves.first().nextMove;
        moves.pop_front();
    }
}
//MoveNode *Pgn::addMove(QString s) {
//}

QString Pgn::getToken(QFile &f) {
    bool escape = false;
    bool instring = false;
    
    QString token = "";
    while (!f.atEnd()) {
        int c=f.getch();
	if (c < 0) break;
	if (!escape) {
            if (!instring) {
		if (isspace(c)) {
                    if (token.isEmpty())
                        continue;
                    return token;
		} else if (c=='[' || c==']' || c == '(' || c == ')' || c == '.' || c=='*' || c=='{' || c=='}') {
                    if (token.isEmpty())
                        return QString(QChar(c));
                    else {
                        f.ungetch(c);
                        return token;
                    }

		}
	    }
	    if (c == '\\') {
		escape = true;
		continue;
	    } else if (c == '"') {
		instring = !instring;
	    }
	}
        token += char(c);
        escape = false;
    }
    return token;
}

void Pgn::clear() {
    while (moves.count()) {
        delete moves.first().nextMove;
        moves.pop_front();
    }
}

bool Pgn::next(VMove **move) {
    if (current->moves.count()) {
	*move = &current->moves.first();
	current = current->moves.first().nextMove;
	return true;
    }
    return false;
}

bool Pgn::prev(VMove **move) {
    if (current == this)
	return false;
    current = current->prevMove;
    *move = &current->moves.first();
    return true;
}

int Pgn::count() {
    int m=0;
    for (MoveTree *i = this; i->moves.count(); i = i->moves.first().nextMove, m++)
	;
    return m;
}

void Pgn::insert(VMove m) {
    qDebug("insert");
    MoveNode mn; 
    mn.capture = m.capture;
    mn.dst = m.dst;
    mn.ln = m.ln;
    mn.piece = m.piece;
    mn.san = m.san;
    mn.special = m.special;
    mn.src = m.src;
    mn.nextMove = new MoveTree;
    mn.nextMove->prevMove = current;
//    current->moves.append(mn);
    current->moves.prepend(mn);
    current = mn.nextMove;
}

