/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@cruncher                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef PGN_H
#define PGN_H

/**
@author Gunther Piez
*/

#include "view.h"
#include <qmap.h>
#include <qfile.h>

class MoveTree;

//struct for internal move representation.
//contains a constructor and a comparison operator
struct VMove {
    CoorInt src, dst;
    int special, capture, piece;
    QString san,ln;

    VMove() {};
    VMove(int p, CoorInt s, CoorInt d, int sp=0, int cap=0) {
        src = s;
        dst = d;
        special = sp;
        capture = cap;
        piece = p;
    }
    bool operator == (const VMove a) const {
        return src == a.src && dst == a.dst;
    }
};


struct MoveNode: public VMove {
    MoveTree *nextMove;
};

class MoveTree {
public:

    ~MoveTree();
    MoveTree *prevMove;
    QValueList<MoveNode> moves;

};

class Pgn: public MoveTree {
public:
    Pgn();
    
    MoveTree *getCurrent() { return current; }
    bool next(VMove **);
    bool prev(VMove **);
    void clear();
    void importPGN(QFile &);
    void getMoveTree(QFile &, MoveTree *mt);
    int count();
    void insert(VMove);
    QString exportPGN();
    QString moveList();
    QString getToken(QFile &);
    
    QMap<QString, QString> tagPairs;
    MoveTree *current;
};
#endif
