/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@cruncher                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "pgndb.h"
#include "pgn.h"

#include <cstdio>
#include <sqlite.h>

#include <qstring.h>
#include <qdatetime.h>

#include <kstandarddirs.h>

PgnDB::PgnDB()
{
    char *zErrMsg = 0;
    int rc;
    
    QString dbfile = locateLocal("appdata", "pgn.db");
    db = sqlite_open(dbfile.latin1(), 0, &zErrMsg);
    
    sqlite_exec(db,"PRAGMA default_synchronous = OFF;", NULL, NULL, &zErrMsg);
    
    int nrow;
    int ncolumn;
    char **result;
    rc = sqlite_get_table(db, "SELECT version FROM version", &result, &nrow, &ncolumn, &zErrMsg);
    if( rc!=SQLITE_OK ) fprintf(stderr, "SQL error: %s\n", zErrMsg);

    if (ncolumn == 0) {
	rc = sqlite_exec(db, "CREATE TABLE version(version TEXT);", NULL, NULL, &zErrMsg);
	if( rc!=SQLITE_OK ) fprintf(stderr, "SQL error: %s\n", zErrMsg);
	rc = sqlite_exec(db, "INSERT INTO version VALUES('0.4.4');", NULL, NULL, &zErrMsg);
	if( rc!=SQLITE_OK ) fprintf(stderr, "SQL error: %s\n", zErrMsg);
	
	rc = sqlite_exec(db, "CREATE TABLE pgn(ctime TIMESTAMP, event TEXT, site TEXT, date DATE, round TEXT,\
					       white TEXT, black TEXT, result TEXT, whiteelo INT,\
					       blackelo INT, eco TEXT, tags TEXT, notation TEXT);",
			 NULL, NULL, &zErrMsg);
	if( rc!=SQLITE_OK ) fprintf(stderr, "SQL error: %s\n", zErrMsg);
    }
    sqlite_free_table(result);
}

//removes " at beginning or end, transforms all
// ' into ", returns the string encapsulated in '
QString PgnDB::trans(QString str) {
    if (str.left(1) = "\"")
	str = str.mid(1);
    if (str.right(1))
	str.setLength(str.length()-1);
    str.replace("'", "\"");
    return '\'' + str + '\'';
}
void PgnDB::insert(Pgn *pgn) {
    QString sql = "INSERT INTO pgn VALUES('" + QDateTime::currentDateTime().toString(Qt::ISODate)
		  + "'," + trans(pgn->tagPairs["Event"])
		  + "," + trans(pgn->tagPairs["Site"])
		  + "," + trans(pgn->tagPairs["Date"])
		  + "," + trans(pgn->tagPairs["Round"])
		  + "," + trans(pgn->tagPairs["White"])
		  + "," + trans(pgn->tagPairs["Black"])
		  + "," + trans(pgn->tagPairs["Result"])
		  + "," + trans(pgn->tagPairs["WhiteElo"])
		  + "," + trans(pgn->tagPairs["BlackElo"])
		  + "," + trans(pgn->tagPairs["ECO"])
		  + ",'','" + pgn->moveList() + "');"; 
    printf("%s\n", sql.latin1());
    char *zErrMsg = 0;
    int rc = sqlite_exec(db, sql.latin1(), NULL, NULL, &zErrMsg);
    if( rc!=SQLITE_OK ) fprintf(stderr, "SQL error: %s\n", zErrMsg);
}

PgnDB::~PgnDB()
{
    sqlite_close(db);
}


