/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "pref.h"
#include "slibo.h"
#include "game.h"
#include "glview.h"

#include <klocale.h>

#include <qlayout.h>
#include <qlabel.h>
#include <qbuttongroup.h>
#include <qcheckbox.h>
#include <qtooltip.h>
#include <qslider.h>
#include <qwhatsthis.h>

SliboPreferences::SliboPreferences(slibo *app)
    : KDialogBase(TreeList, "Slibo Preferences",
                  Help|Default|Ok|Apply|Cancel, Ok)
{
    // this is the base class for your preferences dialog.  it is now
    // a Treelist dialog.. but there are a number of other
    // possibilities (including Tab, Swallow, and just Plain)
    QFrame *frame;
    frame = addPage(i18n("View"), i18n("View Options"));
    m_pageOne = new SliboPrefPageOne(frame, app);

    frame = addPage(i18n("Second Page"), i18n("Page Two Options"));
    m_pageTwo = new SliboPrefPageTwo(frame, app);


}

SliboPrefPageOne::SliboPrefPageOne(QWidget *parent, slibo *app)
    : QFrame(parent)
{
    theApp = app;
    //make a central widget to contain the other widgets
    central = parent;

    // Create a layout to position the widgets
    QHBoxLayout *grid = new QHBoxLayout( central, 10 );

    // Create a group of check boxes
    bg = new QButtonGroup( central, "checkGroup" );
    bg->setTitle( "Rendering" );
    grid->addWidget( bg, 1, 0 );

    // Create a layout for the check boxes
    QVBoxLayout *vbox = new QVBoxLayout(bg, 10);

    vbox->addSpacing( bg->fontMetrics().height() );

    app->cb[0] = new QCheckBox( bg );
    app->cb[1] = new QCheckBox( bg );
    app->cb[2] = new QCheckBox( bg );
    app->cb[0]->setText( "&Shadows" );
    app->cb[1]->setText( "&Reflection" );
    app->cb[2]->setText( "&Textures" );
    vbox->addWidget( app->cb[0] );
    vbox->addWidget( app->cb[1] );
    vbox->addWidget( app->cb[2] );
    QToolTip::add( app->cb[0], "Enables or disables rendering of shadows. This can improve rendering speed by one third." );
    QToolTip::add( app->cb[1], "Enables or disables rendering of reflections. This can improve rendering speed by one third." );
    QToolTip::add( app->cb[2], "Enables or disables rendering of textures. This will improve rendering speed slightly." );

    vbox = new QVBoxLayout(grid, 10);

    vbox->addWidget(new QLabel(QString("Light Position"), central));

    hbox = new QHBoxLayout(vbox, 10);
//    vbox->addLayout(hbox);

}

void SliboPrefPageOne::connectWidgets() {
    disconnect(bg);
    disconnect(theApp->sb[0]);
    disconnect(theApp->sb[1]);
    disconnect(theApp->sb[2]);
    
    GLView *glView;
    if ((glView = theApp->activeGame->getGLView())) {
        connect( bg, SIGNAL(clicked(int)), theApp, SLOT(checkBoxClicked()) );

//TODO: move the cb/sb stuff to this class, where it belongs
//better: remove the whole bullshit and do it new with qtdesigner
        theApp->cb[0]->setChecked(glView->shadows);
        theApp->cb[1]->setChecked(glView->reflections);
        theApp->cb[2]->setChecked(glView->textures);

        theApp->sb[0] = new QSlider( 0, 1500, 30, 1500-glView->lightPosition[1]*100.0, QSlider::Vertical, central, "Slider" );
        theApp->sb[1] = new QSlider(-1500, 1500, 30, glView->lightPosition[0]*100.0, QSlider::Horizontal, central, "Slider" );
        theApp->sb[2] = new QSlider( -1500, 1500, 30, glView->lightPosition[2]*100.0, QSlider::Vertical, central, "Slider" );

        theApp->sb[0]->setFocusPolicy( QWidget::TabFocus );
        theApp->sb[1]->setFocusPolicy( QWidget::TabFocus );
        theApp->sb[2]->setFocusPolicy( QWidget::TabFocus );

        hbox->addWidget( theApp->sb[0] );
        hbox->addWidget( theApp->sb[1] );
        hbox->addWidget( theApp->sb[2] );

        connect( theApp->sb[0], SIGNAL(valueChanged(int)), theApp, SLOT(sliderValueChanged()) );
        connect( theApp->sb[1], SIGNAL(valueChanged(int)), theApp, SLOT(sliderValueChanged()) );
        connect( theApp->sb[2], SIGNAL(valueChanged(int)), theApp, SLOT(sliderValueChanged()) );

        QToolTip::add( theApp->sb[0], "Vertical Light" );
        QToolTip::add( theApp->sb[1], "Horizontal Light" );
        QToolTip::add( theApp->sb[2], "Distance Light" );

        QWhatsThis::add( theApp->sb[0], " This slider controls the vertical position of the light source." );
        QWhatsThis::add( theApp->sb[1], " This slider controls the horizontal position of the light source." );
        QWhatsThis::add( theApp->sb[2], " This slider controls the deep of the light source." );
    }
}

SliboPrefPageTwo::SliboPrefPageTwo(QWidget *parent, slibo *app)
    : QFrame(parent)
{
    QHBoxLayout *layout = new QHBoxLayout(this);
    layout->setAutoAdd(true);

    new QLabel("Add something here", this);
    new QLabel("Add something here", this);
    new QLabel("Add something here", this);
    new QLabel("Add something here", this);
}
#include "pref.moc"
