/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "slibo.h"
#include "glview.h"
#include "flatview.h"
#include "listview.h"
#include "statview.h"
#include "pref.h"
#include "game.h"
#include "engine.h"
#include "playlevel.h"
#include "pgndb.h"

#include <qdragobject.h>
#include <qlineedit.h>
#include <qprinter.h>
#include <qprintdialog.h>
#include <qpainter.h>
#include <qpaintdevicemetrics.h>
#include <qtimer.h>
#include <qworkspace.h>
#include <qvbox.h>
#include <qslider.h>
#include <qcheckbox.h>
#include <qtextedit.h>
#include <qradiobutton.h>
#include <qspinbox.h>
#include <qtabwidget.h>
#include <qfiledialog.h>

#include <kglobal.h>
#include <klocale.h>
#include <kiconloader.h>
#include <kmenubar.h>
#include <kstatusbar.h>
#include <kkeydialog.h>
#include <kaccel.h>
#include <kio/netaccess.h>
#include <kfiledialog.h>
#include <kconfig.h>
#include <kurl.h>
#include <kurlrequesterdlg.h>
#include <kfiledialog.h>
#include <kedittoolbar.h>
#include <kstdaccel.h>
#include <kaction.h>
#include <kstdaction.h>
#include <kpopupmenu.h>

#define STATUSBAR_RIGHT 1
slibo::slibo():
    KMainWindow( 0, "slibo" ),
    activeGame(NULL),
    m_printer(0),
    secondClose(false),
    notInitialized(true)
{
    //at the very first beginning, scan for engines, to have the names available for menus etc
    scanForEngines();

    //setup our mdi-workspace
    ws = new QWorkspace( this );
    //    ws->setScrollBarsEnabled( false );
    ws->setBackgroundMode(PaletteDark);
    setCentralWidget( ws );
    // accept dnd
    setAcceptDrops(true);

    // then, setup our actions
    setupActions();

    // and a status bar
    statusBar()->insertItem("", STATUSBAR_RIGHT, 0, true);
    statusBar()->show();

    playLevel = new PlayLevel(this);
    sliboPreferences = new SliboPreferences(this);    

    connect((QObject *)playLevel->buttonOk, SIGNAL(clicked()), this, SLOT(slotPlayLevelAccept()));
    connect((QObject *)playLevel->engineListBox, SIGNAL(selectionChanged()), this, SLOT(slotPlayLevelSelectionChanged()));
    connect((QObject *)playLevel->blitz4min2, SIGNAL(clicked()), this, SLOT(slotPlayLevelSet4_2()));
    connect((QObject *)playLevel->blitz5min, SIGNAL(clicked()), this, SLOT(slotPlayLevelSet5()));
    connect((QObject *)playLevel->blitz30min15, SIGNAL(clicked()), this, SLOT(slotPlayLevelSet30_15()));
    connect((QObject *)playLevel->blitz30min, SIGNAL(clicked()), this, SLOT(slotPlayLevelSet30()));
    connect((QObject *)playLevel->long5min, SIGNAL(clicked()), this, SLOT(slotPlayLevelSetLong5()));
    connect((QObject *)playLevel->long30min, SIGNAL(clicked()), this, SLOT(slotPlayLevelSetLong30()));
    connect((QObject *)playLevel->long120min, SIGNAL(clicked()), this, SLOT(slotPlayLevelSetLong120()));
}

slibo::~slibo() {
    saveProp(KGlobal::config());
}

//delayed initialization. polish() seems still to early, the child windows don't get properly maximized if they 
void slibo::show() {
    KMainWindow::show();
    if (notInitialized) {
	readProp(KGlobal::config());
	sliboPreferences->m_pageOne->connectWidgets();
	buildEngineMenu();
	notInitialized = false;
    }
}

void slibo::closeEvent(QCloseEvent *e) {
    for (Game *game = games.first(); game; game=games.next())
        for (Engine*  i=game->engines.first(); i; i=game->engines.next())
            i->end();
    QTimer::singleShot(6000, this, SLOT(close()));

    if (secondClose) {
        KMainWindow::closeEvent(e);
        e->accept();
    } else {
        secondClose = true;
        e->ignore();
    }
}

void slibo::setupActions() {
    new KAction(i18n("New game"), "filenew", CTRL+Key_N, this, SLOT( slotGameNew()), actionCollection(), "GameNew");
    new KAction(i18n("New game (swap sides)"), 0, this, SLOT( slotGameNewSwap()), actionCollection(), "GameNewSwap");
//    new KAction(i18n("New game in new window"), 0, this, SLOT( slotGameNewWindow()), actionCollection(), "GameNewWindow");
    new KAction(i18n("Load from DB"), "fileopen", CTRL+Key_O, this, SLOT( slotGameLoadDB()), actionCollection(), "GameLoadDB");    
    new KAction(i18n("Load..."), "fileopen", CTRL+Key_O, this, SLOT( slotGameLoad()), actionCollection(), "GameLoad");
    new KAction(i18n("Save"), "filesave", 0, this, SLOT( slotGameSave()), actionCollection(), "GameSave");
    new KAction(i18n("Save as.."), "filesaveas", 0, this, SLOT( slotGameSaveAs()), actionCollection(), "GameSaveAs");
    new KAction(i18n("Close"), "fileclose", 0, this, SLOT( slotGameQuit()), actionCollection(), "GameQuit");
    new KAction(i18n("Close All"), 0, this, SLOT( slotGameAllQuit()), actionCollection(), "GameAllQuit");
    new KAction(i18n("Quit"), "exit", 0, this, SLOT( slotGamePrgQuit()), actionCollection(), "GamePrgQuit");
    actionView2d = new KToggleAction(i18n("2D Board"), 0, this, SLOT( slotView2d()), actionCollection(), "View2d");
    actionView3d = new KToggleAction(i18n("3D Board"), 0, this, SLOT( slotView3d()), actionCollection(), "View3d");
    actionViewList = new KToggleAction(i18n("Move List"), 0, this, SLOT( slotViewList()), actionCollection(), "ViewList");
    actionViewStat = new KToggleAction(i18n("Statistics"), 0, this, SLOT( slotViewStat()), actionCollection(), "ViewStat");
    new KAction(i18n("Flip Board"), "reload", CTRL+Key_R, this, SLOT( slotViewRotate()), actionCollection(), "ViewRotate");
    new KAction(i18n("Options..."), 0, this, SLOT( slotViewOptions()), actionCollection(), "ViewOptions");
    new KAction(i18n("Fullscreen"), "window_fullscreen", CTRL+Key_F, this, SLOT( slotViewFullScreen()), actionCollection(), "ViewFullScreen");
    new KAction(i18n("Undo move"), "1leftarrow", Key_Left, this, SLOT( slotMoveBack()), actionCollection(), "MoveBack");
    new KAction(i18n("Redo move"), "1rightarrow", Key_Right, this, SLOT( slotMoveForward()), actionCollection(), "MoveForward");
    new KAction(i18n("Move now"), 0, this, SLOT( slotMoveMove()), actionCollection(), "MoveMove");
    new KAction(i18n("Change sides"), 0, this, SLOT( slotMoveChangeSides()), actionCollection(), "MoveChangeSides");
    new KAction(i18n("Time Control..."), 0, this, SLOT( slotSettPlaylevel()), actionCollection(), "SettPlaylevel");
    //    new KAction(i18n("SettEngine"), 0, this, SLOT( slotSettEngine()), actionCollection(), "SettEngine");
    new KAction(i18n("SettConfiguration"), 0, this, SLOT( slotSettConfiguration()), actionCollection(), "SettConfiguration");
    //new KAction(i18n("Load engine..."), 0, this, SLOT( slotEngineLoad()), actionCollection(), "EngineLoad");

    m_toolbarAction = KStdAction::showToolbar(this, SLOT(optionsShowToolbar()), actionCollection());
    m_statusbarAction = KStdAction::showStatusbar(this, SLOT(optionsShowStatusbar()), actionCollection());

    KStdAction::keyBindings(this, SLOT(optionsConfigureKeys()), actionCollection());
    KStdAction::configureToolbars(this, SLOT(optionsConfigureToolbars()), actionCollection());
    KStdAction::preferences(this, SLOT(optionsPreferences()), actionCollection());

    createGUI();
}

void slibo::setupAccel() {}

void slibo::saveProp(KConfig *config) {
    // the 'config' object points to the session managed
    // config file.  anything you write here will be available
    // later when this app is restored
    config->setGroup("");

    config->writeEntry("slibo", geometry());

    config->writeEntry("numGames", games.count());

    int i=0;
    int j=0;
    for (Game *game = games.first(); game; game = games.next(), i++) {
        config->setGroup("Game " + QString::number(i));
        config->writeEntry("numEngines", games.count());
        for (Engine *engine = game->engines.first(); engine; engine = game->engines.next(), j++) {
            config->writeEntry("engine" + QString::number(j), engine->getName());
            config->writeEntry("engine" + QString::number(j) + "color", engine->getColor());
	    if (engine->getStatView()) {
                config->writeEntry("stat" + QString::number(j) , engine->getStatView()->parentWidget()->geometry());
		config->writeEntry("statmax", engine->getStatView()->parentWidget()->isMaximized()); 
		config->writeEntry("statshown", engine->getStatView()->isShown()); 
	    } else
                config->deleteEntry("stat" + QString::number(j) );
        }

        if (game->getGLView()) {
            config->writeEntry("gl", game->getGLView()->parentWidget()->geometry());
	    config->writeEntry("glmax", game->getGLView()->parentWidget()->isMaximized()); 
	    config->writeEntry("glshown", game->getGLView()->isShown()); 
            config->writeEntry("viewangle", game->getGLView()->viewAngle);
            config->writeEntry("viewpos", game->getGLView()->viewPos);
            config->writeEntry("viewdistance", game->getGLView()->viewDistance);
            config->writeEntry("shadows", game->getGLView()->shadows);
            config->writeEntry("reflections", game->getGLView()->reflections);
            config->writeEntry("textures", game->getGLView()->textures);
            config->writeEntry("lightx", game->getGLView()->lightPosition[0]);
            config->writeEntry("lighty", game->getGLView()->lightPosition[1]);
            config->writeEntry("lightz", game->getGLView()->lightPosition[2]);
        } else
            config->deleteEntry("gl");

        if (game->getFlatView()) {
            config->writeEntry("flat", game->getFlatView()->parentWidget()->geometry());
 	    config->writeEntry("flatmax", game->getFlatView()->parentWidget()->isMaximized()); 
 	    config->writeEntry("flatshown", game->getFlatView()->isShown()); 
       } else
            config->deleteEntry("flat");

        if (game->getListView()) {
            config->writeEntry("list", game->getListView()->parentWidget()->geometry());
	    config->writeEntry("listmax", game->getListView()->parentWidget()->isMaximized()); 
	    config->writeEntry("listshown", game->getListView()->isShown()); 
        } else
            config->deleteEntry("list");
    }
}

void slibo::readProp(KConfig *config) {
    // the 'config' object points to the session managed
    // config file.  this function is automatically called whenever
    // the app is being restored.  read in here whatever you wrote
    // in 'saveProperties'
    config->setGroup("");
    QRect rect = config->readRectEntry("slibo", &QRect(0,0,800,700));
    setGeometry(rect);

    //make sure we have at least one game
    int n=config->readNumEntry("numGames", 1);

    for (int i=0; i<n; i++) {
        slotGameNewWindow();
        config->setGroup("Game " + QString::number(i));
        int m=config->readNumEntry("numEngines", 1);
        for (int j=0; j<m; j++) {
            activeGame->engines.append(new Engine(ws, activeGame, config->readEntry("engine" + QString::number(j), "sliboengine")));
            activeGame->engines.current()->setColor(config->readNumEntry("engine" + QString::number(j) + "color", -1));
            rect = config->readRectEntry("stat" + QString::number(j), &QRect(600,0,200,500));
	    if (activeGame->engines.current()->getStatView()) {
		activeGame->engines.current()->getStatView()->parentWidget()->setGeometry(rect);
		if (!config->readBoolEntry("statshown"))
		    activeGame->engines.current()->getStatView()->hide();
		if (config->readBoolEntry("statmax"))
		    activeGame->engines.current()->getStatView()->parentWidget()->showMaximized();
	    }
		
        }
        activeGame->setCaption();

	
        QRect glrect = config->readRectEntry("gl", &QRect(5,5,790,590));
	if (activeGame->getGLView()) {
        activeGame->getGLView()->viewAngle = config->readPointEntry("viewangle", &QPoint(0, 110));
        activeGame->getGLView()->viewPos = config->readPointEntry("viewpos", &QPoint(0, -20));
        activeGame->getGLView()->viewDistance = config->readDoubleNumEntry("viewdistance", 10.75);
        activeGame->getGLView()->shadows = config->readNumEntry("shadows", 1);
        activeGame->getGLView()->reflections = config->readNumEntry("reflections", 1);
        activeGame->getGLView()->textures = config->readNumEntry("textures", 1);
        activeGame->getGLView()->lightPosition[0] = config->readDoubleNumEntry("lightx", 0.0);
        activeGame->getGLView()->lightPosition[1] = config->readDoubleNumEntry("lighty", 15.0);
        activeGame->getGLView()->lightPosition[2] = config->readDoubleNumEntry("lightz", 5.0);
        activeGame->getGLView()->parentWidget()->setGeometry(glrect);
	if (!config->readBoolEntry("glshown", true)) 
	    activeGame->getGLView()->hide();
	else
            actionView3d->activate();

	if (config->readBoolEntry("glmax", true))
	    activeGame->getGLView()->parentWidget()->showMaximized();
    }
	
        rect = config->readRectEntry("flat", &QRect(50,50,300,300));
	if (activeGame->getFlatView()) {
        activeGame->getFlatView()->parentWidget()->setGeometry(rect);
	activeGame->getFlatView()->setMinimumSize(132, 132);
	if (!config->readBoolEntry("flatshown"))
	    activeGame->getFlatView()->hide();
	else
	    actionView2d->activate();

	if (config->readBoolEntry("flatmax"))
	    activeGame->getFlatView()->parentWidget()->showMaximized();
    }
	
        rect = config->readRectEntry("list", &QRect(400,0,200,250));
	if (activeGame->getListView()) {
        activeGame->getListView()->parentWidget()->setGeometry(rect);
	if (!config->readBoolEntry("listshown"))
	    activeGame->getListView()->hide();
	else
	    actionViewList->activate();

	if (config->readBoolEntry("listmax"))
	    activeGame->getListView()->parentWidget()->showMaximized();
    }
	
	if (activeGame->getFlatView()) 
        activeGame->getGLView()->parentWidget()->setGeometry(glrect);

    }
}

void slibo::dragEnterEvent(QDragEnterEvent *event) {
    // accept uri drops only
    event->accept(QUriDrag::canDecode(event));
}

void slibo::dropEvent(QDropEvent *event) {
    // this is a very simplistic implementation of a drop event.  we
    // will only accept a dropped URL.  the Qt dnd code can do *much*
    // much more, so please read the docs there
    QStrList uri;

    // see if we can decode a URI.. if not, just ignore it
    if (QUriDrag::decode(event, uri)) {
        // okay, we have a URI.. process it
        QString url, target;
        url = uri.first();

        // load in the file
        load(KURL(url));
    }
}


void slibo::optionsShowToolbar() {
    // this is all very cut and paste code for showing/hiding the
    // toolbar
    if (m_toolbarAction->isChecked())
        toolBar()->show();
    else
        toolBar()->hide();
}

void slibo::optionsShowStatusbar() {
    // this is all very cut and paste code for showing/hiding the
    // statusbar
    if (m_statusbarAction->isChecked())
        statusBar()->show();
    else
        statusBar()->hide();
}

void slibo::optionsConfigureKeys() {
    KKeyDialog::configureKeys(actionCollection(), "sliboui.rc");
}

void slibo::optionsConfigureToolbars() {
    // use the standard toolbar editor
    KEditToolbar dlg(actionCollection());
    if (dlg.exec()) {
        // recreate our GUI
        createGUI();
    }
}

void slibo::optionsPreferences() {
    sliboPreferences->show();
}

void slibo::slotGameNew() {
    activeGame->newGame();
}

void slibo::slotGameNewSwap() {
    Engine *wengine = NULL;
    Engine *bengine = NULL;
    for (Engine*  i=activeGame->engines.first(); i; i=activeGame->engines.next()) {
	if (i->getColor() == WHITE) 
	    wengine = i;
	if (i->getColor() == BLACK)
	    bengine = i;
	i->setColor(0);
    }
//    activeGame->changeSides();
    activeGame->newGame();
    activeGame->getGLView()->rotate180();
    if (wengine) wengine->setColor(BLACK);
    if (bengine) bengine->setColor(WHITE);
    updateEngineMenu();
}

void slibo::slotGameNewWindow() {
    activeGame = new Game(ws);
    games.append(activeGame);
    connect (activeGame->getGLView(), SIGNAL(enterEvent(QEvent *)), this, SLOT(slotEnterGL()));
    connect (activeGame->getFlatView(), SIGNAL(enterEvent(QEvent *)), this, SLOT(slotEnterFlat()));
    connect (activeGame->getListView(), SIGNAL(enterEvent(QEvent *)), this, SLOT(slotEnterList()));
    connect (activeGame->getGLView(), SIGNAL(leaveEvent(QEvent *)), this, SLOT(slotLeave()));
    connect (activeGame->getFlatView(), SIGNAL(leaveEvent(QEvent *)), this, SLOT(slotLeave()));
    connect (activeGame->getListView(), SIGNAL(leaveEvent(QEvent *)), this, SLOT(slotLeave()));
}

void slibo::load(const KURL &url) {
    activeGame->openGame(url, NULL);
}

void slibo::slotGameLoad() {
//    KURL gameurl = QFileDialog::getOpenFileName ( 0, "*.pgn" );
    KURL gameurl = KFileDialog::getOpenURL(":gamedir", "application/x-chess-pgn application/x-chess-epd application/x-chess-fen", this, "Select Game");
    if (gameurl.isEmpty())
        return;
    activeGame->openGame(gameurl, NULL);
}

void slibo::slotGameLoadDB() {
    KURL url = KFileDialog::getOpenURL(":gamedir", "application/x-chess-pgn", this, "Select PGN Library");
    QString tmpfile;
    KIO::NetAccess::download(url, tmpfile);
    QFile file(tmpfile);
    if (url.filename().right(4).lower() == ".pgn")  {
	if (file.open(IO_ReadOnly)) {
	    while(!file.atEnd()) {
		Pgn *pgn = new Pgn;
		pgn->importPGN(file);
		pgnDB.insert(pgn);
		delete pgn;
	    }
	    file.close();
        }
    }
    KIO::NetAccess::removeTempFile(tmpfile);
}

void slibo::slotGameSave() {
    if (activeGame->hasUrl)
        activeGame->saveGame(activeGame->getURL(), NULL);
    else
        slotGameSaveAs();
};
void slibo::slotGameSaveAs() {
//    KURL gameurl = QFileDialog::getSaveFileName ( 0, "*.pgn" );
    KURL gameurl = KFileDialog::getSaveURL(":gamedir", "*.pgn", this, "Select Game");
    if (gameurl.isEmpty())
        return;
    activeGame->saveGame(gameurl, NULL);
    activeGame->hasUrl = true;
};
void slibo::slotGameQuit() {
    close();
};
void slibo::slotGameAllQuit() {}
;
void slibo::slotGamePrgQuit() {
    slotGameQuit();
}
;
void slibo::slotView2d() {
    if (actionView2d->isChecked()) {
        activeGame->getFlatView()->show();
    } else {
        activeGame->getFlatView()->hide();
    }
};
void slibo::slotView3d() {
    if (actionView3d->isChecked()) {
        activeGame->getGLView()->show();
    } else {
        activeGame->getGLView()->hide();
    }
};
void slibo::slotViewList() {
    if (actionViewList->isChecked()) {
        activeGame->getListView()->show();
    } else {
        activeGame->getListView()->hide();
    }
};

void slibo::slotViewStat() {
    if (actionViewStat->isChecked())
        for (Engine*  i=activeGame->engines.first(); i; i=activeGame->engines.next())
            i->getStatView()->show();
    else
        for (Engine*  i=activeGame->engines.first(); i; i=activeGame->engines.next())
            i->getStatView()->hide();
};

void slibo::slotViewRotate() {
    activeGame->getGLView()->rotate180();
}

void slibo::slotViewOptions() {
    sliboPreferences->show();
}

void slibo::slotViewFullScreen() {
    if (isFullScreen()) {
	toolBar()->show();
	menuBar()->show();
	statusBar()->show();
	ws->activeWindow()->showNormal();
	showNormal();
    } else {
	toolBar()->hide();
	menuBar()->hide();
	statusBar()->hide();
	ws->activeWindow()->showMaximized();
	showFullScreen();
    }	
}

void slibo::slotMoveChangeSides() {
    activeGame->changeSides();
    updateEngineMenu();
}

void slibo::slotMoveForward() {
    activeGame->redoMove();
    updateEngineMenu();
}

void slibo::slotMoveBack() {
    activeGame->undoMove();
    updateEngineMenu();
}

void slibo::slotMoveMove() {}

void slibo::slotSettEngine() {}

void slibo::slotSettConfiguration() {}

void slibo::slotEngineLoad() {
    QString enginepath = KFileDialog::getOpenFileName(":enginedir", "application/x-executable", this, "Select Engine");
    if (!enginepath)
        return;

    QString enginename = enginepath.section("/", -1, -1);
    qDebug("Loading %s", enginepath.latin1());
    qDebug("Engines already loaded: %d", activeGame->engines.count());
    activeGame->engines.append(new Engine(ws, activeGame, enginepath));
    /*    while (games->current()->engines.find(enginename)) {
            qDebug("%s already exists", enginename.latin1());
            enginename = enginename.section("<", 0, 0)
                + "<"
                + QString::number(enginename.section("<", 1, 1).section(">", -2, -2).toInt() + 1)
                + ">";
            qDebug("trying %s", enginename.latin1());
        }
        qDebug("Inserting engine %s", enginename.latin1());
        engines.insert(enginename, new Engine(enginepath, "debug 19"));*/
}

void slibo::slotPlayerAnalyze(int intenginepath) {
    QString enginepath = (char *)intenginepath;
    if (!activeGame)
        return;
    Engine *olde = 0;
    for (Engine *e = activeGame->engines.first(); e; e = activeGame->engines.next()) {
        if (e->getName().section('/', -1, -1) == enginepath.section('/', -1, -1) && e->getColor() == 0) {
            olde = e;
            break;
        }
    }
    if (olde) {
	activeGame->engines.remove(olde);
        olde->end();
	updateEngineMenu();
	return;
    }
    if ("human" != QString(enginepath)) {
	activeGame->engines.append(new Engine(ws, activeGame, enginepath));
	activeGame->engines.current()->setColor( 0 );
    }

    updateEngineMenu();
}

void slibo::slotPlayerWhite(int intenginepath) {
    const char *enginepath = (char *)intenginepath;
    if (!activeGame)
        return;
    Engine *olde = 0;
    for (Engine *e = activeGame->engines.first(); e; e = activeGame->engines.next())
        if (e->getColor() == WHITE) {
            olde = e;
            break;
        }

    if (olde) {
        if (olde->getName() == enginepath)
            return;
        activeGame->engines.remove(olde);
        olde->end();
    }
    if ("human" != QString(enginepath)) {
	activeGame->engines.append(new Engine(ws, activeGame, enginepath));
	activeGame->engines.current()->setColor( WHITE );
    }
    updateEngineMenu();
}

void slibo::slotPlayerBlack(int intenginepath) {
    const char *enginepath = (char *)intenginepath;
    if (!activeGame)
        return;
    Engine *olde = 0;
    for (Engine *e = activeGame->engines.first(); e; e = activeGame->engines.next())
        if (e->getColor() == BLACK) {
            olde = e;
            break;
        }

    if (olde) {
        if (olde->getName() == enginepath)
            return;
        activeGame->engines.remove(olde);
        olde->end();
    }
    if ("human" != QString(enginepath)) {
	activeGame->engines.append(new Engine(ws, activeGame, enginepath));
	activeGame->engines.current()->setColor( BLACK );
    }
    updateEngineMenu();
}

void slibo::scanForEngines() {
    const char *files[] = {"crafty", "sliboengine", "phalanx", "gnuchess", "faile", "KinghtCap", "baby_xboard_engine", "blah", 0
                          };
    const char **f = files;
    FILE *fp;
    char *path;
    size_t len;

    enginenames.append("human");
    while (*f) {
        QString cmd = QString("which ") + *f;
        fp = popen(cmd.latin1(), "r");
        path = NULL;
        len = 0;
        if (getline(&path, &len, fp)>1) {
            path[strlen(path)-1]=0;
            qDebug("found %s at %s", *f, path);
            enginenames.append(path);
        }
        pclose(fp);
        f++;
    }
}

void slibo::buildEngineMenu() {
    emenu = new KPopupMenu( this );
    //    QLabel *wtxt = new QLabel("<font color=""#FFF0E0"">White</font>", menu);
    wmenu = new KPopupMenu( this );
    for (QStringList::iterator i=enginenames.begin(); i!=enginenames.end(); ++i) {
        int id=wmenu->insertItem(*i, this, SLOT(slotPlayerWhite(int)));
        wmenu->setItemParameter(id, (int)(*i).ascii());
    }
    wmenuid = emenu->insertItem("White: Human", wmenu);

    bmenu = new KPopupMenu( this );
    for (QStringList::iterator i=enginenames.begin(); i!=enginenames.end(); ++i) {
        int id=bmenu->insertItem(*i, this, SLOT(slotPlayerBlack(int)));
        bmenu->setItemParameter(id,  (int)(*i).ascii());
    }
    bmenuid = emenu->insertItem("Black: Human", bmenu);

    amenu = new KPopupMenu( this );
    for (QStringList::iterator i=enginenames.begin(); i!=enginenames.end(); ++i) {
        int id=amenu->insertItem(*i, this, SLOT(slotPlayerAnalyze(int)));
        amenu->setItemParameter(id,  (int)(*i).ascii());
    }
    amenuid = emenu->insertItem("Analyzing:", amenu);

    bool haveWhite = false;
    bool haveBlack = false;
    if (activeGame) {
        for (Engine *e = activeGame->engines.first(); e; e = activeGame->engines.next()) {
            if (e->getColor() == WHITE && !haveWhite) {
                emenu->changeItem(wmenuid, "White: "+e->nickName);
                haveWhite = true;
            } else if (e->getColor() == BLACK && !haveBlack) {
                emenu->changeItem(bmenuid, "Black: "+e->nickName);
                haveBlack = true;
            } else {
                e->setColor(0);
                int id = emenu->insertItem("       " + e->nickName);
                emenu->setItemEnabled(id, false);
            }
        }
        activeGame->setCaption();
    }

    emenuid = menuBar()->insertItem("Players", emenu);

}

void slibo::updateEngineMenu() {
    bool haveWhite = false;
    bool haveBlack = false;
    emenu->changeItem(wmenuid, "White: Human");
    emenu->changeItem(bmenuid, "Black: Human");
    int id = -1;
    for (int index = 0; (id = emenu->idAt(index)) != -1; index++) {
	if (!emenu->isItemEnabled(id)) {
	    emenu->removeItem(id);
	    index--;
	}
    }
    
    if (activeGame) {
        for (Engine *e = activeGame->engines.first(); e; e = activeGame->engines.next()) {
            if (e->getColor() == WHITE && !haveWhite) {
                emenu->changeItem(wmenuid, "White: "+e->nickName);
                haveWhite = true;
            } else if (e->getColor() == BLACK && !haveBlack) {
                emenu->changeItem(bmenuid, "Black: "+e->nickName);
                haveBlack = true;
            } else {
                e->setColor(0);
                int id = emenu->insertItem("       " + e->nickName);
                emenu->setItemEnabled(id, false);
            }
        }
        activeGame->setCaption();
    }
}

//
// these slots are activated if something in the options dialog happens
//

void slibo::sliderValueChanged() {
    activeGame->getGLView()->lightPosition[1] = (1500-sb[0]->value()) / 100.0;
    activeGame->getGLView()->lightPosition[0] = sb[1]->value() / 100.0;
    activeGame->getGLView()->lightPosition[2] = sb[2]->value() / 100.0;
    activeGame->getGLView()->repaintNeeded = true;

}

void slibo::checkBoxClicked() {
    activeGame->getGLView()->shadows = cb[0]->isChecked();
    activeGame->getGLView()->reflections = cb[1]->isChecked();
    activeGame->getGLView()->textures = cb[2]->isChecked();
    activeGame->getGLView()->repaintNeeded = true;
}

//
// these slots are activated if something in the playlevel widget happens
//

void slibo::slotPlayLevelAccept() {
    Engine *selected = activeGame->engines.at(playLevel->engineListBox->index(playLevel->engineListBox->selectedItem()));
    selected->write("level "+QString::number(playLevel->t1moves->value())+" "+QString::number(playLevel->t1time->value())+" "+QString::number(playLevel->t1inc->value()));

}

void slibo::playLevelSet(int moves, int time, int inc) {
    Engine *selected = activeGame->engines.at(playLevel->engineListBox->index(playLevel->engineListBox->selectedItem()));
    selected->t1time = time;
    selected->t1moves = moves;
    selected->t1inc = inc;
    slotPlayLevelSelectionChanged();
}

void slibo::slotPlayLevelSet4_2() {
    playLevelSet(0, 4*60,2);
}
void slibo::slotPlayLevelSet5() {
    playLevelSet(0, 5*60,0);
}
void slibo::slotPlayLevelSet30_15() {
    playLevelSet(0, 30*60,15);
}
void slibo::slotPlayLevelSet30() {
    playLevelSet(0, 30*60,0);
}
void slibo::slotPlayLevelSetLong5() {
    playLevelSet(40, 5*60,0);
}
void slibo::slotPlayLevelSetLong30() {
    playLevelSet(40, 30*60,0);
}
void slibo::slotPlayLevelSetLong120() {
    playLevelSet(40, 120*60,0);
}

void slibo::slotSettPlaylevel() {
    if (activeGame->engines.count() == 0) {
        return;
    }
    playLevel->engineListBox->clear();
    for (Engine *engine = activeGame->engines.first(); engine; engine = activeGame->engines.next()) {
        playLevel->engineListBox->insertItem(engine->getName());
    }
    playLevel->engineListBox->setSelected(0, true);
    playLevel->show();
}

void slibo::slotPlayLevelSelectionChanged() {
    Engine *selected = activeGame->engines.at(playLevel->engineListBox->index(playLevel->engineListBox->selectedItem()));

    playLevel->blitz4min2->setChecked(selected->t1moves == 0 && selected->t1time == 4*60 && selected->t1inc == 2);
    playLevel->blitz5min->setChecked(selected->t1moves == 0 && selected->t1time == 5*60 && selected->t1inc == 0);
    playLevel->blitz30min15->setChecked(selected->t1moves == 0 && selected->t1time == 30*60 && selected->t1inc == 15);
    playLevel->blitz30min->setChecked(selected->t1moves == 0 && selected->t1time == 30*60 && selected->t1inc == 0);
    playLevel->long5min->setChecked(selected->t1moves == 40 && selected->t1time == 5*60 && selected->t1inc == 0);
    playLevel->long30min->setChecked(selected->t1moves == 40 && selected->t1time == 30*60 && selected->t1inc == 0);
    playLevel->long120min->setChecked(selected->t1moves == 40 && selected->t1time == 120*60 && selected->t1inc == 0);

    playLevel->t1time->setValue(selected->t1time/60);
    playLevel->t2time->setValue(selected->t2time/60);
    playLevel->t3time->setValue(selected->t3time/60);
    playLevel->t1total->setValue(selected->t1time/60);
    playLevel->t1moves->setValue(selected->t1moves);
    playLevel->t2moves->setValue(selected->t2moves);
    playLevel->t3moves->setValue(selected->t3moves);
    playLevel->t1inc->setValue(selected->t1inc);

    if (selected->t1moves == 0)
        playLevel->tabWidget->setCurrentPage(0);
    else
        playLevel->tabWidget->setCurrentPage(1);
}

void slibo::slotEnterGL() {
    statusBar()->changeItem("click right mouse button to rotate, middle to move, use wheel to zoom", STATUSBAR_RIGHT);
}

void slibo::slotEnterFlat() {
    statusBar()->changeItem("pick piece with left mouse button", STATUSBAR_RIGHT);
}

void slibo::slotEnterList() {
    statusBar()->changeItem("leftclick on a move to go to this position", STATUSBAR_RIGHT);
}

void slibo::slotLeave() {
    statusBar()->changeItem("", STATUSBAR_RIGHT);
}
#include "slibo.moc"
