/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#ifndef _SLIBO_H_
#define _SLIBO_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif 

#include "pgndb.h"

#include <kapp.h>
#include <kmainwindow.h>

class QPrinter;
class KToggleAction;
class KURL;
class Game;
class Stats;
class Info;
class Engine;
class QWorkspace;
class KPopupMenu;
class SliboPreferences;
class QCheckBox;
class QSlider;
class PlayLevel;

/**
 * This class serves as the main window for slibo.  It handles the
 * menus, toolbars, and status bars.
 *
 * @short Main window class
 * @author $AUTHOR <$EMAIL>
 * @version $APP_VERSION
 */
class slibo : public KMainWindow
{
    Q_OBJECT
public:
    /**
     * Default Constructor
     */
    slibo();

    /**
     * Default Destructor
     */
    virtual ~slibo();

    /**
     * Overridden virtuals for Qt drag 'n drop (XDND)
     */
    virtual void dragEnterEvent(QDragEnterEvent *event);
    virtual void dropEvent(QDropEvent *event);

    /**
     * Use this method to load whatever file/URL you have
     */
    void load(const KURL& url);
    void show();

    KToggleAction *actionView2d;
    KToggleAction *actionView3d;
    KToggleAction *actionViewList;
    KToggleAction *actionViewStat;
    QCheckBox *cb[3];
    QSlider *sb[3];
    Game *activeGame;
    PgnDB pgnDB;

public slots:
    void sliderValueChanged();
    void checkBoxClicked();

private:
    /**
     * This function is called when it is time for the app to save its
     * properties for session management purposes.
     */
    void saveProp(KConfig *);

    /**
     * This function is called when this app is restored.  The KConfig
     * object points to the session management config file that was saved
     * with @ref saveProperties
     */
    void readProp(KConfig *);
    void buildEngineMenu();
    void updateEngineMenu();
    void scanForEngines();
    void playLevelSet(int time, int moves, int inc);

    QPtrList<Game> games;
    QStringList enginenames;
    SliboPreferences *sliboPreferences;
    PlayLevel *playLevel;
    QPrinter   *m_printer;
    QWorkspace *ws;
    KToggleAction *m_toolbarAction;
    KToggleAction *m_statusbarAction;
    KToggleAction *viewInfoAction, *viewMoveListAction, *view2DViewAction, *view3DviewAction, *viewStatsAction;
    KPopupMenu *amenu, *wmenu, *bmenu, *emenu;
    int amenuid, wmenuid, bmenuid, emenuid;
    bool secondClose;
    bool notInitialized;
    
private slots:
    void slotGameNew();
    void slotGameNewSwap();
    void slotGameNewWindow();
    void slotGameLoad();
    void slotGameLoadDB();
    void slotGameSave();
    void slotGameSaveAs();
    void slotGameQuit();
    void slotGameAllQuit();
    void slotGamePrgQuit();
    void slotView2d();
    void slotView3d();
    void slotViewList();
    void slotViewStat();
    void slotViewRotate();
    void slotViewOptions();
    void slotViewFullScreen();
    void slotMoveForward();
    void slotMoveBack();
    void slotMoveMove();
    void slotMoveChangeSides();
    void slotSettPlaylevel();
    void slotSettEngine();
    void slotSettConfiguration();
    void optionsShowToolbar();
    void optionsShowStatusbar();
    void optionsConfigureKeys();
    void optionsConfigureToolbars();
    void optionsPreferences();
	void setupAccel();
	void setupActions();
    void slotEngineLoad();
    void slotPlayerWhite(int);
    void slotPlayerBlack(int);
    void slotPlayerAnalyze(int);
    void closeEvent(QCloseEvent *);
    
    //slots for various actions in playlevel widget
    void slotPlayLevelAccept();
    void slotPlayLevelSelectionChanged();
    void slotPlayLevelSet4_2();
    void slotPlayLevelSet5();
    void slotPlayLevelSet30_15();
    void slotPlayLevelSet30();
    void slotPlayLevelSetLong5();
    void slotPlayLevelSetLong30();
    void slotPlayLevelSetLong120();
    
    void slotEnterGL();
    void slotEnterFlat();
    void slotEnterList();
    void slotLeave();
};

#endif // _SLIBO_H_
