/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "board.h"
#include "hash.h"
#include "eval.h"
#include "search.h"
#include "main.h"
#include "util.h"
#include "opening.h"
#include "debug.h"

#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <stdio.h>
#include <ncurses.h>
#include <math.h>

Board board __attribute__ ((aligned(64)));

#ifdef MYDEBUG
Board debugboard[1000];
#endif

void convert2EPD(const Board * b, char *r)
{
    int x, type;
    int i, j;

    for (i = 0; i < 8; i++) {
        x = 0;
        for (j = 0; j < 8; j++) {
            type = TYPE(7 - i, j);
            if (!type)
                x++;
            else {
                if (x) {
                    *r++ = (char) ('0' + x);
                    x = 0;
                }
                *r++ = "kpnqbr RBQNPK"[type + 6];
            }
        }
        if (x)
            *r++ = (char) ('0' + x);
        if (i < 7)
            *r++ = '/';
    }

    *r++ = ' ';
    *r++ = b->color > 0 ? 'w' : 'b';

    *r++ = ' ';
    if (b->o_o) {
        if (b->o_o & CAST_WK)
            *r++ = 'K';
        if (b->o_o & CAST_WQ)
            *r++ = 'Q';
        if (b->o_o & CAST_BK)
            *r++ = 'k';
        if (b->o_o & CAST_BQ)
            *r++ = 'q';
    } else
        *r++ = '-';

    *r++ = ' ';
    if (b->epstatus) {
        *r++ = (char) (X(b->epstatus) + 'a');
        *r++ = (char) (Y(b->epstatus) + '1');
    } else
        *r++ = '-';

    *r++ = 0;
}

void clearBoard(Board * b)
{
    int i, j;

    for (i = 0; i < 12 * 16; i++) {     //set border
        b->board[0][i].w = 1;
        b->board[0][i].b = -1;
    }

#ifndef SIMPLEBOARD
    memset(b->attacklen, 0, sizeof(b->attacklen));
    memset(b->attack, 0, sizeof(b->attack));
    memset(b->attackvec, 0, sizeof(b->attackvec));
#endif
    memset(b->repTable, 0, sizeof(b->repTable));

    for (i = 2; i < 10; i++)
        for (j = 2; j < 10; j++)
            b->board[i][j].all = 0;

    for (i = 0; i < 13; i++)
        b->pieces[i].cnt = 0;

    b->epstatus = 0;
    b->o_o = 0;
    b->color = 1;
    b->startMove = b->currMove = b->moveList;

    b->key = (HashKey) { .main.key = 0, .pawn.key = 0, .white.key = 0, .black.key = 0 };
    b->material = 0;

#ifndef SIMPLEBOARD
    for (int dir = 0; dir < 8; dir++)
        for (int x = 0; x < 8; x++)
            for (int y = 0; y < 8; y++) {
                for (j = 1; !(((x + y * 16) * sizeof(int) + dirtab[dir] * j) & (0x88 * sizeof(int))); j++) ;
                b->attacklen[dir][y + 2][x + 2] = j - 1;
            }
#endif
}

void initBoard()
{
    int x, y, p;

    static Board *const b = &board;

    importEPD("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq -");
    initBook();

    eval(b);

    for (p = 1; p < 7; p++)
        for (x = 0; x < 8; x++)
            for (y = 0; y < 8; y++)
                if (VAL(p, y, x) != -VAL(-p, 7 - y, x))
                    printf("%c%c %f!=%f\n", 'a' + x, '1' + y, VAL(p, y, x), VAL(-p, y, x));
}

Move *maxr;

void sortw(Move * l, Move * r)
{
    Move *i = l - 1;
    Move *j = r;
    float v = r->val;
    if (l->val < v) {
        Move m = *l;
        *l = *r;
        *r = m;
        v = r->val;
        i = l;
    }
    if (r-l <= 1) return;
    for (;;) {
        while ((++i)->val > v) ;
        while ((--j)->val < v) ;
        if (j <= i)
            break;
        Move m = *i;
        *i = *j;
        *j = m;
    }
    {
        Move m = *i;
        *i = *r;
        *r = m;
    }
    if (l > maxr) return;
    if ((char *) i - (char *) l > (signed) sizeof(Move))
        sortw(l, i - 1);
    if (i+1 > maxr) return;    
    if ((char *) r - (char *) i > (signed) sizeof(Move))
        sortw(i + 1, r);
}

void sortb(Move * l, Move * r)
{
    Move *i = l - 1;
    Move *j = r;
    float v = r->val;
    if (l->val > v) {
        Move m = *l;
        *l = *r;
        *r = m;
        v = r->val;
        i = l;
    }
    if (r-l <= 1) return;
    for (;;) {
        while ((++i)->val < v) ;
        while ((--j)->val > v) ;
        if (j <= i)
            break;
        Move m = *i;
        *i = *j;
        *j = m;
    }
    {
        Move m = *i;
        *i = *r;
        *r = m;
    }
    if (l > maxr) return;
    if ((char *) i - (char *) l > (signed) sizeof(Move))
        sortb(l, i - 1);
    if (i+1 > maxr) return;    
    if ((char *) r - (char *) i > (signed) sizeof(Move))
        sortb(i + 1, r);
}

void sort(Move * l, Move * r, int color)
{
    maxr = l+4;
    if (color > 0)
        sortw(l, r);
    else
        sortb(l, r);
}

int insertRep(Board * b)
{
    int repIdx;

    repIdx = REPIDX;
    while (b->repTable[repIdx] && b->repTable[repIdx] != b->key.main.key)
        repIdx = (repIdx + 1) & (REPTABLESIZE - 1);
    if (b->repTable[repIdx] == b->key.main.key)
        return 1;
    b->repTable[repIdx] = b->key.main.key;
    return 0;
}

void deleteRep(Board * b)
{
    int repIdx;

    repIdx = REPIDX;
    while (b->repTable[repIdx] && b->repTable[repIdx] != b->key.main.key)
        repIdx = (repIdx + 1) & (REPTABLESIZE - 1);
    b->repTable[repIdx] = 0;
}

extern unsigned long long currWork;
void doMove(Board * b, Move * m)
{
#ifdef MYDEBUG
    if (flags.debug & DEBUG_MOVES)
        debug("doMove %2x-%2x %d\n", m->src, m->dst, m->special);
    if (flags.debug & DEBUG_INTEGRITY)
        debugboard[b->currMove-b->moveList] = *b;
#endif
    BEGIN_PROFILE(TIME_DOMOVE);
    b->currMove++;
    b->currMove->move = *m;
    b->currMove->epstatus = b->epstatus;
    b->currMove->o_o = b->o_o;
    b->currMove->fiftyMoves = b->fiftyMoves;
    b->epstatus = 0;

    switch (m->special) {
    case SPEC_PROMOQ:
        clrPiece(b, m->src);
        b->currMove->captured = putPiece(b, m->dst, QUEEN * b->color);
        break;
    case SPEC_PROMON:
        clrPiece(b, m->src);
        b->currMove->captured = putPiece(b, m->dst, KNIGHT * b->color);
        break;
    case SPEC_PROMOB:
        clrPiece(b, m->src);
        b->currMove->captured = putPiece(b, m->dst, BISHOP * b->color);
        break;
    case SPEC_PROMOR:
        clrPiece(b, m->src);
        b->currMove->captured = putPiece(b, m->dst, ROOK * b->color);
        break;
    case SPEC_PAWNDBL:
        b->currMove->captured = movePiece(b, m->src, m->dst);
        b->epstatus = m->dst;
        break;
    case SPEC_PAWNEP:
        b->currMove->captured = movePiece(b, m->src, m->dst);
        clrPiece(b, m->dst + ADR(-b->color, 0));
        break;
    case SPEC_CASTLEQ:
        movePiece(b, m->src, m->dst);
        movePiece(b, m->src + ADR(0, -4), m->src + ADR(0, -1));
        if (b->color == 1)
            b->o_o &= ~CAST_WQ & ~CAST_WK;
        else
            b->o_o &= ~CAST_BQ & ~CAST_BK;
        break;
    case SPEC_CASTLEK:
        movePiece(b, m->src + ADR(0, 3), m->src + ADR(0, 1));
    case SPEC_KING:
        b->currMove->captured = movePiece(b, m->src, m->dst);
        if (b->color == 1)
            b->o_o &= ~CAST_WQ & ~CAST_WK;
        else
            b->o_o &= ~CAST_BQ & ~CAST_BK;
        break;
    case SPEC_ROOK:
        b->currMove->captured = movePiece(b, m->src, m->dst);
        if (b->color == 1) {
            if (m->src == ADR(0, 0))
                b->o_o &= ~CAST_WQ;
            else if (m->src == ADR(0, 7))
                b->o_o &= ~CAST_WK;
        } else {
            if (m->src == ADR(7, 0))
                b->o_o &= ~CAST_BQ;
            else if (m->src == ADR(7, 7))
                b->o_o &= ~CAST_BK;
        }
        break;
    default:
        b->currMove->captured = movePiece(b, m->src, m->dst);
    }

    prefetch(HASH_PTR(b->key.main, b->color));
    prefetch(HASH_PTR(b->key.main, -b->color));
    prefetch( (PawnEntry *) pawnHash.table + (b->key.pawn.key0 & pawnHash.mask));
    prefetch( (HashEntry *) whiteHash.table + (b->key.white.key0 & whiteHash.mask));
    prefetch( (HashEntry *) blackHash.table + (b->key.black.key0 & blackHash.mask));
    if (b->currMove->captured || m->piece == PAWN || m->piece == -PAWN) {
        b->fiftyMoves = 0;
    } else
        b->fiftyMoves++;
    END_PROFILE();
#ifdef MYDEBUG
    if (flags.debug & DEBUG_INTEGRITY) {
//        float score = slowEval(b);
//        if (score != b->currMove->move.val) {
//            debug("score mismatch %f should be %f\n", b->currMove->move.val, score);
//            asm("int3");
//        }
        unsigned long long hash = slowHash(b);
        if (hash != b->key.main.key) {
            debug("hash mismatch %llx should be %llx\n", b->key.main.key, hash);
            asm("int3");
        }
    }
#endif
}

void undoMove(Board * b)
{
#ifdef MYDEBUG
    if (flags.debug & DEBUG_MOVES)
        debug("undoMove %2x-%2x %d\n", b->currMove->move.src, b->currMove[-1].move.dst, b->currMove[-1].move.special);
#endif
    BEGIN_PROFILE(TIME_UNDOMOVE);
    switch (b->currMove->move.special) {
    case SPEC_PROMOQ:
    case SPEC_PROMON:
    case SPEC_PROMOB:
    case SPEC_PROMOR:
        if (b->currMove->captured)
            putPiece(b, b->currMove->move.dst, b->currMove->captured);
        else
            clrPiece(b, b->currMove->move.dst);
        putPiece(b, b->currMove->move.src, PAWN * b->color);
        break;
    case SPEC_PAWNEP:
        putPiece(b, b->currMove->move.dst + ADR(-b->color, 0), -PAWN * b->color);
        movePiece(b, b->currMove->move.dst, b->currMove->move.src);
        break;
    case SPEC_CASTLEQ:
        movePiece(b, b->currMove->move.src + ADR(0, -1), b->currMove->move.src + ADR(0, -4));
        movePiece(b, b->currMove->move.dst, b->currMove->move.src);
        break;
    case SPEC_CASTLEK:
        movePiece(b, b->currMove->move.src + ADR(0, 1), b->currMove->move.src + ADR(0, 3));
        movePiece(b, b->currMove->move.dst, b->currMove->move.src);
        break;
    default:
        unmovePiece(b, b->currMove->move.dst, b->currMove->move.src, b->currMove->captured);
    }

    prefetchw(HASH_PTR(b->key.main, -b->color));
    b->epstatus = b->currMove->epstatus;
    b->o_o = b->currMove->o_o;
    b->fiftyMoves = b->currMove->fiftyMoves;
    b->currMove--;
    END_PROFILE();

#ifdef MYDEBUG
    if (flags.debug & DEBUG_INTEGRITY) {
        Board *b1 = &debugboard[b->currMove-b->moveList];
        for (int y = 0; y < 8; y++)
            for (int x = 0; x < 8; x++) {
                if (b1->board[y + 2][x + 2].all != b->board[y + 2][x + 2].all)
                    debug("board error\n");
            }
        for (int p = 0; p < 13; p++)
            if (p != 6) {
                if (b1->pieces[p].cnt != b->pieces[p].cnt)
                    debug("piece count error\n");
                for (unsigned i = 0; i < b->pieces[p].cnt; i++) {
                    bool found = false;
                    for (unsigned j = 0; j < b->pieces[p].cnt; j++)
                        if (b1->pieces[p].pos[i] == b->pieces[p].pos[j]) {
                            found = true;
                            break;
                        }
                    if (!found) {
                        debug("piece position error: piece %d is @%x, should be @%x\n", p - 6, b->pieces[p].pos[i], b1->pieces[p].pos[i]);
                        exit(1);
                    }
                }
            }
    }
#endif
}

unsigned long long searchTime()
{
    struct timeval m;
    gettimeofday(&m, 0);
    return (long long) m.tv_sec * 1000000 + m.tv_usec;
}

/** No descriptions */
void setTimeout(double t_0, double t_1)
{
    t0 = t_0;
    t1 = t_1;
}

/** set breakMove */
void setBreakMove(const Move * m)
{
    breakMove = *m;
}
