/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
 
#include "eval.h"
#include "main.h"
#include "debug.h"
#include "squares.h"
#include "movegen_c.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define PAWNBIT (1<<PAWN)
#define LIGHTBIT ((1<<BISHOP)|(1<<KNIGHT))
#define ROOKBIT (1<<ROOK)
#define QUEENBIT (1<<QUEEN)

#define ENDGAME_MAT 30

//values in quants
#define EVAL_ISOLANI            -8
#define EVAL_ISOLANI_OPEN       -8
#define EVAL_CLOSED_POSITION    8
#define EVAL_ROOK_BLOCKED       -32
#define EVAL_KING_SAFETY        20
#define EVAL_KING_ATTACK        96
#define EVAL_BISHOP_PAIR        16
#define EVAL_ROOK_CONNECTED     16
#define EVAL_ROOK_QUEEN         15
#define EVAL_CLOSED_POS         8

static const int doublepawn[] = {
                                    0, 0, -12, -48, -96, -160, -224, -288
                                };
static const int passedPawn[8] = {
                                     0, 92, 72, 54, 38, 24, 24, 0
                                 };
static const int connectedPPawn[8] = {
                                         0, 64, 32, 16, 0, 0, 0, 0
                                     };
static const int supportedPPawn[8] = {
                                         0, 16, 16, 16, 8, 0, 0, 0
                                     };
static const int possiblePPawn[8] = {
                                        0, 0, 36, 27, 18, 0, 0, 0
                                    };
//static const int safetyTab[20] = { 0, 1, 3, 6, 10, 13, 15, 16,

static float mwght[256];   //middlegame wheight
static float ewght[256];   //endgame wheight

//attack strenght table.
//index is roughly four times the number of attacking pieces.
static int attstr[64];

//counts number ot pieces attacking square src.
//static int attackcntw(const Board *b, int src);
//static int attackcntb(const Board *b, int src);

#define ATTACKTABLE(piece, ptr) ( *(int *)((char *) &attackTable[piece][8][8] + (ptr)) )
static int attackTable[7][16][16];

float drawValue = 0.0;

//static inline void dynEvalPawn(const Board *b, PawnEntry *pawn) __attribute__ ((always_inline));
static inline void dynEvalPawn(const Board *b, PawnEntry *pawn) {
    int wmin[10], bmin[10];
    int wcnt[10], bcnt[10];
    int wmax[10], bmax[10];

    int struc = 0;  //in quants
    //count the number of pawns per line and find the position of the most backward
    for (int i = 0; i < 10; i++) {
        wmin[i] = 7;
        wmax[i] = 0;
        bmin[i] = 0;
        bmax[i] = 7;
        wcnt[i] = bcnt[i] = 0;
    }
    for (unsigned i = 0; i < CNT(PAWN); i++) {
        int x = X(POS(PAWN, i)) + 1;
        int y = Y(POS(PAWN, i));
        if (wmin[x] > y)
            wmin[x] = y;
        if (wmax[x] < y)
            wmax[x] = y;
        wcnt[x]++;
    }

    for (unsigned i = 0; i < CNT(-PAWN); i++) {
        int x = X(POS(-PAWN, i)) + 1;
        int y = Y(POS(-PAWN, i));
        if (bmin[x] < y)
            bmin[x] = y;
        if (bmax[x] > y)
            bmax[x] = y;
        bcnt[x]++;
    }

    int wlastpp = 99;
    int blastpp = 99;

    for (int i = 1; i < 9; i++) {

        if (wcnt[i]) {
            struc += doublepawn[wcnt[i]];
            //                printf("w double %d\n", doublepawn[wcnt[i]]);
            if (wmin[i - 1] > wmin[i] && wmin[i + 1] > wmin[i]) {
                struc += EVAL_ISOLANI;
                //                printf("w isolani %d\n", EVAL_ISOLANI);
                if (!bcnt[i]) {
                    struc += EVAL_ISOLANI_OPEN;
                    //                    printf("w isolani open %d\n", EVAL_ISOLANI_OPEN);
                }
            }

            if (bmin[i - 1] <= wmax[i] && bmin[i] < wmax[i] && bmin[i + 1] <= wmax[i]) {
                int wdist;
                for (wdist = 1; TYPE(7 - wdist, i - 1) != PAWN; wdist++)
                    ;
                struc += passedPawn[wdist];
                //                printf("w passed pawn %d\n", passedPawn[wdist]);
                if (wlastpp == i - 1) {
                    struc += connectedPPawn[wdist];
                    //                    printf("w connected passed pawn %d\n", connectedPPawn[wdist]);
                }
                if (abs(X(POS(KING,0)) - i+1) <= 1 && Y(POS(KING, 0)) >= 7 - wdist) {
                    struc += supportedPPawn[wdist];
                    //                    printf("w supported passed pawn %d\n", supportedPPawn[wdist]);
                }
                wlastpp = i;
            } else if (bmin[i] < wmax[i])
                if (((bmin[i-1] > wmax[i]) + (bmin[i+1] > wmax[i])) <=
                        ((wmin[i-1] < bmin[i-1]-1) + (wmin[i+1] < bmin[i+1]-1))) {
                    struc += possiblePPawn[wmax[i]];
                    //                    printf("w possible passed pawn %d\n", possiblePPawn[7-wmax[i]]);
                }
        }

        if (bcnt[i]) {
            struc -= doublepawn[bcnt[i]];

            if (bmin[i - 1] < bmin[i] && bmin[i + 1] < bmin[i]) {
                struc -= EVAL_ISOLANI;
                if (!wcnt[i])
                    struc -= EVAL_ISOLANI_OPEN;
            }

            if (wmin[i - 1] >= bmax[i] && wmin[i] > bmax[i] && wmin[i + 1] >= bmax[i]) {
                int bdist;
                for (bdist = 1; TYPE(bdist, i - 1) != -PAWN; bdist++)
                    ;
                struc -= passedPawn[bdist];
                if (blastpp == i - 1)
                    struc -= connectedPPawn[bdist];
                if (abs(X(POS(-KING,0)) - i+1) <=1 && Y(POS(-KING, 0)) <= bdist)
                    struc -= supportedPPawn[bdist];
                blastpp = i;
            } else if (wmin[i] > bmax[i])
                if (((wmin[i-1] < bmax[i]) + (wmin[i+1] < bmax[i])) <=
                        ((bmin[i-1] > wmin[i-1]-1) + (bmin[i+1] > wmin[i+1]-1)))
                    struc -= possiblePPawn[bmax[i]];
        }
    }

    if (CNT(PAWN) == 8 || CNT(-PAWN) == 8)
        struc += EVAL_CLOSED_POS;
    //king safety
    //if castling still possible, evaluate value of pawn shield of the king/queenside,
    //which ever is better, maximum score is 0.
    //if castling isn't possible, score is -3 if king is still on d/e file
    //else score is in a range -3..3, depending on the pawn shield. blocked rooks
    //have a malus, to prevent moving the king towards the corner without castling
    int wsafety = 0;
    if (b->o_o & (CAST_WQ | CAST_WK)) {
        int lshield = 0;
        lshield += BOARD22(A2)->w == (PAWN << BOARD_SHIFT);
        lshield += BOARD22(B2)->w == (PAWN << BOARD_SHIFT);
        lshield += BOARD22(C2)->w == (PAWN << BOARD_SHIFT);
        lshield *=2;
        lshield += BOARD22(A3)->w == (PAWN << BOARD_SHIFT);
        lshield += BOARD22(B3)->w == (PAWN << BOARD_SHIFT);
        lshield += BOARD22(C3)->w == (PAWN << BOARD_SHIFT);

        int rshield = 0;
        rshield += BOARD22(F2)->w == (PAWN << BOARD_SHIFT);
        rshield += BOARD22(G2)->w == (PAWN << BOARD_SHIFT);
        rshield += BOARD22(H2)->w == (PAWN << BOARD_SHIFT);
        rshield *=2;
        rshield += BOARD22(F3)->w == (PAWN << BOARD_SHIFT);
        rshield += BOARD22(G3)->w == (PAWN << BOARD_SHIFT);
        rshield += BOARD22(H3)->w == (PAWN << BOARD_SHIFT);

        //6 is the maximum value for l/rshield
        int shield = -6;
        if (rshield >= lshield && b->o_o & CAST_WK)
            shield += rshield;
        else
            shield += lshield;

        //so in the case of a perfect pawn shelter this results in no safety change.
        wsafety += shield*(EVAL_KING_SAFETY/2);

    } else {
        //no castling possible
        if (POS(KING, 0) == G1) {
            if (BOARD22(H1)->w == ROOK << BOARD_SHIFT && wcnt[8])
                wsafety += EVAL_ROOK_BLOCKED;
        } else if (POS(KING, 0) == F1) {
            if ((BOARD22(H1)->w == ROOK << BOARD_SHIFT && wcnt[8]) || (BOARD22(G1)->w == ROOK << BOARD_SHIFT && wcnt[7]))
                wsafety += EVAL_ROOK_BLOCKED;
        } else if (POS(KING, 0) == B1) {
            if (BOARD22(A1)->w == ROOK << BOARD_SHIFT && wcnt[1])
                wsafety += EVAL_ROOK_BLOCKED;
        } else if (POS(KING, 0) == C1) {
            if ((BOARD22(A1)->w == ROOK << BOARD_SHIFT && wcnt[1]) || (BOARD22(B1)->w == ROOK << BOARD_SHIFT && wcnt[7]))
                wsafety += EVAL_ROOK_BLOCKED;
        }
        //safety without pawn shelter
        wsafety -= 3*EVAL_KING_SAFETY;
        int x = X(POS(KING, 0));
        //look for open files
        if (wcnt[x+1] == 0)
            wsafety -= 2*EVAL_KING_SAFETY;
        if (bcnt[x+1] == 0)
            wsafety -= EVAL_KING_SAFETY;

        if (x < 3) {
            wsafety += EVAL_KING_SAFETY * ( BOARD22(A2)->w == (PAWN << BOARD_SHIFT) ) * 2;
            wsafety += EVAL_KING_SAFETY * ( BOARD22(B2)->w == (PAWN << BOARD_SHIFT) ) * 2;
            wsafety += EVAL_KING_SAFETY * ( BOARD22(C2)->w == (PAWN << BOARD_SHIFT) ) * 2;
            wsafety += EVAL_KING_SAFETY * ( BOARD22(A3)->w == (PAWN << BOARD_SHIFT) );
            wsafety += EVAL_KING_SAFETY * ( BOARD22(B3)->w == (PAWN << BOARD_SHIFT) );
            wsafety += EVAL_KING_SAFETY * ( BOARD22(C3)->w == (PAWN << BOARD_SHIFT) );
        } else if (x > 4) {
            wsafety += EVAL_KING_SAFETY * ( BOARD22(F2)->w == (PAWN << BOARD_SHIFT) ) * 2;
            wsafety += EVAL_KING_SAFETY * ( BOARD22(G2)->w == (PAWN << BOARD_SHIFT) ) * 2;
            wsafety += EVAL_KING_SAFETY * ( BOARD22(H2)->w == (PAWN << BOARD_SHIFT) ) * 2;
            wsafety += EVAL_KING_SAFETY * ( BOARD22(F3)->w == (PAWN << BOARD_SHIFT) );
            wsafety += EVAL_KING_SAFETY * ( BOARD22(G3)->w == (PAWN << BOARD_SHIFT) );
            wsafety += EVAL_KING_SAFETY * ( BOARD22(H3)->w == (PAWN << BOARD_SHIFT) );
        }
    }

    int bsafety = 0;
    if (b->o_o & (CAST_BQ | CAST_BK)) {
        //if castling still possible, evaluate value of a possible castling
        int lshield = 0;
        lshield += BOARD22(A7)->w == -(PAWN << BOARD_SHIFT);
        lshield += BOARD22(B7)->w == -(PAWN << BOARD_SHIFT);
        lshield += BOARD22(C7)->w == -(PAWN << BOARD_SHIFT);
        lshield *=2;
        lshield += BOARD22(A6)->w == -(PAWN << BOARD_SHIFT);
        lshield += BOARD22(B6)->w == -(PAWN << BOARD_SHIFT);
        lshield += BOARD22(C6)->w == -(PAWN << BOARD_SHIFT);

        int rshield = 0;
        rshield += BOARD22(F7)->w == -(PAWN << BOARD_SHIFT);
        rshield += BOARD22(G7)->w == -(PAWN << BOARD_SHIFT);
        rshield += BOARD22(H7)->w == -(PAWN << BOARD_SHIFT);
        rshield *=2;
        rshield += BOARD22(F6)->w == -(PAWN << BOARD_SHIFT);
        rshield += BOARD22(G6)->w == -(PAWN << BOARD_SHIFT);
        rshield += BOARD22(H6)->w == -(PAWN << BOARD_SHIFT);

        //6 is the maximum value for l/rshield
        int shield = -6;
        if (rshield >= lshield && b->o_o & CAST_BK)
            shield += rshield;
        else
            shield += lshield;

        //so in the case of a perfect -(PAWN << BOARD_SHIFT) shelter this results in no safety change.
        bsafety -= shield*(EVAL_KING_SAFETY/2);

    } else {
        //no castling possible
        if (POS(-KING, 0) == G8) {
            if (BOARD22(H8)->w == -ROOK << BOARD_SHIFT && bcnt[8])
                bsafety -= EVAL_ROOK_BLOCKED;
        } else if (POS(-KING, 0) == F8) {
            if ((BOARD22(H8)->w == -ROOK << BOARD_SHIFT && bcnt[8]) || (BOARD22(G8)->w == -ROOK << BOARD_SHIFT && bcnt[7]))
                bsafety -= EVAL_ROOK_BLOCKED;
        } else if (POS(-KING, 0) == B8) {
            if (BOARD22(A8)->w == -ROOK << BOARD_SHIFT && bcnt[1])
                bsafety -= EVAL_ROOK_BLOCKED;
        } else if (POS(-KING, 0) == C8) {
            if ((BOARD22(A8)->w == -ROOK << BOARD_SHIFT && bcnt[1]) || (BOARD22(B8)->w == -ROOK << BOARD_SHIFT && bcnt[7]))
                bsafety -= EVAL_ROOK_BLOCKED;
        }
        //safety without -(PAWN << BOARD_SHIFT) shelter
        bsafety += 3*EVAL_KING_SAFETY;
        int x = X(POS(-KING, 0));
        //look for open files
        if (bcnt[x+1] == 0)
            bsafety += 2*EVAL_KING_SAFETY;
        if (wcnt[x+1] == 0)
            bsafety += EVAL_KING_SAFETY;

        if (x < 3) {
            bsafety -= EVAL_KING_SAFETY * ( BOARD22(A7)->w == -(PAWN << BOARD_SHIFT) ) * 2;
            bsafety -= EVAL_KING_SAFETY * ( BOARD22(B7)->w == -(PAWN << BOARD_SHIFT) ) * 2;
            bsafety -= EVAL_KING_SAFETY * ( BOARD22(C7)->w == -(PAWN << BOARD_SHIFT) ) * 2;
            bsafety -= EVAL_KING_SAFETY * ( BOARD22(A6)->w == -(PAWN << BOARD_SHIFT) );
            bsafety -= EVAL_KING_SAFETY * ( BOARD22(B6)->w == -(PAWN << BOARD_SHIFT) );
            bsafety -= EVAL_KING_SAFETY * ( BOARD22(C6)->w == -(PAWN << BOARD_SHIFT) );
        } else if (x > 4) {
            bsafety -= EVAL_KING_SAFETY * ( BOARD22(F7)->w == -(PAWN << BOARD_SHIFT) ) * 2;
            bsafety -= EVAL_KING_SAFETY * ( BOARD22(G7)->w == -(PAWN << BOARD_SHIFT) ) * 2;
            bsafety -= EVAL_KING_SAFETY * ( BOARD22(H7)->w == -(PAWN << BOARD_SHIFT) ) * 2;
            bsafety -= EVAL_KING_SAFETY * ( BOARD22(F6)->w == -(PAWN << BOARD_SHIFT) );
            bsafety -= EVAL_KING_SAFETY * ( BOARD22(G6)->w == -(PAWN << BOARD_SHIFT) );
            bsafety -= EVAL_KING_SAFETY * ( BOARD22(H6)->w == -(PAWN << BOARD_SHIFT) );
        }
    }

    pawn->struc = struc;
    pawn->wsafety = wsafety;
    pawn->bsafety = bsafety;
}

static inline void dynEvalPieces(const Board *b, HashEntry *hash, const int color) __attribute__ ((always_inline));
static inline void dynEvalPieces(const Board *b, HashEntry *hash, const int color) {
    int coor = 0;
    //     if (CNT(color*ROOK)>1) {
    //         int offset;
    //         if ((offset = DIST(ROOK, POS(color*ROOK, 0) - POS(color*ROOK, 1)))) {
    //             int src = POS(color*ROOK, 0);
    //             while (!BOARD22(src -= offset)->all) ;
    //             if (src == POS(color*ROOK, 1)) {
    //                 coor += color*EVAL_ROOK_CONNECTED;
    //             }
    //         }
    //     }

    if (CNT(color*BISHOP)>1)
        coor += color*EVAL_BISHOP_PAIR;

    int attcnt = 0;
    for (int piece = ROOK; piece <= KNIGHT; piece++)
        for (const int *ppp = &POS(color*piece, CNT(color*piece)-1); ppp >= &POS(color*piece, 0); ppp--) {
            attcnt += ATTACKTABLE(piece, *ppp - POS(-color*KING,0));
            //            printf("attack piece %d %d\n",color*piece, ATTACKTABLE(piece, *ppp - POS(-color*KING,0)));
        }
    int attack = color*attstr[attcnt];
    //    printf("attcnt %d, attstr %d\n", attcnt, attstr[attcnt]);

    hash->coor = coor;
    hash->attack = attack;
}
//dynamic part of evaluation. three values are calculated:
//pawn structure/king safety (uses positions of kings, pawns and rooks in the startposition only),
//piece coordination/attack strength (uses positions of pieces of one color and position
//of opposite king) for both colors.
//all values are hashed, the hashkeys generated should only change iff the value changes.
//this is accomplished by using "0" as a key in the random key table, if the type/position
//of a piece can't change the value calculated here. see initHash()
//always look in hashtable first, if we don't find something calculate it.
//hit ratio is better than 90%, usually 99%+, so speed in the calculation isn't that important.

float dynEval(const Board * b) {
    BEGIN_PROFILE(TIME_DYNEVAL);

    PawnEntry *pawn = (PawnEntry *) pawnHash.table + (b->key.pawn.key0 & pawnHash.mask);
    pawnHash.read++;
    if (pawn->key == b->key.pawn.key1) {
        pawnHash.hit++;
    } else {
        dynEvalPawn(b, pawn);
        if (pawn->key == 0)
            pawnHash.fill++;
        pawnHash.write++;
        pawn->key = b->key.pawn.key1;
    }

    HashEntry *white = (HashEntry *) whiteHash.table + (b->key.white.key0 & whiteHash.mask);
    whiteHash.read++;
    if (white->key == b->key.white.key1) {
        whiteHash.hit++;
    } else {
        dynEvalPieces(b, white, WHITE);
        white->key = b->key.white.key1;
        whiteHash.write++;
    }

    HashEntry *black = (HashEntry *) blackHash.table + (b->key.black.key0 & blackHash.mask);
    blackHash.read++;
    if (black->key == b->key.black.key1) {
        blackHash.hit++;
    } else {
        dynEvalPieces(b, black, BLACK);
        black->key = b->key.black.key1;
        blackHash.write++;
    }

    END_PROFILE();
    //king safety plays only a role in the middlegame, so adjust it.
    //attack strength is scaled with the opponents king safety weakness. for a perfect
    //king position the scaling term below yields 1.0, for a non existent pawn shield 2.0.
    //    printBoard(b, stdout);
    /*    printf("pawn structure: %4d, white safety: %4d, black safety: %4d\n", pawn->struc, pawn->wsafety, pawn->bsafety);
        printf("white pieces coordiantion: %4d, attack: %4d\n", white->coor, white->attack);
        printf("black pieces coordiantion: %4d, attack: %4d\n", black->coor, black->attack);*/
    return QUANT * rint
           ( pawn->struc + ((double)pawn->wsafety+(double)pawn->bsafety) * mwght[b->material]
             + white->coor + ((double)pawn->bsafety + 15*EVAL_KING_SAFETY)*white->attack * (1.0/(12.0*EVAL_KING_SAFETY))
             + black->coor - ((double)pawn->wsafety - 15*EVAL_KING_SAFETY)*black->attack * (1.0/(12.0*EVAL_KING_SAFETY))
           );
}

static inline void rayw(Board *b, int src, int offset, unsigned mask) {
    int dst = src + offset;
    while ((BOARD22(dst)->all) == 0) {
        ATTACK22(dst)->w += mask;
        dst += offset;
    }
    ATTACK22(dst)->w += mask;
}

static inline void rayb(Board *b, int src, int offset, unsigned mask) {
    int dst = src + offset;
    while ((BOARD22(dst)->all) == 0) {
        ATTACK22(dst)->b += mask;
        dst += offset;
    }
    ATTACK22(dst)->b += mask;
}

#if 0
static void buildAttackTable(Board *b) {
    for (int y=0; y<8; y++) {
        b->attack[2+y][2].all = 0;
        b->attack[2+y][3].all = 0;
        b->attack[2+y][4].all = 0;
        b->attack[2+y][5].all = 0;
        b->attack[2+y][6].all = 0;
        b->attack[2+y][7].all = 0;
        b->attack[2+y][8].all = 0;
        b->attack[2+y][9].all = 0;
    }

    for (const int *ppp = &POS(ROOK, CNT(ROOK) - 1); ppp >= &POS(ROOK, 0); ppp--) {
        int src = *ppp;
        rayw(b, src, dirtab[0], ROOK_ATTACK);
        rayw(b, src, dirtab[2], ROOK_ATTACK);
        rayw(b, src, dirtab[4], ROOK_ATTACK);
        rayw(b, src, dirtab[6], ROOK_ATTACK);
    }

    for (const int *ppp = &POS(BISHOP, CNT(BISHOP) - 1); ppp >= &POS(BISHOP, 0); ppp--) {
        int src = *ppp;
        rayw(b, src, dirtab[1], BISHOP_ATTACK);
        rayw(b, src, dirtab[3], BISHOP_ATTACK);
        rayw(b, src, dirtab[5], BISHOP_ATTACK);
        rayw(b, src, dirtab[7], BISHOP_ATTACK);
    }

    for (const int *ppp = &POS(QUEEN, CNT(QUEEN) - 1); ppp >= &POS(QUEEN, 0); ppp--) {
        int src = *ppp;
        rayw(b, src, dirtab[0], QUEEN_ATTACK);
        rayw(b, src, dirtab[2], QUEEN_ATTACK);
        rayw(b, src, dirtab[4], QUEEN_ATTACK);
        rayw(b, src, dirtab[6], QUEEN_ATTACK);
        rayw(b, src, dirtab[1], QUEEN_ATTACK);
        rayw(b, src, dirtab[3], QUEEN_ATTACK);
        rayw(b, src, dirtab[5], QUEEN_ATTACK);
        rayw(b, src, dirtab[7], QUEEN_ATTACK);
    }

    for (const int *ppp = &POS(KNIGHT, CNT(KNIGHT) - 1); ppp >= &POS(KNIGHT, 0); ppp--) {
        int src = *ppp;
        ATTACK22(src + kdirtab[0])->w += KNIGHT_ATTACK;
        ATTACK22(src + kdirtab[1])->w += KNIGHT_ATTACK;
        ATTACK22(src + kdirtab[2])->w += KNIGHT_ATTACK;
        ATTACK22(src + kdirtab[3])->w += KNIGHT_ATTACK;
        ATTACK22(src + kdirtab[4])->w += KNIGHT_ATTACK;
        ATTACK22(src + kdirtab[5])->w += KNIGHT_ATTACK;
        ATTACK22(src + kdirtab[6])->w += KNIGHT_ATTACK;
        ATTACK22(src + kdirtab[7])->w += KNIGHT_ATTACK;
    }

    {
        int src = POS(KING, 0);
        ATTACK22(src + dirtab[0])->w += KING_ATTACK;
        ATTACK22(src + dirtab[1])->w += KING_ATTACK;
        ATTACK22(src + dirtab[2])->w += KING_ATTACK;
        ATTACK22(src + dirtab[3])->w += KING_ATTACK;
        ATTACK22(src + dirtab[4])->w += KING_ATTACK;
        ATTACK22(src + dirtab[5])->w += KING_ATTACK;
        ATTACK22(src + dirtab[6])->w += KING_ATTACK;
        ATTACK22(src + dirtab[7])->w += KING_ATTACK;
    }

    for (const int *ppp = &POS(PAWN, CNT(PAWN) - 1); ppp >= &POS(PAWN, 0); ppp--) {
        int src = *ppp;
        ATTACK22(src + dirtab[1])->w += PAWN_ATTACK;
        ATTACK22(src + dirtab[3])->w += PAWN_ATTACK;
    }

    for (const int *ppp = &POS(-ROOK, CNT(-ROOK) - 1); ppp >= &POS(-ROOK, 0); ppp--) {
        int src = *ppp;
        rayb(b, src, dirtab[0], ROOK_ATTACK);
        rayb(b, src, dirtab[2], ROOK_ATTACK);
        rayb(b, src, dirtab[4], ROOK_ATTACK);
        rayb(b, src, dirtab[6], ROOK_ATTACK);
    }

    for (const int *ppp = &POS(-BISHOP, CNT(-BISHOP) - 1); ppp >= &POS(-BISHOP, 0); ppp--) {
        int src = *ppp;
        rayb(b, src, dirtab[1], BISHOP_ATTACK);
        rayb(b, src, dirtab[3], BISHOP_ATTACK);
        rayb(b, src, dirtab[5], BISHOP_ATTACK);
        rayb(b, src, dirtab[7], BISHOP_ATTACK);
    }

    for (const int *ppp = &POS(-QUEEN, CNT(-QUEEN) - 1); ppp >= &POS(-QUEEN, 0); ppp--) {
        int src = *ppp;
        rayb(b, src, dirtab[0], QUEEN_ATTACK);
        rayb(b, src, dirtab[2], QUEEN_ATTACK);
        rayb(b, src, dirtab[4], QUEEN_ATTACK);
        rayb(b, src, dirtab[6], QUEEN_ATTACK);
        rayb(b, src, dirtab[1], QUEEN_ATTACK);
        rayb(b, src, dirtab[3], QUEEN_ATTACK);
        rayb(b, src, dirtab[5], QUEEN_ATTACK);
        rayb(b, src, dirtab[7], QUEEN_ATTACK);
    }

    for (const int *ppp = &POS(-KNIGHT, CNT(-KNIGHT) - 1); ppp >= &POS(-KNIGHT, 0); ppp--) {
        int src = *ppp;
        ATTACK22(src + kdirtab[0])->b += KNIGHT_ATTACK;
        ATTACK22(src + kdirtab[1])->b += KNIGHT_ATTACK;
        ATTACK22(src + kdirtab[2])->b += KNIGHT_ATTACK;
        ATTACK22(src + kdirtab[3])->b += KNIGHT_ATTACK;
        ATTACK22(src + kdirtab[4])->b += KNIGHT_ATTACK;
        ATTACK22(src + kdirtab[5])->b += KNIGHT_ATTACK;
        ATTACK22(src + kdirtab[6])->b += KNIGHT_ATTACK;
        ATTACK22(src + kdirtab[7])->b += KNIGHT_ATTACK;
    }

    {
        int src = POS(-KING, 0);
        ATTACK22(src + dirtab[0])->b += KING_ATTACK;
        ATTACK22(src + dirtab[1])->b += KING_ATTACK;
        ATTACK22(src + dirtab[2])->b += KING_ATTACK;
        ATTACK22(src + dirtab[3])->b += KING_ATTACK;
        ATTACK22(src + dirtab[4])->b += KING_ATTACK;
        ATTACK22(src + dirtab[5])->b += KING_ATTACK;
        ATTACK22(src + dirtab[6])->b += KING_ATTACK;
        ATTACK22(src + dirtab[7])->b += KING_ATTACK;
    }

    for (const int *ppp = &POS(-PAWN, CNT(-PAWN) - 1); ppp >= &POS(-PAWN, 0); ppp--) {
        int src = *ppp;
        ATTACK22(src + dirtab[1])->b += PAWN_ATTACK;
        ATTACK22(src + dirtab[3])->b += PAWN_ATTACK;
    }
}
#endif
//static part of attack evaluation
static void ray(Board *b, int x, int y, int dx, int dy, int p1, int p2, float d) {
    for (x += dx, y += dy; x < 8 && y < 8 && x >= 0 && y >= 0; x += dx, y += dy) {
        VAL(p1, y, x) += d;
        VAL(p2, y, x) += d * 0.5;
        if (abs(TYPE(y, x)) == PAWN)
            d *= 0.5;
    }
}

static void evalAttack(Board *b, float a) {
    int px, py, k, x;

    for (k = -1; k <= 1; k += 2) {
        px = X(POS(k * KING, 0));
        py = Y(POS(k * KING, 0));
        if (px == 7)
            px = 6;
        else if (px == 0)
            px = 1;

        for (x = -1; x <= 1; x++) {

            ray(b, px + x, py + k, 0, k, -ROOK * k, -QUEEN * k, -a * k);
            ray(b, px + x, py + k, 1, k, -BISHOP * k, -QUEEN * k, -a * k);
            ray(b, px + x, py + k, -1, k, -BISHOP * k, -QUEEN * k, -a * k);

            ray(b, px + x, py, 1, k, -BISHOP * k, -QUEEN * k, -a * k);
            ray(b, px + x, py, -1, k, -BISHOP * k, -QUEEN * k, -a * k);

            VAL(-PAWN*k, (7 - 5*k)/2, x+px) -= 4*a*k;
            VAL(-PAWN*k, (7 - 3*k)/2, x+px) -= 2*a*k;
            VAL(-PAWN*k, (7 - 1*k)/2, x+px) -= 2*a*k;
            VAL(-PAWN*k, (7 + 1*k)/2, x+px) -= a*k;
        }
    }
}

static void fillVal(Board *b, int type, const float midval[][8], const float endval[][8], float gw, float bias) {
    for (int y = 0; y < 8; y++)
        for (int x = 0; x < 8; x++) {
            VAL( type, y, x) = bias + gw + (midval[  y][x]*mwght[b->material] + endval[  y][x]*ewght[b->material]) * QUANT;
            VAL(-type, y, x) = bias - gw - (midval[7-y][x]*mwght[b->material] + endval[7-y][x]*ewght[b->material]) * QUANT;
        }
}

static void evalMobility(Board *b, float w) {
    w=w;
}

static void evalCenter(Board *b, float w) {
    int x, y;

    for (int col = -1; col <= 1; col += 2) {
        for (x = 2; x < 6; x++)
            for (y = 2; y < 6; y++) {
                VAL(col * PAWN, col == 1 ? y : 7 - y, x) += col * w;
            }
        for (x = 3; x < 5; x++)
            for (y = 3; y < 6; y++) {
                VAL(col * PAWN, col == 1 ? y : 7 - y, x) += col * w;
            }

        VAL(col * PAWN, col == 1 ? 1 : 7 - 1, 3) -= col * (w + w);
        VAL(col * PAWN, col == 1 ? 1 : 7 - 1, 4) -= col * (w + w);
        VAL(col * BISHOP, col == 1 ? 0 : 7 - 0, 2) -= col * (w);
        VAL(col * BISHOP, col == 1 ? 0 : 7 - 0, 5) -= col * (w);
        VAL(col * KNIGHT, col == 1 ? 0 : 7 - 0, 1) -= col * (w + w);
        VAL(col * KNIGHT, col == 1 ? 0 : 7 - 0, 6) -= col * (w + w);
    }

}
static void evalRook() {
    /*
       int x, w;

       if (Y(POS(-KING, 0)) == 7)
       w = ROOK7KING8;
       else
       w = 0;

       for (x = 7; x >= 0; x--)
       if (TYPE(6, x) == -PAWN) {
       w += ROOK7PAWN;
       break;
       }

       for (x = 7; x >= 0; x--)
       VAL(ROOK, 6, x) += w;
     */

}

static void evalBishop(Board *b, float w) {
    //give pawns on opposite colors a malus
    for (int col = -1; col <= 1; col += 2)
        if (CNT(col * BISHOP) == 1) {
            int bcol = (X(POS(col * BISHOP, 0)) + Y(POS(col * BISHOP, 0))) & 1;
            for (int y = 0; y < 8; y++)
                for (int x = (y & 1) ^ bcol; x < 8; x += 2) {
                    VAL(PAWN, y, x) -= w;
                    VAL(-PAWN, y, x) += w;
                }
        }
}

static void evalPawn(Board *b, int piecebits, float tw) {

    int bweakpawn = 0;
    int wweakpawn = 0;
    unsigned int ii;
    int x, y;
    int wkx, wkydist;
    int bkx, bkydist;
    int dist;

    int wmin[10], bmin[10]; //pos des jeweils rckstndigsten bauern auf [linie-1]
    unsigned int wcnt[10], bcnt[10];        //anazhl der bauern auf [linie-1]

    for (ii = 0; ii < 10; ii++) {
        wmin[ii] = 7;
        bmin[ii] = 0;
        wcnt[ii] = bcnt[ii] = 0;
    }

    for (ii = 0; ii < CNT(PAWN); ii++) {
        x = X(POS(PAWN, ii)) + 1;
        y = Y(POS(PAWN, ii));
        if (wmin[x] > y)
            wmin[x] = y;
        wcnt[x]++;
    }

    for (ii = 0; ii < CNT(-PAWN); ii++) {
        x = X(POS(-PAWN, ii)) + 1;
        y = Y(POS(-PAWN, ii));
        if (bmin[x] < y)
            bmin[x] = y;
        bcnt[x]++;
    }

    wkx = X(POS(KING, 0));
    wkydist = Y(POS(KING, 0));
    bkx = X(POS(-KING, 0));
    bkydist = 7-Y(POS(-KING, 0));

    for (ii = 1; ii < 9; ii++) {

        y = bmin[ii - 1];
        if (bmin[ii + 1] > y)
            y = bmin[ii + 1];
        for (; y < 7; y++)
            VAL(KNIGHT, y, ii - 1) += 0.125;

        y = wmin[ii - 1];
        if (wmin[ii + 1] < y)
            y = wmin[ii + 1];
        for (; y > 0; y--)
            VAL(-KNIGHT, y, ii - 1) -= 0.125;

        if (wcnt[ii]) {

            if (!bcnt[ii] && wmin[ii - 1] > wmin[ii] && wmin[ii + 1] > wmin[ii]) {
                wweakpawn++;
                for (y = wmin[ii] + 1; y < 8; y++)
                    VAL(-ROOK, y, ii - 1) -= tw;
            }

            if (bmin[ii - 1] <= wmin[ii] && bmin[ii] < wmin[ii] && bmin[ii + 1] <= wmin[ii]) {
                //passedPawn
                for (y = 7; TYPE(y, ii - 1) != PAWN; y--) {
                    //move pieces out of the way
                    VAL(ROOK, y, ii - 1) -= tw;
                    VAL(KNIGHT, y, ii - 1) -= tw;
                    VAL(BISHOP, y, ii - 1) -= tw;
                    //use knight or bishop for blocking pp
                    VAL(-KNIGHT, y, ii - 1) -= tw * 0.5;
                    VAL(-BISHOP, y, ii - 1) -= tw * 0.5;
                }

//this need to be moved to dynamic eval
#if 0
                if (!(piecebits & (QUEENBIT | ROOKBIT | LIGHTBIT))) {
                    dist = abs(ii - 1 - bkx);
                    if (bkydist > dist)
                        dist = bkydist;
                    if (7 - y < dist - (b->color == BLACK))
                        b->currMove->move.val += 5.0;
                }
#endif
                for (; y >= 0; y--) {
                    //rooks behind pp
                    VAL(ROOK, y, ii - 1) += tw;
                    VAL(-ROOK, y, ii - 1) -= tw;
                }
            }
        }

        if (bcnt[ii]) {

            if (!wcnt[ii] && bmin[ii - 1] < bmin[ii] && bmin[ii + 1] < bmin[ii]) {
                bweakpawn++;
                for (y = bmin[ii] - 1; y >= 0; y--)
                    VAL(ROOK, y, ii - 1) += tw;
            }

            if (wmin[ii - 1] >= bmin[ii] && wmin[ii] > bmin[ii] && wmin[ii + 1] >= bmin[ii]) {
                for (y = 0; TYPE(y, ii - 1) != -PAWN; y++) {
                    VAL(-ROOK, y, ii - 1) += tw;
                    VAL(-KNIGHT, y, ii - 1) += tw;
                    VAL(-BISHOP, y, ii - 1) += tw;
                    VAL(KNIGHT, y, ii - 1) += tw * 0.5;
                    VAL(BISHOP, y, ii - 1) += tw * 0.5;
                }

//this need to be moved to dynamic eval
#if 0
                if (!(piecebits & (QUEENBIT | ROOKBIT | LIGHTBIT))) {
                    dist = abs(ii - 1 - wkx);
                    if (wkydist > dist)
                        dist = wkydist;
                    if (y < dist - (b->color == WHITE))
                        b->currMove->move.val -= 5.0;
                }
#endif
                for (; y < 8; y++) {
                    VAL(ROOK, y, ii - 1) += tw;
                    VAL(-ROOK, y, ii - 1) -= tw;
                }
            }
        }

        if (!bcnt[ii] && !wcnt[ii])
            //open file
            for (y = 7; y >= 0; y--) {
                VAL(ROOK, y, ii - 1) += tw;
                VAL(-ROOK, y, ii - 1) -= tw;
            }
    }
}

static void fillk(Board *b, float f1, float f2) {

    int x, y;

    for (y = 0; y < 8; y++)
        for (x = 0; x < 8; x++) {
            VAL(KING, y, x) = (abs(7 - x - x) + abs(7 - y - y)) * f1 * 0.0078125000;
            VAL(-KING, y, x) = (abs(7 - x - x) + abs(7 - y - y)) * f2 * 0.0078125000;
        }
}

static void fillk2(Board *b, int k, float f1, float f2, float f3) {

    int x, y;

    for (y = 0; y < 8; y++)
        for (x = 0; x < 8; x++) {
            VAL(k, y, x) = abs(x + y * f3) * f1 * 0.0078125;
            VAL(-k, y, x) = (abs(7 - x - x) + abs(7 - y - y)) * f2 * 0.0078125;
        }
}

int eval(Board * b) {

    float bias = 0.0;
    unsigned int piecebits;


    //values are in QUANTS (1/64 of a pawn)
    static const float pawn1[8][8] = {
                                         {
                                             0,   0,   0,   0,   0,   0,   0,   0
                                         },
                                         {    1,   1,   1,   2,   2,   1,   1,   1},
                                         {    0,   0,   2,   3,   3,   2,   0,   0},
                                         {    0,   0,   2,   6,   5,   2,   0,   0},
                                         {    0,   2,   3,   6,   5,   2,   2,   0},
                                         {    2,   3,   4,   6,   4,   3,   3,   2},
                                         {    2,   2,   2,   2,   2,   2,   2,   2},
                                         {    0,   0,   0,   0,   0,   0,   0,   0}
                                     };

    static const float knight1[8][8] = {
                                           {
                                               -7,  -6,  -5,  -5,  -5,  -5,  -6,  -7
                                           },
                                           {   -6,  -3,   0,   1,   1,   0,  -3,  -6},
                                           {   -5,   1,   2,   2,   2,   2,   1,  -5},
                                           {   -5,   1,   3,   4,   4,   3,   1,  -5},
                                           {   -5,   2,   4,   5,   5,   4,   2,  -5},
                                           {   -5,   2,   4,   4,   4,   4,   2,  -5},
                                           {  -11,  -7,   2,   3,   3,   2,  -7, -11},
                                           {  -15,  -8,  -5,  -5,  -5,  -5,  -8, -15}
                                       };

    static const float bishop1[8][8] = {
                                           {
                                               -7,  -6,  -5,  -5,  -5,  -5,  -6,  -7
                                           },
                                           {   -6,  -3,   0,   1,   1,   0,  -3,  -6},
                                           {   -5,   1,   2,   2,   2,   2,   1,  -5},
                                           {   -5,   1,   3,   4,   4,   3,   1,  -5},
                                           {   -5,   2,   4,   5,   5,   4,   2,  -5},
                                           {   -5,   2,   4,   4,   4,   4,   2,  -5},
                                           {  -11,  -7,   2,   3,   3,   2,  -7, -11},
                                           {  -15,  -8,  -5,  -5,  -5,  -5,  -8, -15}
                                       };

    static const float rook1[8][8] = {
                                         {
                                             0,   0,   2,   3,   3,   2,   0,   0
                                         },
                                         {    0,   0,   2,   3,   3,   2,   0,   0},
                                         {    0,   0,   2,   3,   3,   2,   0,   0},
                                         {    0,   0,   2,   3,   3,   2,   0,   0},
                                         {    0,   0,   2,   3,   3,   2,   0,   0},
                                         {    2,   3,   4,   5,   5,   4,   3,   2},
                                         {    3,   3,   5,   6,   6,   5,   3,   3},
                                         {    3,   3,   5,   6,   6,   5,   3,   3}
                                     };

    static const float queen1[8][8] = {
                                          {
                                              -3,  -2,   1,   1,   1,   1,  -2,  -3
                                          },
                                          {   -2,   2,   4,   5,   5,   4,   2,  -2},
                                          {   -2,  -1,   3,   6,   6,   3,  -1,  -2},
                                          {   -1,   3,   6,   7,   7,   6,   3,  -1},
                                          {    3,   3,   6,   7,   7,   6,   3,   3},
                                          {    3,   4,   5,   7,   7,   5,   4,   3},
                                          {   -2,   3,   5,   5,   5,   5,   3,  -2},
                                          {    2,   2,   2,   3,   3,   2,   2,   2}
                                      };

    static const float king1[8][8] = {
                                         {
                                             0,   0,  -4,  -4,  -4,  -4,   0,   0
                                         },
                                         {   -4,  -8, -16  -16, -16, -16,  -8,  -4},
                                         {  -64, -64, -64, -64, -64, -64, -64, -64},
                                         {  -64, -64, -64, -64, -64, -64, -64, -64},
                                         {  -64, -64, -64, -64, -64, -64, -64, -64},
                                         {  -64, -64, -64, -64, -64, -64, -64, -64},
                                         {  -64, -64, -64, -64, -64, -64, -64, -64},
                                         {  -64, -64, -64, -64, -64, -64, -64, -64},
                                     };

    static const float king2[8][8] = {
                                         {
                                             -16,  -8,  -4,  -4,  -4,  -4,  -8, -16
                                         },
                                         {   -8,  -4,   0,   0,   0,   0,  -4,  -8},
                                         {   -4,   0,   4,   4,   4,   4,   0,  -4},
                                         {   -4,   0,   4,   6,   6,   4,   0,  -4},
                                         {   -4,   0,   6,   6,   6,   6,   0,  -4},
                                         {   -4,   0,   4,   6,   6,   4,   0,  -4},
                                         {   -6,  -2,   2,   2,   2,   2,  -2,  -6},
                                         {  -16,  -8,  -4,  -4,  -4,  -4,  -8, -16},
                                     };

    static const float pawn2[8][8] = {
                                         {
                                             0,   0,   0,   0,   0,   0,   0,   0
                                         },
                                         {    0,   0,   0,   0,   0,   0,   0,   0},
                                         {    0,   0,   0,   0,   0,   0,   0,   0},
                                         {    1,   1,   1,   1,   1,   1,   1,   1},
                                         {    3,   3,   3,   3,   3,   3,   3,   3},
                                         {    6,   6,   6,   6,   6,   6,   6,   6},
                                         {   10,  10,  10,  10,  10,  10,  10,  10},
                                         {    0,   0,   0,   0,   0,   0,   0,   0}
                                     };

    static const int matlist[] = {
                                     0, 5, 3, 9, 3, 1, 0
                                 };

    unsigned int i, j;
    int p;

    piecebits = 0;
    int wmat = 0;
    int bmat = 0;

    b->currMove->move.val = 0.0;

    for (i = 1; i <= 6; i++) {
        wmat += matlist[i] * CNT(i);
        bmat += matlist[i] * CNT(-i);
        if (CNT(i) || CNT(-i))
            piecebits |= 1 << i;
    }
    int mat = wmat + bmat;
    int matadv = wmat - bmat;

    if ((mat == 3 && piecebits & LIGHTBIT)      //only one light pieces left
            || (mat == 0)           //no piece left
            || (mat == 6 && matadv == 0 && !(piecebits & PAWNBIT))) {       //one light piece on every side
        return (1);             //return immediately with remis flag
    }

    if (!(piecebits & PAWNBIT)) {       //no pawns left
        fillVal(b, ROOK, rook1, rook1, 5.0, bias);
        fillVal(b, BISHOP, bishop1, bishop1, 3.0, bias);
        fillVal(b, KNIGHT, knight1, knight1, 2.75, bias);
        fillVal(b, QUEEN, queen1, queen1, 9.0, bias);
        fillVal(b, KING, king1, king2, 0, bias);
        fillVal(b, PAWN, pawn1, pawn2, 0.875, bias);

        if (mat == 6 && abs(matadv) == 6 && (piecebits & LIGHTBIT) == LIGHTBIT)
            //create special king tables for KNBK endgame
            if (matadv > 0)
                fillk2(b, KING, 3, -1, 2 - (Y(POS(BISHOP, 0)) ^ X(POS(BISHOP, 0))));
            else
                fillk2(b, -KING, 1, -3, 2 - (Y(POS(-BISHOP, 0)) ^ X(POS(-BISHOP, 0))));
        else if (matadv > 0)
            fillk(b, -1, 3);
        else
            fillk(b, -3, 1);
    } else {
        fillVal(b, ROOK, rook1, rook1, 5.0, bias);
        fillVal(b, BISHOP, bishop1, bishop1, 3.0, bias);
        fillVal(b, KNIGHT, knight1, knight1, 2.75, bias);
        fillVal(b, QUEEN, queen1, queen1, 9.0, bias);
        fillVal(b, KING, king1, king2, 0, bias);
        fillVal(b, PAWN, pawn1, pawn2, 0.875, bias);

        evalPawn(b, piecebits, 0.25);
        evalRook();
        evalBishop(b, 0x0.08p0);
        evalCenter(b, 0x0.2p0 * mwght[b->material]);
        evalAttack(b, 0.125 * mwght[b->material]);
    }

    //if we got huge material advantage or more pawns,
    //make our pieces less valuable to encourage exchanges
    if (matadv) {
        bias = matadv * 0x0.2p0;
        bias = min(bias, 0x0.4p0);
        bias = max(bias, -0x0.4p0);
        for (p = -4; p <= 4; p++)
            if (p) {
                for (i = 0; i < 8; i++)
                    for (j = 0; j < 8; j++)
                        VAL(p, j, i) -= bias;
                b->currMove->move.val += CNT(p) * bias;
            }
    }

    for (p = -KING; p <= KING; p++)
        if (p)
            for (i = 0; i < 8; i++)
                for (j = 0; j < 8; j++)
                    VAL(p, j, i) = QUANT * rint(VAL(p, j, i) * (1/QUANT));

    for (i = 0; i < 8; i++)
        for (j = 0; j < 8; j++)

            if (TYPE(j, i))
                b->currMove->move.val += VAL(TYPE(j, i), j, i);

#include <stdio.h>

    FILE *f;
    int x, y;
    f = fopen("eval.csv", "w");
    if (f) {
        for (p = -KING; p <= KING; p++)
            if (p)
                for (y = 7; y >= 0; y--) {
                    for (x = 0; x < 8; x++) {
                        fprintf(f, "%f", VAL(p, y, x));
                        if (x < 7)
                            fprintf(f, " ");
                    }
                    fprintf(f, "\n");
                }

        fclose(f);
    }
    return 0;

}

void initEval() {
    for (int i=0; i<256; i++) {
        mwght[i] = 0.5 + erf((i-ENDGAME_MAT)*0.333)*0.5; //50% endgame is at material 30, 3 material more or less 10/%90%
        ewght[i] = 0.5 - erf((i-ENDGAME_MAT)*0.333)*0.5;
    }

    //this gives us 0.25*KING_ATTACK with 3 attacking pieces
    //0.5 with 4 and 0.75 with 5.
    for (int i=0; i<64; i++) {
        attstr[i] = lrint(0.5 + erf((i-16)*0.121) * EVAL_KING_ATTACK);
        //        printf("%d %d\n", i/4, attstr[i]);
    }

#define NDIST(ptr) ( *(int *)( (char *)&ndistTable[8][8] + (ptr) ) )
    int ndistTable[18][16];
    for (int x=0; x<18*16; x++)
        ndistTable[0][x] = 999;

    NDIST(0) = 0;
    for (int i=0; i<10; i++)
        for (int x=-6; x<7; x++)
            for (int y=-6; y<7; y++) {
                int src = ADR(y, x);
                if (NDIST(src) == 999)
                    for (int dir = 0; dir < 8; dir++) {
                        int dst = src + kdirtab[dir];
                        if ( dst >= -(signed)sizeof(ndistTable)/2 && dst < (signed)sizeof(ndistTable)/2 )
                            if (NDIST(dst) == i) {
                                NDIST(src) = i+1;
                                break;
                            }
                    }
            }

    for (int x=-7; x<=7; x++)
        for (int y=-7; y<=7; y++) {
            int dist = max(abs(x), abs(y));
            int dia = abs(abs(x) - abs(y)) + max(2*dist-8, 0);
            int line = abs(abs(x-y) - abs(x+y)) + max(2*dist-9, 0);

            attackTable[ROOK][y+8][x+8] = min(max(5-line, 0), 4);
            attackTable[BISHOP][y+8][x+8] = min(max(5-dia, 0), 4);
            attackTable[QUEEN][y+8][x+8] = max(min(max(6-line, 6-dia), 5), 0);
            attackTable[KNIGHT][y+8][x+8] = max(min(8-(   ndistTable[y+7][x+7] + ndistTable[y+8][x+7] + ndistTable[y+9][x+7]
                                                    + ndistTable[y+7][x+8] + ndistTable[y+8][x+8] + ndistTable[y+9][x+8]
                                                    + ndistTable[y+7][x+9] + ndistTable[y+8][x+9] + ndistTable[y+9][x+9])/4, 4), 0);
            attackTable[KING][y+8][x+8] = min(max(8-2*dist, 0), 4);
            attackTable[PAWN][y+8][x+8] = min(max(8-2*dist, 0), 4);
        }
    /*   for (int x=-7; x<=7; x++) {
           for (int y=-7; y<=7; y++) 
               printf("%4d", attackTable[ROOK][y+8][x+8]);
           printf("\n");
       }            */
}

float slowEval(const Board * b) {
    float score = 0.0;
    for (int p = 0; p < 13; p++)
        if (p != 6) {
            for (unsigned i = 0; i < b->pieces[p].cnt; i++) {
                score += b->values[p][0][b->pieces[p].pos[i] / 4];
            }
        }
    return score;
}

unsigned long long slowHash(const Board * b) {
    unsigned long long hash = 0;
    for (int p = 0; p < 13; p++)
        if (p != 6) {
            for (unsigned i = 0; i < b->pieces[p].cnt; i++) {
                hash ^= hashKeys[p][0][b->pieces[p].pos[i] / 4].main.key;
            }
        }
    return hash;
}

#if 0
static inline int attackray(const Board *b, int src, const int offset, const int type1, const int type2) {
    int dst = src + offset;
    int ret = 0;
    int attack = 4;
    while (!(dst & 0x220)) {
        if (BOARD22(dst)->w == type1)
            ret += (attack*3)/2;
        else if (BOARD22(dst)->w == type2)
            ret += attack;
        else if (BOARD22(dst)->w == PAWN << BOARD_SHIFT || BOARD22(dst)->w == -PAWN << BOARD_SHIFT)
            attack >>= 1;
        dst += offset;
    }
    return ret;
}

int attackcntw(const Board *b, int src) {
    return attackray(b, src, dirtab[0], QUEEN << BOARD_SHIFT, ROOK << BOARD_SHIFT)
           + attackray(b, src, dirtab[4], QUEEN << BOARD_SHIFT, ROOK << BOARD_SHIFT)
           + attackray(b, src, dirtab[5], QUEEN << BOARD_SHIFT, BISHOP << BOARD_SHIFT)
           + attackray(b, src, dirtab[6], QUEEN << BOARD_SHIFT, ROOK << BOARD_SHIFT)
           + attackray(b, src, dirtab[7], QUEEN << BOARD_SHIFT, BISHOP << BOARD_SHIFT)
           + 4 * ( (BOARD22(src + kdirtab[0])->w == KNIGHT << BOARD_SHIFT)
                   + (BOARD22(src + kdirtab[1])->w == KNIGHT << BOARD_SHIFT)
                   + (BOARD22(src + kdirtab[2])->w == KNIGHT << BOARD_SHIFT)
                   + (BOARD22(src + kdirtab[3])->w == KNIGHT << BOARD_SHIFT)
                   + (BOARD22(src + kdirtab[4])->w == KNIGHT << BOARD_SHIFT)
                   + (BOARD22(src + kdirtab[5])->w == KNIGHT << BOARD_SHIFT)
                   + (BOARD22(src + kdirtab[6])->w == KNIGHT << BOARD_SHIFT)
                   + (BOARD22(src + kdirtab[7])->w == KNIGHT << BOARD_SHIFT));
}

int attackcntb(const Board *b, int src) {
    return attackray(b, src, dirtab[0], -QUEEN << BOARD_SHIFT, -ROOK << BOARD_SHIFT)
           + attackray(b, src, dirtab[4], -QUEEN << BOARD_SHIFT, -ROOK << BOARD_SHIFT)
           + attackray(b, src, dirtab[1], -QUEEN << BOARD_SHIFT, -BISHOP << BOARD_SHIFT)
           + attackray(b, src, dirtab[2], -QUEEN << BOARD_SHIFT, -ROOK << BOARD_SHIFT)
           + attackray(b, src, dirtab[3], -QUEEN << BOARD_SHIFT, -BISHOP << BOARD_SHIFT)
           + 4 * ( (BOARD22(src + kdirtab[0])->w == -KNIGHT << BOARD_SHIFT)
                   + (BOARD22(src + kdirtab[1])->w == -KNIGHT << BOARD_SHIFT)
                   + (BOARD22(src + kdirtab[2])->w == -KNIGHT << BOARD_SHIFT)
                   + (BOARD22(src + kdirtab[3])->w == -KNIGHT << BOARD_SHIFT)
                   + (BOARD22(src + kdirtab[4])->w == -KNIGHT << BOARD_SHIFT)
                   + (BOARD22(src + kdirtab[5])->w == -KNIGHT << BOARD_SHIFT)
                   + (BOARD22(src + kdirtab[6])->w == -KNIGHT << BOARD_SHIFT)
                   + (BOARD22(src + kdirtab[7])->w == -KNIGHT << BOARD_SHIFT));
}

#endif
