/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#ifndef GLOBALS_H
#define GLOBALS_H

#define SIMPLEBOARD

#define PLY 0x100
#define MAXDEPTH (64*PLY)            //max plys
#define MAXGAMELEN 1024
#define MAXMOVES 256		    //maximum number of moves per position.
#define QUANT (0x0.04P0)        //smallest evaluation step
#define MATE_VALUE (-32768.0*QUANT)
#define INF (128.0)             //9 queens, 2 rooks+bishops+knights

#define ROOK 1
#define BISHOP 2
#define QUEEN 3
#define KNIGHT 4
#define PAWN 5
#define KING 6

#define WHITE 1
#define BLACK -1

//castling bits
#define CAST_WQ 0x1
#define CAST_WK 0x2
#define CAST_BQ 0x4
#define CAST_BK 0x8

//special moves
#define SPEC_PROMOQ  0x1
#define SPEC_PROMON  0x2
#define SPEC_PROMOB  0x3
#define SPEC_PAWNEP  0x5
#define SPEC_CASTLEQ 0x6
#define SPEC_CASTLEK 0x7
#define SPEC_PAWNDBL 0x9
#define SPEC_ROOK    0x8
#define SPEC_KING    0xa
#define SPEC_PROMOR  0xb      
#define SPEC_ILLEGAL 0xf

#define S       *1000000                //S for converting s to s

#define REPTABLESIZE 1024


#ifndef ASSERT
#ifdef MYDEBUG
#define ASSERT(x)   if (!(x)) { \
                    fprintf(stderr, "Assertion %s failed. File %s, line %d (%s)", \
                    #x, __FILE__, __LINE__, __PRETTY_FUNCTION__); \
                    asm("int3");}
#else 
#define ASSERT(x)
#endif
#endif

#define likely(x)	__builtin_expect((x),1)
#define unlikely(x)	__builtin_expect((x),0)
#define prefetch(addr) __builtin_prefetch(addr, 0, 3)
#define prefetchw(addr) __builtin_prefetch(addr, 1, 3)

#define max(a,b) \
       ({ typeof (a) _a = (a); \
          typeof (b) _b = (b); \
          _a > _b ? _a : _b; })
#define min(a,b) \
       ({ typeof (a) _a = (a); \
          typeof (b) _b = (b); \
          _a < _b ? _a : _b; })
          
typedef struct _Move {
    float val;
    char piece;
    char attack;
    char capture;
    char special;
    int src;
    int dst;
} Move;

typedef union _piece_t {
    long all;
    struct {
        short w;
        short b;
    };
} piece_t;

typedef struct _piecel_t {
    unsigned int cnt;
    int pos[15];
} piecel_t;

#endif
