/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#ifndef HASH_H
#define HASH_H

#define HASH_BOUND_STR " =><"
#define PERM_VALUE 0
#define EXACT_VALUE 1
#define UPPER_BOUND 2
#define LOWER_BOUND 3

//SMALL_HASH currently no longer functional
//#define SMALLHASH
#define DOUBLEHASH

#define HASH_PTR(key, color) ((MainEntry *) mainHash.table + ((key.key0 ^ color) & mainHash.mask))
#define CURR_HASH_PTR HASH_PTR(b->key.main, b->color)
#define HASHKEY(piece, pos) (*((HashKey *)((char *)&hashKeys[6+(piece)][0][0] + (pos)*(sizeof(HashKey)/sizeof(int)))))

#define HASH_VALUE(p) ((p)->val * QUANT)
#define HASH_CMP(p, key) ((p)->key1 == key.key1 && p->key0 == ((unsigned short *)&(key))[1]) 

struct _constBoard;
struct _Move;

union u64 {
    unsigned long long key;
    struct {
        unsigned key0;
        unsigned key1;
    };    
};

typedef struct _MainEntry {
    unsigned key1;          //bit 32-63 of key
    int val:16; //score in quants
    int ply:16;         //the ply of the stored pos
    unsigned src:7;         //the src member of struct move >> 2
    unsigned endofsearch:1;
    unsigned dst:7;         //the dst member of move >> 2
    unsigned overwrite:1;
    unsigned special:4;     //special member of struct move
    unsigned bound:2;
} MainEntry;

typedef struct _PermEntry {
    union u64 key;          //bit 32-63 of key
    int val:16;         //score in quants
    int ply:16;         //the ply of the stored pos
    unsigned src:7;         //the src member of struct move >> 2
    unsigned dst:7;         //the dst member of move >> 2
    unsigned special:4;     //special member of struct move
} PermEntry;

typedef struct _HashEntry {
    unsigned key;
    short coor;
    short attack;
} HashEntry;

typedef struct _PawnEntry {
    unsigned key;
    short struc;
    short wsafety;
    short bsafety;
} PawnEntry;

typedef struct _HashInfo {
    unsigned size;
    unsigned count;
    unsigned mask;
    void *table;
    unsigned long long hit;
    unsigned long long read;
    unsigned long long write;
    unsigned fill;
    unsigned entrySize;
} HashInfo;

    
typedef struct _HashKey {
    union u64 main, pawn, white, black;          
} HashKey;

extern HashKey hashKeys[13][8][16];

extern HashInfo mainHash, pawnHash, whiteHash, blackHash, permHash;
extern char permInfoFileName[256];
extern char permHashFileName[256];
extern char permTempFileName[256];
extern int permPly;
    
void initHash(void);
void allocHash(HashInfo * h, unsigned bits);
void purgeHash(void);
void clearHash(void);
void writeHash(const struct _constBoard * b, int ply, float val, float alpha, float beta, const struct _Move *m, int endofsearch);
void flushHash(void);
void preloadHash(void);
void exitHash(void);
#endif

