/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "board.h"
#include "eval.h"
#include "movegen_c.h"
#include "squares.h"
#include "debug.h"

#include <string.h>

#define BOARD_SHIFT 9

//array[p][d] with a value dirtab[dir] for every distance d from a piece p,
//if p is attacking the square in distance d from direction dir, otherwise 0.
//maximum distance betwen 2 pieces is 7*(1,1), with a row = 16 columns. for negative distances too, thus 2*

int distTable[6][16][16];

int ndistTable[16][16];

#ifdef SIMPLEBOARD

  #include "movegen_c_simple.c" 
  Move *movegenq1(const constBoard *b, Move *m, float alpha) {
    BEGIN_PROFILE(TIME_MOVEGENQ);
    Move *endlist = b->color>0 ? movegenqw1(b, m, alpha) : movegenqb1(b, m, alpha);
    END_PROFILE();
    return endlist;
  }

#else
  
  #include "movegen_c_complex.c"
  Move *movegenq1(const constBoard *b, Move *m, float alpha) {
    BEGIN_PROFILE(TIME_MOVEGENQ);
    Move *endlist = b->color>0 ? movegenqw(b, m, alpha) : movegenqb(b, m, alpha);
    END_PROFILE();
    return endlist;
  }

  
#endif


Move *movegen(const constBoard *b, Move *m) {
    BEGIN_PROFILE(TIME_MOVEGEN);
    Move *endlist = b->color>0 ? movegenw(b, m) : movegenb(b, m);
    END_PROFILE();
    return endlist;
}

Move *movegenq(const constBoard *b, Move *m, float alpha) {
    BEGIN_PROFILE(TIME_MOVEGENQ);
    Move *endlist = b->color>0 ? movegenqw(b, m, alpha) : movegenqb(b, m, alpha);
    END_PROFILE();
    return endlist;
}

//usually called after domove, to verify the move done was legal (the side wich was moving
//isn't in check) and to build a attack signature pattern for the opposite color, which
//can be used in the following call of movegen.
int inCheckSig(Board *b) {
    BEGIN_PROFILE(TIME_INCHECKSIG);
    int ret;
    if (b->color > 0) {
        /*if (b->currMove->move.piece != KING)
            return attackb(b, POS(KING, 0));
        else */{
            b->bsig += 2;
            ret = attackbsig(b, POS(KING, 0), b->bsig);
        }
    } else {
        /*if (b->currMove->move.piece != -KING)
            return attackw(b, POS(KING, 0));
        else */{
            b->wsig += 2;
            ret = attackwsig(b, POS(-KING, 0), b->wsig);
        }
    }
    END_PROFILE();
    return ret;
}


void movegenqInit() {
    for (int d = 1; d < 8; d++) {
        for (int dir = 0; dir < 8; dir++)
            DIST(QUEEN, dirtab[dir] * d) = dirtab[dir];
        for (int dir = 0; dir < 8; dir += 2)
            DIST(ROOK, dirtab[dir] * d) = dirtab[dir];
        for (int dir = 1; dir < 8; dir += 2)
            DIST(BISHOP, dirtab[dir] * d) = dirtab[dir];
    }
    for (int dir = 0; dir < 8; dir++)
        DIST(KING, dirtab[dir]) = 1;

    DIST(PAWN, dirtab[1]) = 1;
    DIST(PAWN, dirtab[3]) = 1;

    for (int dir = 0; dir < 8; dir++)
        DIST(KNIGHT, kdirtab[dir]) = 1;
}

