//writes a move into the move list, additionally sets the attackflag, if the signature
//matches.
inline static Move *insertMovew(const constBoard *b, Move * m, int src, int dst, int special, int piece, float val) {
    m->val = val;
    m->attack = PEXT(piece, dst) == b->wsig;
    m->special = special;
    m->piece = piece;
    m->src = src;
    m->dst = dst;
    return ++m;
}

inline static Move *insertMoveb(const constBoard *b, Move * m, int src, int dst, const int special, int piece, float val) {
    m->val = val;
    m->attack = PEXT(piece, dst) == b->bsig;
    m->special = special;
    m->piece = piece;
    m->src = src;
    m->dst = dst;
    return ++m;
}

//checks if any square from square pos in direction dir is occupied by a piece of type
//type3 or type4. additionally, the first square (nearby pos) is checked for type1 or type2
//used by attack() to determine if a sqaure is attacked.
inline static int checkray(const constBoard * b, int pos, const int dir, const int type1, const int type2, const int type3, const int type4) __attribute__ ((always_inline));
inline static int checkray(const constBoard * b, int pos, const int dir, const int type1, const int type2, const int type3, const int type4) {
    int dst = pos + dirtab[dir];
    int p;
    if ((p = BOARD22(dst)->w)) {
        if (p == type1 << BOARD_SHIFT || p == type2 << BOARD_SHIFT || p == type3 << BOARD_SHIFT || p == type4 << BOARD_SHIFT)
            return 1;
        else
            return 0;
    } else {
        while (!(p = BOARD22(dst += dirtab[dir])->w))
            ;
        if (p == type3 << BOARD_SHIFT || p == type4 << BOARD_SHIFT)
            return 1;
        else
            return 0;
    }
}

//checks if square pos is attacked by col.
//
//function template.
#define ATTACK_FUNC(name, col) \
static inline int name(const constBoard * b, int pos)\
{\
    if (BOARD22(pos + kdirtab[0])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    if (BOARD22(pos + kdirtab[1])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    if (BOARD22(pos + kdirtab[2])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    if (BOARD22(pos + kdirtab[3])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    if (BOARD22(pos + kdirtab[4])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    if (BOARD22(pos + kdirtab[5])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    if (BOARD22(pos + kdirtab[6])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    if (BOARD22(pos + kdirtab[7])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    if (checkray(b, pos, 0, col * KING, col * ROOK, col * ROOK, col * QUEEN)) return 1;\
    if (checkray(b, pos, 2, col * KING, col * ROOK, col * ROOK, col * QUEEN)) return 1;\
    if (checkray(b, pos, 4, col * KING, col * ROOK, col * ROOK, col * QUEEN)) return 1;\
    if (checkray(b, pos, 6, col * KING, col * ROOK, col * ROOK, col * QUEEN)) return 1;\
    if (checkray(b, pos, 4+col*3, col * KING, col * PAWN, col * BISHOP, col * QUEEN)) return 1;\
    if (checkray(b, pos, 4+col  , col * KING, col * PAWN, col * BISHOP, col * QUEEN)) return 1;\
    if (checkray(b, pos, 4-col*3, col * KING, col * BISHOP, col * BISHOP, col * QUEEN)) return 1;\
    if (checkray(b, pos, 4-col  , col * KING, col * BISHOP, col * BISHOP, col * QUEEN)) return 1;\
    return 0;\
}

//function definitions.
ATTACK_FUNC(attackw, 1);
ATTACK_FUNC(attackb, -1);

//returns true if color is in check.
int inCheck(const constBoard *b, int color) {
    BEGIN_PROFILE(TIME_INCHECK);
    int ret = ( color>0 ? attackb(b, POS(KING,0)) : attackw(b, POS(-KING,0)));
    END_PROFILE();
    return ret;
}

//same as checkray(), additionally writes attack signatures for the target squares of pieces
//of type1..4. attack signatures are used in the following call of a movegenerator
//for a fast determination if a move will check.
inline static int checkraysig(const constBoard * b, int pos, const int dir, const int type1, const int type2, const int type3, const int type4, int sig) __attribute__ ((always_inline));
inline static int checkraysig(const constBoard * b, int pos, const int dir, const int type1, const int type2, const int type3, const int type4, int sig) {
    int dst = pos + dirtab[dir];
    int p;
    if ((p = BOARD22(dst)->w)) {
        if (p == type1 << BOARD_SHIFT || p == type2 << BOARD_SHIFT || p == type3 << BOARD_SHIFT || p == type4 << BOARD_SHIFT)
            return 1;
        //        PEXT(type1, dst) = sig; type one is always a king
        PEXT(type2, dst) = sig;
        PEXT(type3, dst) = sig;
        PEXT(type4, dst) = sig;
    } else {
        //        PEXT(type1, dst) = sig;
        PEXT(type2, dst) = sig;
        PEXT(type3, dst) = sig;
        PEXT(type4, dst) = sig;
        while (!(p = BOARD22(dst += dirtab[dir])->w)) {
            PEXT(type3, dst) = sig;
            PEXT(type4, dst) = sig;
        }
        if (p == type3 << BOARD_SHIFT || p == type4 << BOARD_SHIFT)
            return 1;
        PEXT(type3, dst) = sig;
        PEXT(type4, dst) = sig;
    }
    return 0;
}

//checks if pos is attacked by col and writes attack signatures for the following call
//of a movegenerator.
//
//function template
#define ATTACKSIG_FUNC(name, col) \
static inline int name(const Board * b, int pos, int sig)\
{\
    if (BOARD22(pos + kdirtab[0])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    PEXT(col*KNIGHT, pos + kdirtab[0]) = sig;\
    if (BOARD22(pos + kdirtab[1])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    PEXT(col*KNIGHT, pos + kdirtab[1]) = sig;\
    if (BOARD22(pos + kdirtab[2])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    PEXT(col*KNIGHT, pos + kdirtab[2]) = sig;\
    if (BOARD22(pos + kdirtab[3])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    PEXT(col*KNIGHT, pos + kdirtab[3]) = sig;\
    if (BOARD22(pos + kdirtab[4])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    PEXT(col*KNIGHT, pos + kdirtab[4]) = sig;\
    if (BOARD22(pos + kdirtab[5])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    PEXT(col*KNIGHT, pos + kdirtab[5]) = sig;\
    if (BOARD22(pos + kdirtab[6])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    PEXT(col*KNIGHT, pos + kdirtab[6]) = sig;\
    if (BOARD22(pos + kdirtab[7])->w == col * KNIGHT << BOARD_SHIFT) return 1;\
    PEXT(col*KNIGHT, pos + kdirtab[7]) = sig;\
    if (checkraysig((constBoard *)b, pos, 0, col * KING, col * ROOK, col * ROOK, col * QUEEN, sig)) return 1;\
    if (checkraysig((constBoard *)b, pos, 2, col * KING, col * ROOK, col * ROOK, col * QUEEN, sig)) return 1;\
    if (checkraysig((constBoard *)b, pos, 4, col * KING, col * ROOK, col * ROOK, col * QUEEN, sig)) return 1;\
    if (checkraysig((constBoard *)b, pos, 6, col * KING, col * ROOK, col * ROOK, col * QUEEN, sig)) return 1;\
    if (checkraysig((constBoard *)b, pos, 4+col*3, col * KING, col * PAWN, col * BISHOP, col * QUEEN, sig)) return 1;\
    if (checkraysig((constBoard *)b, pos, 4+col  , col * KING, col * PAWN, col * BISHOP, col * QUEEN, sig)) return 1;\
    if (checkraysig((constBoard *)b, pos, 4-col*3, col * KING, col * BISHOP, col * BISHOP, col * QUEEN, sig)) return 1;\
    if (checkraysig((constBoard *)b, pos, 4-col  , col * KING, col * BISHOP, col * BISHOP, col * QUEEN, sig)) return 1;\
    return 0;\
}

//function definition
ATTACKSIG_FUNC(attackwsig, 1);
ATTACKSIG_FUNC(attackbsig, -1);

//writes attack signatures for the target squares of pieces
//of type1..4. attack signatures are used in the following call of a movegenerator
//for a fast determination if a move will check.
inline static int writesig(Board * b, int pos, const int dir, const int type1, const int type2, const int type3, const int type4, int sig) __attribute__ ((always_inline));
inline static int writesig(Board * b, int pos, const int dir, const int type1, const int type2, const int type3, const int type4, int sig) {
    int dst = pos + dirtab[dir];
    int p;
    if ((p = BOARD22(dst)->w)) {
        if (p == type1 << BOARD_SHIFT || p == type2 << BOARD_SHIFT || p == type3 << BOARD_SHIFT || p == type4 << BOARD_SHIFT)
            return 1;
        //        PEXT(type1, dst) = sig; type one is always a king
        PEXT(type2, dst) = sig;
        PEXT(type3, dst) = sig;
        PEXT(type4, dst) = sig;
    } else {
        //        PEXT(type1, dst) = sig;
        PEXT(type2, dst) = sig;
        PEXT(type3, dst) = sig;
        PEXT(type4, dst) = sig;
        while (!(p = BOARD22(dst += dirtab[dir])->w)) {
            PEXT(type3, dst) = sig;
            PEXT(type4, dst) = sig;
        }
        if (p == type3 << BOARD_SHIFT || p == type4 << BOARD_SHIFT)
            return 1;
        PEXT(type3, dst) = sig;
        PEXT(type4, dst) = sig;
    }
    return 0;
}

//checks if pos is attacked by col and writes attack signatures for the following call
//of a movegenerator.
//
//function template
#define WRITESIG_FUNC(name, col) \
static inline void name(Board * b, int pos, int sig)\
{\
    PEXT(col*KNIGHT, pos + kdirtab[0]) = sig;\
    PEXT(col*KNIGHT, pos + kdirtab[1]) = sig;\
    PEXT(col*KNIGHT, pos + kdirtab[2]) = sig;\
    PEXT(col*KNIGHT, pos + kdirtab[3]) = sig;\
    PEXT(col*KNIGHT, pos + kdirtab[4]) = sig;\
    PEXT(col*KNIGHT, pos + kdirtab[5]) = sig;\
    PEXT(col*KNIGHT, pos + kdirtab[6]) = sig;\
    PEXT(col*KNIGHT, pos + kdirtab[7]) = sig;\
    writesig(b, pos, 0, col * KING, col * ROOK, col * ROOK, col * QUEEN, sig);\
    writesig(b, pos, 2, col * KING, col * ROOK, col * ROOK, col * QUEEN, sig);\
    writesig(b, pos, 4, col * KING, col * ROOK, col * ROOK, col * QUEEN, sig);\
    writesig(b, pos, 6, col * KING, col * ROOK, col * ROOK, col * QUEEN, sig);\
    writesig(b, pos, 4+col*3, col * KING, col * PAWN, col * BISHOP, col * QUEEN, sig);\
    writesig(b, pos, 4+col  , col * KING, col * PAWN, col * BISHOP, col * QUEEN, sig);\
    writesig(b, pos, 4-col*3, col * KING, col * BISHOP, col * BISHOP, col * QUEEN, sig);\
    writesig(b, pos, 4-col  , col * KING, col * BISHOP, col * BISHOP, col * QUEEN, sig);\
}

//function definition
WRITESIG_FUNC(writewsig, 1);
WRITESIG_FUNC(writebsig, -1);


//
//white movegen functions
//

//movegenerator for capture moves
inline static Move *movegenqw(const constBoard * b, Move * m, float alpha) {
    //macro to determine if a sliding piece attacks another piece
    #define rayqw(p, offset)  {\
        int to = src + offset;\
        int cap;\
        while((cap = BOARD22(to)->w) == 0) to += offset;\
        if (cap < 0) {\
            m->val = srcval + *VALUE6(p, to) - *VALUE6(0, cap + to); \
            m->special = 0; \
            m->src = src; \
            m->dst = to; \
            m++; \
        }\
    }

    //macro for non-sliding pieces
    #define squareqw(p, offset) {\
        int to = src + offset;\
        int cap = BOARD22(to)->w;\
        if (cap < 0) {\
            m->val = srcval + *VALUE6(p, to) - *VALUE6(0, cap + to); \
            m->special = 0; \
            m->src = src; \
            m->dst = to; \
            m++; \
        }\
    }

    for (const int *ppp = &POS(ROOK, CNT(ROOK) - 1); ppp >= &POS(ROOK, 0); ppp--) {     //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(ROOK, src);
        rayqw(ROOK, dirtab[0]);
        rayqw(ROOK, dirtab[2]);
        rayqw(ROOK, dirtab[4]);
        rayqw(ROOK, dirtab[6]);
    }

    //bishop moves
    for (const int *ppp = &POS(BISHOP, CNT(BISHOP) - 1); ppp >= &POS(BISHOP, 0); ppp--) {       //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(BISHOP, src);
        rayqw(BISHOP, dirtab[1]);
        rayqw(BISHOP, dirtab[3]);
        rayqw(BISHOP, dirtab[5]);
        rayqw(BISHOP, dirtab[7]);
    }

    //queen moves
    for (const int *ppp = &POS(QUEEN, CNT(QUEEN) - 1); ppp >= &POS(QUEEN, 0); ppp--) {  //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(QUEEN, src);
        rayqw(QUEEN, dirtab[0]);
        rayqw(QUEEN, dirtab[2]);
        rayqw(QUEEN, dirtab[4]);
        rayqw(QUEEN, dirtab[6]);
        rayqw(QUEEN, dirtab[1]);
        rayqw(QUEEN, dirtab[3]);
        rayqw(QUEEN, dirtab[5]);
        rayqw(QUEEN, dirtab[7]);
    }

    //knight moves
    for (const int *ppp = &POS(KNIGHT, CNT(KNIGHT) - 1); ppp >= &POS(KNIGHT, 0); ppp--) {       //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(KNIGHT, src);
        squareqw(KNIGHT, kdirtab[0]);
        squareqw(KNIGHT, kdirtab[2]);
        squareqw(KNIGHT, kdirtab[4]);
        squareqw(KNIGHT, kdirtab[6]);
        squareqw(KNIGHT, kdirtab[1]);
        squareqw(KNIGHT, kdirtab[3]);
        squareqw(KNIGHT, kdirtab[5]);
        squareqw(KNIGHT, kdirtab[7]);
    }

    //king moves
    {
        int src = POS(KING, 0);
        float srcval = b->currMove->move.val - *VALUE6(KING, src);
        squareqw(KING, dirtab[0]);
        squareqw(KING, dirtab[2]);
        squareqw(KING, dirtab[4]);
        squareqw(KING, dirtab[6]);
        squareqw(KING, dirtab[1]);
        squareqw(KING, dirtab[3]);
        squareqw(KING, dirtab[5]);
        squareqw(KING, dirtab[7]);

    }
    //pawn moves
    for (const int *ppp = &POS(PAWN, CNT(PAWN) - 1); ppp >= &POS(PAWN, 0); ppp--) {     //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(PAWN, src);
        //promotions
        if (src >= 6 * ROW) {
            int to;
            if (BOARD22(to = src + dirtab[1])->w < 0)
                m = insertMovew(b, m, src, to, SPEC_PROMOQ, PAWN, srcval + *VALUE6(QUEEN, to) - *VALUE6(0, BOARD22(to)->w + to));
            if (BOARD22(to = src + dirtab[3])->w < 0)
                m = insertMovew(b, m, src, to, SPEC_PROMOQ, PAWN, srcval + *VALUE6(QUEEN, to) - *VALUE6(0, BOARD22(to)->w + to));
            if (BOARD22(to = src + dirtab[2])->w == 0)
                m = insertMovew(b, m, src, to, SPEC_PROMOQ, PAWN, srcval + *VALUE6(QUEEN, to));
        } else {
            int to;
            if (BOARD22(to = src + dirtab[1])->w < 0)
                m = insertMovew(b, m, src, to, 0, PAWN, srcval + *VALUE6(PAWN, to) - *VALUE6(0, BOARD22(to)->w + to));
            if (BOARD22(to = src + dirtab[3])->w < 0)
                m = insertMovew(b, m, src, to, 0, PAWN, srcval + *VALUE6(PAWN, to) - *VALUE6(0, BOARD22(to)->w + to));
        }
    }

    return m;
}

inline static Move *movegenqw1(const constBoard * b, Move * m, float alpha) {
    #define rayqw1(p, offset)  {\
        int to = src + offset;\
        int cap;\
        while((cap = BOARD22(to)->w) == 0) {\
            if (PEXT(p, to) == b->wsig) {\
                m->val = srcval + *VALUE6(p, to); \
                m->special = 0; \
                m->attack = 1;\
                m->src = src; \
                m->dst = to; \
                m++; \
            }\
            to += offset;\
        }\
        if (cap < 0) {\
            m->val = srcval + *VALUE6(p, to) - *VALUE6(0, cap + to); \
            m->special = 0; \
            m->attack = PEXT(p, to) == b->wsig;\
            m->src = src; \
            m->dst = to; \
            m++; \
        }\
    }

    #define squareqw1(p, offset) {\
        int to = src + offset;\
        int cap = BOARD22(to)->w;\
        if (cap < 0 || (cap == 0 && PEXT(p, to) == b->wsig)) {\
            m->val = srcval + *VALUE6(p, to) - *VALUE6(0, cap + to); \
            m->special = 0; \
            m->attack = PEXT(p, to) == b->bsig;\
            m->src = src; \
            m->dst = to; \
            m++; \
        }\
    }

    for (const int *ppp = &POS(ROOK, CNT(ROOK) - 1); ppp >= &POS(ROOK, 0); ppp--) {     //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(ROOK, src);
        rayqw1(ROOK, dirtab[0]);
        rayqw1(ROOK, dirtab[2]);
        rayqw1(ROOK, dirtab[4]);
        rayqw1(ROOK, dirtab[6]);
    }

    //bishop moves
    for (const int *ppp = &POS(BISHOP, CNT(BISHOP) - 1); ppp >= &POS(BISHOP, 0); ppp--) {       //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(BISHOP, src);
        rayqw1(BISHOP, dirtab[1]);
        rayqw1(BISHOP, dirtab[3]);
        rayqw1(BISHOP, dirtab[5]);
        rayqw1(BISHOP, dirtab[7]);
    }

    //queen moves
    for (const int *ppp = &POS(QUEEN, CNT(QUEEN) - 1); ppp >= &POS(QUEEN, 0); ppp--) {  //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(QUEEN, src);
        rayqw1(QUEEN, dirtab[0]);
        rayqw1(QUEEN, dirtab[2]);
        rayqw1(QUEEN, dirtab[4]);
        rayqw1(QUEEN, dirtab[6]);
        rayqw1(QUEEN, dirtab[1]);
        rayqw1(QUEEN, dirtab[3]);
        rayqw1(QUEEN, dirtab[5]);
        rayqw1(QUEEN, dirtab[7]);
    }

    //knight moves
    for (const int *ppp = &POS(KNIGHT, CNT(KNIGHT) - 1); ppp >= &POS(KNIGHT, 0); ppp--) {       //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(KNIGHT, src);
        squareqw1(KNIGHT, kdirtab[0]);
        squareqw1(KNIGHT, kdirtab[2]);
        squareqw1(KNIGHT, kdirtab[4]);
        squareqw1(KNIGHT, kdirtab[6]);
        squareqw1(KNIGHT, kdirtab[1]);
        squareqw1(KNIGHT, kdirtab[3]);
        squareqw1(KNIGHT, kdirtab[5]);
        squareqw1(KNIGHT, kdirtab[7]);
    }

    //king moves
    {
        int src = POS(KING, 0);
        float srcval = b->currMove->move.val - *VALUE6(KING, src);
        squareqw(KING, dirtab[0]);
        squareqw(KING, dirtab[2]);
        squareqw(KING, dirtab[4]);
        squareqw(KING, dirtab[6]);
        squareqw(KING, dirtab[1]);
        squareqw(KING, dirtab[3]);
        squareqw(KING, dirtab[5]);
        squareqw(KING, dirtab[7]);

    }
    //pawn moves
    for (const int *ppp = &POS(PAWN, CNT(PAWN) - 1); ppp >= &POS(PAWN, 0); ppp--) {     //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(PAWN, src);
        //promotions
        if (src >= 6 * ROW) {
            int to;
            if (BOARD22(to = src + dirtab[1])->w < 0)
                m = insertMovew(b, m, src, to, SPEC_PROMOQ, PAWN, srcval + *VALUE6(QUEEN, to) - *VALUE6(0, BOARD22(to)->w + to));
            if (BOARD22(to = src + dirtab[3])->w < 0)
                m = insertMovew(b, m, src, to, SPEC_PROMOQ, PAWN, srcval + *VALUE6(QUEEN, to) - *VALUE6(0, BOARD22(to)->w + to));
            if (BOARD22(to = src + dirtab[2])->w == 0)
                m = insertMovew(b, m, src, to, SPEC_PROMOQ, PAWN, srcval + *VALUE6(QUEEN, to));
        } else {
            int to;
            if (BOARD22(to = src + dirtab[1])->w < 0)
                m = insertMovew(b, m, src, to, 0, PAWN, srcval + *VALUE6(PAWN, to) - *VALUE6(0, BOARD22(to)->w + to));
            if (BOARD22(to = src + dirtab[3])->w < 0)
                m = insertMovew(b, m, src, to, 0, PAWN, srcval + *VALUE6(PAWN, to) - *VALUE6(0, BOARD22(to)->w + to));
        }
    }

    return m;
}

//checks if a step from square src to square src+offset is possible and in this case
//writes a move to the move list. calculates pre-value of move and sets attack flag
//if appropriate
//TODO: use simpler version (without wsig) for king moves.
inline static Move *squarew(const constBoard * b, Move * m, int src, const int piece, const int offset, const int special, float srcval) __attribute__ ((always_inline));
inline static Move *squarew(const constBoard * b, Move * m, int src, const int piece, const int offset, const int special, float srcval) {
    int dst = src + offset;
    int cap = BOARD22(dst)->w;
    if (cap <= 0) {
        m->val = srcval + *VALUE6(piece, dst) - *VALUE6(0, cap + dst);
        m->special = special;
        m->attack = PEXT(piece, dst) == b->wsig;
        m->piece = piece;
        m->src = src;
        m->dst = dst;
        m++;
    }
    return m;
}

//same as squarew, but for sliding pieces
inline static Move *rayw(const constBoard * b, Move * m, int src, const int piece, const int offset, const int special, float srcval) __attribute__ ((always_inline));
inline static Move *rayw(const constBoard * b, Move * m, int src, const int piece, const int offset, const int special, float srcval) {
    int dst = src + offset;
    int cap;
    while ((cap = BOARD22(dst)->w) == 0) {
        m->val = srcval + *VALUE6(piece, dst);
        m->special = special;
        m->attack = PEXT(piece, dst) == b->wsig;
        m->piece = piece;
        m->src = src;
        m->dst = dst;
        m++;
        dst += offset;
    }
    if (cap < 0) {
        m->val = srcval + *VALUE6(piece, dst) - *VALUE6(0, cap + dst);
        m->special = special;
        m->attack = PEXT(piece, dst) == b->wsig;
        m->piece = piece;
        m->src = src;
        m->dst = dst;
        m++;
    }
    return m;
}


//generates all possible (and probably some illegal) moves for position b and writes
//them to the move list m.
inline static Move *movegenw(const constBoard *b, Move *m) {
    //rook moves. if castling is still possible, set SPEC_ROOK flag. in this case,
    //when the move is executed later, the appropriate castling flag will be reset.
    for (const int *ppp = &POS(ROOK, CNT(ROOK) - 1); ppp >= &POS(ROOK, 0); ppp--) {     //piece position pointer
        int src = *ppp;
        int spec = b->o_o & (CAST_WK | CAST_WQ) ? SPEC_ROOK : 0;
        float srcval = b->currMove->move.val - *VALUE6(ROOK, src);
        m = rayw(b, m, src, ROOK, dirtab[0], spec, srcval);
        m = rayw(b, m, src, ROOK, dirtab[2], spec, srcval);
        m = rayw(b, m, src, ROOK, dirtab[4], spec, srcval);
        m = rayw(b, m, src, ROOK, dirtab[6], spec, srcval);
    }

    //bishop moves
    for (const int *ppp = &POS(BISHOP, CNT(BISHOP) - 1); ppp >= &POS(BISHOP, 0); ppp--) {       //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(BISHOP, src);
        m = rayw(b, m, src, BISHOP, dirtab[1], 0, srcval);
        m = rayw(b, m, src, BISHOP, dirtab[3], 0, srcval);
        m = rayw(b, m, src, BISHOP, dirtab[5], 0, srcval);
        m = rayw(b, m, src, BISHOP, dirtab[7], 0, srcval);
    }

    //queen moves
    for (const int *ppp = &POS(QUEEN, CNT(QUEEN) - 1); ppp >= &POS(QUEEN, 0); ppp--) {  //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(QUEEN, src);
        m = rayw(b, m, src, QUEEN, dirtab[0], 0, srcval);
        m = rayw(b, m, src, QUEEN, dirtab[2], 0, srcval);
        m = rayw(b, m, src, QUEEN, dirtab[4], 0, srcval);
        m = rayw(b, m, src, QUEEN, dirtab[6], 0, srcval);
        m = rayw(b, m, src, QUEEN, dirtab[1], 0, srcval);
        m = rayw(b, m, src, QUEEN, dirtab[3], 0, srcval);
        m = rayw(b, m, src, QUEEN, dirtab[5], 0, srcval);
        m = rayw(b, m, src, QUEEN, dirtab[7], 0, srcval);
    }

    //knight moves
    for (const int *ppp = &POS(KNIGHT, CNT(KNIGHT) - 1); ppp >= &POS(KNIGHT, 0); ppp--) {       //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(KNIGHT, src);
        m = squarew(b, m, src, KNIGHT, kdirtab[0], 0, srcval);
        m = squarew(b, m, src, KNIGHT, kdirtab[2], 0, srcval);
        m = squarew(b, m, src, KNIGHT, kdirtab[4], 0, srcval);
        m = squarew(b, m, src, KNIGHT, kdirtab[6], 0, srcval);
        m = squarew(b, m, src, KNIGHT, kdirtab[1], 0, srcval);
        m = squarew(b, m, src, KNIGHT, kdirtab[3], 0, srcval);
        m = squarew(b, m, src, KNIGHT, kdirtab[5], 0, srcval);
        m = squarew(b, m, src, KNIGHT, kdirtab[7], 0, srcval);
    }

    //king moves
    {
        int src = POS(KING, 0);
        int spec = b->o_o & (CAST_WK | CAST_WQ) ? SPEC_KING : 0;
        float srcval = b->currMove->move.val - *VALUE6(KING, src);
        m = squarew(b, m, src, KING, dirtab[0], spec, srcval);
        m = squarew(b, m, src, KING, dirtab[2], spec, srcval);
        m = squarew(b, m, src, KING, dirtab[4], spec, srcval);
        m = squarew(b, m, src, KING, dirtab[6], spec, srcval);
        m = squarew(b, m, src, KING, dirtab[1], spec, srcval);
        m = squarew(b, m, src, KING, dirtab[3], spec, srcval);
        m = squarew(b, m, src, KING, dirtab[5], spec, srcval);
        m = squarew(b, m, src, KING, dirtab[7], spec, srcval);

        //castling
        if (b->o_o & CAST_WQ && BOARD22(D1)->all == 0 && BOARD22(C1)->all == 0 && BOARD22(B1)->all == 0 && BOARD22(A1)->w == ROOK << BOARD_SHIFT
                && attackb(b, E1) == 0 && attackb(b, D1) == 0 && attackb(b, C1) == 0) {
            m = insertMovew(b, m, E1, C1, SPEC_CASTLEQ, KING, srcval + *VALUE6(KING, C1) - *VALUE6(ROOK, A1) + *VALUE6(ROOK, D1));
        }
        if (b->o_o & CAST_WK && BOARD22(F1)->all == 0 && BOARD22(G1)->all == 0 && BOARD22(H1)->w == ROOK << BOARD_SHIFT
                && attackb(b, E1) == 0 && attackb(b, F1) == 0 && attackb(b, G1) == 0)
            m = insertMovew(b, m, E1, G1, SPEC_CASTLEK, KING, srcval + *VALUE6(KING, G1) - *VALUE6(ROOK, H1) + *VALUE6(ROOK, F1));
    }
    //pawn moves
    for (const int *ppp = &POS(PAWN, CNT(PAWN) - 1); ppp >= &POS(PAWN, 0); ppp--) {     //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(PAWN, src);
        //promotions
        if (src >= 6 * ROW) {
            int dst;
            if (BOARD22(dst = src + dirtab[1])->w < 0) {
                m = insertMovew(b, m, src, dst, SPEC_PROMOQ, PAWN, srcval + *VALUE6(QUEEN, dst) - *VALUE6(0, BOARD22(dst)->w + dst));
                m = insertMovew(b, m, src, dst, SPEC_PROMOR, PAWN, srcval + *VALUE6(ROOK, dst) - *VALUE6(0, BOARD22(dst)->w + dst));
                m = insertMovew(b, m, src, dst, SPEC_PROMON, PAWN, srcval + *VALUE6(KNIGHT, dst) - *VALUE6(0, BOARD22(dst)->w + dst));
                m = insertMovew(b, m, src, dst, SPEC_PROMOB, PAWN, srcval + *VALUE6(BISHOP, dst) - *VALUE6(0, BOARD22(dst)->w + dst));
            }
            if (BOARD22(dst = src + dirtab[3])->w < 0) {
                m = insertMovew(b, m, src, dst, SPEC_PROMOQ, PAWN, srcval + *VALUE6(QUEEN, dst) - *VALUE6(0, BOARD22(dst)->w + dst));
                m = insertMovew(b, m, src, dst, SPEC_PROMOR, PAWN, srcval + *VALUE6(ROOK, dst) - *VALUE6(0, BOARD22(dst)->w + dst));
                m = insertMovew(b, m, src, dst, SPEC_PROMON, PAWN, srcval + *VALUE6(KNIGHT, dst) - *VALUE6(0, BOARD22(dst)->w + dst));
                m = insertMovew(b, m, src, dst, SPEC_PROMOB, PAWN, srcval + *VALUE6(BISHOP, dst) - *VALUE6(0, BOARD22(dst)->w + dst));
            }
            if (BOARD22(dst = src + dirtab[2])->w == 0) {
                m = insertMovew(b, m, src, dst, SPEC_PROMOQ, PAWN, srcval + *VALUE6(QUEEN, dst));
                m = insertMovew(b, m, src, dst, SPEC_PROMOR, PAWN, srcval + *VALUE6(ROOK, dst));
                m = insertMovew(b, m, src, dst, SPEC_PROMON, PAWN, srcval + *VALUE6(KNIGHT, dst));
                m = insertMovew(b, m, src, dst, SPEC_PROMOB, PAWN, srcval + *VALUE6(BISHOP, dst));
            }
        } else {
            //single and double step
            int dst;
            if (BOARD22(dst = src + dirtab[2])->all == 0) {
                m = insertMovew(b, m, src, dst, 0, PAWN, srcval + *VALUE6(PAWN, dst));
                if (BOARD22(dst = src + 2 * dirtab[2])->all == 0 && src < 2 * ROW)
                    m = insertMovew(b, m, src, dst, SPEC_PAWNDBL, PAWN, srcval + *VALUE6(PAWN, dst));
            }
            if (BOARD22(dst = src + dirtab[1])->w < 0)
                m = insertMovew(b, m, src, dst, 0, PAWN, srcval + *VALUE6(PAWN, dst) - *VALUE6(0, BOARD22(dst)->w + dst));
            if (BOARD22(dst = src + dirtab[3])->w < 0)
                m = insertMovew(b, m, src, dst, 0, PAWN, srcval + *VALUE6(PAWN, dst) - *VALUE6(0, BOARD22(dst)->w + dst));
        }

        //en passant capture
        if (src >= 4 * ROW && (src + dirtab[0] == b->epstatus || src + dirtab[4] == b->epstatus))
            m = insertMovew(b, m, src, b->epstatus + dirtab[2], SPEC_PAWNEP, PAWN, srcval + *VALUE6(PAWN, b->epstatus + ROW) - *VALUE6(-PAWN, b->epstatus));
    }
    return m;
}

//
//black movegen functions
//

inline static Move *squareb(const constBoard * b, Move * m, int src, const int piece, const int offset, const int special, float srcval) __attribute__ ((always_inline));
inline static Move *squareb(const constBoard * b, Move * m, int src, const int piece, const int offset, const int special, float srcval) {
    int dst = src + offset;
    int cap = BOARD22(dst)->b;
    if (cap >= 0) {
        m->val = srcval + *VALUE6(piece, dst) - *VALUE6(0, cap + dst);
        m->attack = PEXT(piece, dst) == b->bsig;
        m->special = special;
        m->piece = piece;
        m->src = src;
        m->dst = dst;
        m++;
    }
    return m;
}
inline static Move *rayb(const constBoard * b, Move * m, int src, const int piece, const int offset, const int special, float srcval) __attribute__ ((always_inline));
inline static Move *rayb(const constBoard * b, Move * m, int src, const int piece, const int offset, const int special, float srcval) {
    int dst = src + offset;
    int cap;
    while ((cap = BOARD22(dst)->b) == 0) {
        m->val = srcval + *VALUE6(piece, dst);
        m->attack = PEXT(piece, dst) == b->bsig;
        m->special = special;
        m->piece = piece;
        m->src = src;
        m->dst = dst;
        m++;
        dst += offset;
    }
    if (cap > 0) {
        m->val = srcval + *VALUE6(piece, dst) - *VALUE6(0, cap + dst);
        m->attack = PEXT(piece, dst) == b->bsig;
        m->special = special;
        m->piece = piece;
        m->src = src;
        m->dst = dst;
        m++;
    }
    return m;
}

//TODO: make movegenb and movegenqb values negative
inline static Move *movegenb(const constBoard * b, Move * m) {
    //rook moves. if castling is still possible, set SPEC_ROOK flag. in this case,
    //when the move is executed later, the appropriate castling flag will be reset.
    for (const int *ppp = &POS(-ROOK, CNT(-ROOK) - 1); ppp >= &POS(-ROOK, 0); ppp--) {  //piece position pointer
        int spec = b->o_o & (CAST_BK | CAST_BQ) ? SPEC_ROOK : 0;
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(-ROOK, src);
        m = rayb(b, m, src, -ROOK, dirtab[0], spec, srcval);
        m = rayb(b, m, src, -ROOK, dirtab[2], spec, srcval);
        m = rayb(b, m, src, -ROOK, dirtab[4], spec, srcval);
        m = rayb(b, m, src, -ROOK, dirtab[6], spec, srcval);
    }

    for (const int *ppp = &POS(-BISHOP, CNT(-BISHOP) - 1); ppp >= &POS(-BISHOP, 0); ppp--) {    //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(-BISHOP, src);
        m = rayb(b, m, src, -BISHOP, dirtab[1], 0, srcval);
        m = rayb(b, m, src, -BISHOP, dirtab[3], 0, srcval);
        m = rayb(b, m, src, -BISHOP, dirtab[5], 0, srcval);
        m = rayb(b, m, src, -BISHOP, dirtab[7], 0, srcval);
    }

    for (const int *ppp = &POS(-QUEEN, CNT(-QUEEN) - 1); ppp >= &POS(-QUEEN, 0); ppp--) {       //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(-QUEEN, src);
        m = rayb(b, m, src, -QUEEN, dirtab[0], 0, srcval);
        m = rayb(b, m, src, -QUEEN, dirtab[2], 0, srcval);
        m = rayb(b, m, src, -QUEEN, dirtab[4], 0, srcval);
        m = rayb(b, m, src, -QUEEN, dirtab[6], 0, srcval);
        m = rayb(b, m, src, -QUEEN, dirtab[1], 0, srcval);
        m = rayb(b, m, src, -QUEEN, dirtab[3], 0, srcval);
        m = rayb(b, m, src, -QUEEN, dirtab[5], 0, srcval);
        m = rayb(b, m, src, -QUEEN, dirtab[7], 0, srcval);
    }

    for (const int *ppp = &POS(-KNIGHT, CNT(-KNIGHT) - 1); ppp >= &POS(-KNIGHT, 0); ppp--) {    //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(-KNIGHT, src);
        m = squareb(b, m, src, -KNIGHT, kdirtab[0], 0, srcval);
        m = squareb(b, m, src, -KNIGHT, kdirtab[2], 0, srcval);
        m = squareb(b, m, src, -KNIGHT, kdirtab[4], 0, srcval);
        m = squareb(b, m, src, -KNIGHT, kdirtab[6], 0, srcval);
        m = squareb(b, m, src, -KNIGHT, kdirtab[1], 0, srcval);
        m = squareb(b, m, src, -KNIGHT, kdirtab[3], 0, srcval);
        m = squareb(b, m, src, -KNIGHT, kdirtab[5], 0, srcval);
        m = squareb(b, m, src, -KNIGHT, kdirtab[7], 0, srcval);
    }

    //king moves. if castling is still possible, set SPEC_KING flag. in this case,
    //when the move is executed later, the castling flags will be reset.
    {
        int spec = b->o_o & (CAST_BK | CAST_BQ) ? SPEC_KING : 0;
        int src = POS(-KING, 0);
        float srcval = b->currMove->move.val - *VALUE6(-KING, src);
        m = squareb(b, m, src, -KING, dirtab[0], spec, srcval);
        m = squareb(b, m, src, -KING, dirtab[2], spec, srcval);
        m = squareb(b, m, src, -KING, dirtab[4], spec, srcval);
        m = squareb(b, m, src, -KING, dirtab[6], spec, srcval);
        m = squareb(b, m, src, -KING, dirtab[1], spec, srcval);
        m = squareb(b, m, src, -KING, dirtab[3], spec, srcval);
        m = squareb(b, m, src, -KING, dirtab[5], spec, srcval);
        m = squareb(b, m, src, -KING, dirtab[7], spec, srcval);

        //castling
        if (b->o_o & CAST_BQ && BOARD22(D8)->all == 0 && BOARD22(C8)->all == 0 && BOARD22(B8)->all == 0 && BOARD22(A8)->w == -ROOK << BOARD_SHIFT
                && attackw(b, E8) == 0 && attackw(b, D8) == 0 && attackw(b, C8) == 0)
            m = insertMoveb(b, m, E8, C8, SPEC_CASTLEQ, -KING, srcval + *VALUE6(-KING, C8) - *VALUE6(-ROOK, A8) + *VALUE6(-ROOK, D8));
        if (b->o_o & CAST_BK && BOARD22(F8)->all == 0 && BOARD22(G8)->all == 0 && BOARD22(H8)->w == -ROOK << BOARD_SHIFT
                && attackw(b, E8) == 0 && attackw(b, F8) == 0 && attackw(b, G8) == 0)
            m = insertMoveb(b, m, E8, G8, SPEC_CASTLEK, -KING, srcval + *VALUE6(-KING, G8) - *VALUE6(-ROOK, H8) + *VALUE6(-ROOK, F8));
    }

    for (const int *ppp = &POS(-PAWN, CNT(-PAWN) - 1); ppp >= &POS(-PAWN, 0); ppp--) {  //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(-PAWN, src);
        if (src < 2 * ROW) {
            int to;
            if (BOARD22(to = src + dirtab[5])->b > 0) {
                m = insertMoveb(b, m, src, to, SPEC_PROMOQ, -PAWN, srcval + *VALUE6(-QUEEN, to) - *VALUE6(0, BOARD22(to)->w + to));
                m = insertMoveb(b, m, src, to, SPEC_PROMOR, -PAWN, srcval + *VALUE6(-ROOK, to) - *VALUE6(0, BOARD22(to)->w + to));
                m = insertMoveb(b, m, src, to, SPEC_PROMON, -PAWN, srcval + *VALUE6(-KNIGHT, to) - *VALUE6(0, BOARD22(to)->w + to));
                m = insertMoveb(b, m, src, to, SPEC_PROMOB, -PAWN, srcval + *VALUE6(-BISHOP, to) - *VALUE6(0, BOARD22(to)->w + to));
            }
            if (BOARD22(to = src + dirtab[7])->b > 0) {
                m = insertMoveb(b, m, src, to, SPEC_PROMOQ, -PAWN, srcval + *VALUE6(-QUEEN, to) - *VALUE6(0, BOARD22(to)->w + to));
                m = insertMoveb(b, m, src, to, SPEC_PROMOR, -PAWN, srcval + *VALUE6(-ROOK, to) - *VALUE6(0, BOARD22(to)->w + to));
                m = insertMoveb(b, m, src, to, SPEC_PROMON, -PAWN, srcval + *VALUE6(-KNIGHT, to) - *VALUE6(0, BOARD22(to)->w + to));
                m = insertMoveb(b, m, src, to, SPEC_PROMOB, -PAWN, srcval + *VALUE6(-BISHOP, to) - *VALUE6(0, BOARD22(to)->w + to));
            }
            if (BOARD22(to = src + dirtab[6])->b == 0) {
                m = insertMoveb(b, m, src, to, SPEC_PROMOQ, -PAWN, srcval + *VALUE6(-QUEEN, to));
                m = insertMoveb(b, m, src, to, SPEC_PROMOR, -PAWN, srcval + *VALUE6(-ROOK, to));
                m = insertMoveb(b, m, src, to, SPEC_PROMON, -PAWN, srcval + *VALUE6(-KNIGHT, to));
                m = insertMoveb(b, m, src, to, SPEC_PROMOB, -PAWN, srcval + *VALUE6(-BISHOP, to));
            }
        } else {
            int to;
            if (BOARD22(to = src + dirtab[6])->all == 0) {
                m = insertMoveb(b, m, src, to, 0, -PAWN, srcval + *VALUE6(-PAWN, to));

                //pawn double step. set the SPEC_PAWNDBL flag. when the move is
                //executed later, the en passant status will be updated.
                if (BOARD22(to = src + 2 * dirtab[6])->all == 0 && src >= 6 * ROW)
                    m = insertMoveb(b, m, src, to, SPEC_PAWNDBL, -PAWN, srcval + *VALUE6(-PAWN, to));
            }
            if (BOARD22(to = src + dirtab[5])->b > 0)
                m = insertMoveb(b, m, src, to, 0, -PAWN, srcval + *VALUE6(-PAWN, to) - *VALUE6(0, BOARD22(to)->b + to));
            if (BOARD22(to = src + dirtab[7])->b > 0)
                m = insertMoveb(b, m, src, to, 0, -PAWN, srcval + *VALUE6(-PAWN, to) - *VALUE6(0, BOARD22(to)->b + to));
        }

        //en passant capture
        if (src < 4 * ROW && (src + dirtab[0] == b->epstatus || src + dirtab[4] == b->epstatus))
            m = insertMoveb(b, m, src, b->epstatus + dirtab[6], SPEC_PAWNEP, -PAWN, srcval + *VALUE6(-PAWN, b->epstatus - ROW) - *VALUE6(-PAWN, b->epstatus));
    }
    return m;
}

#define rayqb(p, offset)  {\
    int to = src + offset;\
    int cap;\
    while((cap = BOARD22(to)->b) == 0) to += offset;\
    if (cap > 0) {\
        m->val = srcval + *VALUE6(p, to) - *VALUE6(0, cap + to); \
        m->special = 0; \
        m->src = src; \
        m->dst = to; \
        m++; \
    }\
}

#define squareqb(p, offset) {\
    int to = src + offset;\
    int cap = BOARD22(to)->b;\
    if (cap > 0) {\
        m->val = srcval + *VALUE6(p, to) - *VALUE6(0, cap + to); \
        m->special = 0; \
        m->src = src; \
        m->dst = to; \
        m++; \
    }\
}

inline static Move *movegenqb(const constBoard * b, Move * m, float alpha) {
    for (const int *ppp = &POS(-ROOK, CNT(-ROOK) - 1); ppp >= &POS(-ROOK, 0); ppp--) {  //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(-ROOK, src);
        rayqb(-ROOK, dirtab[0]);
        rayqb(-ROOK, dirtab[2]);
        rayqb(-ROOK, dirtab[4]);
        rayqb(-ROOK, dirtab[6]);
    }

    //bishop moves
    for (const int *ppp = &POS(-BISHOP, CNT(-BISHOP) - 1); ppp >= &POS(-BISHOP, 0); ppp--) {    //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(-BISHOP, src);
        rayqb(-BISHOP, dirtab[1]);
        rayqb(-BISHOP, dirtab[3]);
        rayqb(-BISHOP, dirtab[5]);
        rayqb(-BISHOP, dirtab[7]);
    }

    //queen moves
    for (const int *ppp = &POS(-QUEEN, CNT(-QUEEN) - 1); ppp >= &POS(-QUEEN, 0); ppp--) {       //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(-QUEEN, src);
        rayqb(-QUEEN, dirtab[0]);
        rayqb(-QUEEN, dirtab[2]);
        rayqb(-QUEEN, dirtab[4]);
        rayqb(-QUEEN, dirtab[6]);
        rayqb(-QUEEN, dirtab[1]);
        rayqb(-QUEEN, dirtab[3]);
        rayqb(-QUEEN, dirtab[5]);
        rayqb(-QUEEN, dirtab[7]);
    }

    //knight moves
    for (const int *ppp = &POS(-KNIGHT, CNT(-KNIGHT) - 1); ppp >= &POS(-KNIGHT, 0); ppp--) {    //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(-KNIGHT, src);
        squareqb(-KNIGHT, kdirtab[0]);
        squareqb(-KNIGHT, kdirtab[2]);
        squareqb(-KNIGHT, kdirtab[4]);
        squareqb(-KNIGHT, kdirtab[6]);
        squareqb(-KNIGHT, kdirtab[1]);
        squareqb(-KNIGHT, kdirtab[3]);
        squareqb(-KNIGHT, kdirtab[5]);
        squareqb(-KNIGHT, kdirtab[7]);
    }

    //king moves
    {
        int src = POS(-KING, 0);
        float srcval = b->currMove->move.val - *VALUE6(-KING, src);
        squareqb(-KING, dirtab[0]);
        squareqb(-KING, dirtab[2]);
        squareqb(-KING, dirtab[4]);
        squareqb(-KING, dirtab[6]);
        squareqb(-KING, dirtab[1]);
        squareqb(-KING, dirtab[3]);
        squareqb(-KING, dirtab[5]);
        squareqb(-KING, dirtab[7]);

    }
    //pawn moves
    for (const int *ppp = &POS(-PAWN, CNT(-PAWN) - 1); ppp >= &POS(-PAWN, 0); ppp--) {  //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(-PAWN, src);
        //promotions
        if (src < 2 * ROW) {
            int to;
            if (BOARD22(to = src + dirtab[5])->b > 0)
                m = insertMoveb(b, m, src, to, SPEC_PROMOQ, -PAWN, srcval + *VALUE6(-QUEEN, to) - *VALUE6(0, BOARD22(to)->w + to));
            if (BOARD22(to = src + dirtab[7])->b > 0)
                m = insertMoveb(b, m, src, to, SPEC_PROMOQ, -PAWN, srcval + *VALUE6(-QUEEN, to) - *VALUE6(0, BOARD22(to)->w + to));
            if (BOARD22(to = src + dirtab[6])->b == 0)
                m = insertMoveb(b, m, src, to, SPEC_PROMOQ, -PAWN, srcval + *VALUE6(-QUEEN, to));
        } else {
            int to;
            if (BOARD22(to = src + dirtab[5])->b > 0)
                m = insertMoveb(b, m, src, to, 0, -PAWN, srcval + *VALUE6(-PAWN, to) - *VALUE6(0, BOARD22(to)->w + to));
            if (BOARD22(to = src + dirtab[7])->b > 0)
                m = insertMoveb(b, m, src, to, 0, -PAWN, srcval + *VALUE6(-PAWN, to) - *VALUE6(0, BOARD22(to)->w + to));
        }
    }
    return m;
}

#define rayqb1(p, offset)  {\
    int to = src + offset;\
    int cap;\
    while((cap = BOARD22(to)->b) == 0) {\
        if (PEXT(p, to) == b->bsig) {\
            m->val = srcval + *VALUE6(p, to); \
            m->special = 0; \
            m->attack = 1;\
            m->src = src; \
            m->dst = to; \
            m++; \
        }\
        to += offset;\
    }\
    if (cap > 0) {\
        m->val = srcval + *VALUE6(p, to) - *VALUE6(0, cap + to); \
        m->special = 0; \
        m->attack = PEXT(p, to) == b->bsig;\
        m->src = src; \
        m->dst = to; \
        m++; \
    }\
}
 
#define squareqb1(p, offset) {\
    int to = src + offset;\
    int cap = BOARD22(to)->b;\
    if (cap > 0 || (cap == 0 && PEXT(p, to) == b->bsig)) {\
        m->val = srcval + *VALUE6(p, to) - *VALUE6(0, cap + to); \
        m->special = 0; \
        m->attack = PEXT(p, to) == b->bsig;\
        m->src = src; \
        m->dst = to; \
        m++; \
    }\
}

inline static Move *movegenqb1(const constBoard * b, Move * m, float alpha) {
    for (const int *ppp = &POS(-ROOK, CNT(-ROOK) - 1); ppp >= &POS(-ROOK, 0); ppp--) {  //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(-ROOK, src);
        rayqb1(-ROOK, dirtab[0]);
        rayqb1(-ROOK, dirtab[2]);
        rayqb1(-ROOK, dirtab[4]);
        rayqb1(-ROOK, dirtab[6]);
    }

    //bishop moves
    for (const int *ppp = &POS(-BISHOP, CNT(-BISHOP) - 1); ppp >= &POS(-BISHOP, 0); ppp--) {    //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(-BISHOP, src);
        rayqb1(-BISHOP, dirtab[1]);
        rayqb1(-BISHOP, dirtab[3]);
        rayqb1(-BISHOP, dirtab[5]);
        rayqb1(-BISHOP, dirtab[7]);
    }

    //queen moves
    for (const int *ppp = &POS(-QUEEN, CNT(-QUEEN) - 1); ppp >= &POS(-QUEEN, 0); ppp--) {       //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(-QUEEN, src);
        rayqb1(-QUEEN, dirtab[0]);
        rayqb1(-QUEEN, dirtab[2]);
        rayqb1(-QUEEN, dirtab[4]);
        rayqb1(-QUEEN, dirtab[6]);
        rayqb1(-QUEEN, dirtab[1]);
        rayqb1(-QUEEN, dirtab[3]);
        rayqb1(-QUEEN, dirtab[5]);
        rayqb1(-QUEEN, dirtab[7]);
    }

    //knight moves
    for (const int *ppp = &POS(-KNIGHT, CNT(-KNIGHT) - 1); ppp >= &POS(-KNIGHT, 0); ppp--) {    //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(-KNIGHT, src);
        squareqb1(-KNIGHT, kdirtab[0]);
        squareqb1(-KNIGHT, kdirtab[2]);
        squareqb1(-KNIGHT, kdirtab[4]);
        squareqb1(-KNIGHT, kdirtab[6]);
        squareqb1(-KNIGHT, kdirtab[1]);
        squareqb1(-KNIGHT, kdirtab[3]);
        squareqb1(-KNIGHT, kdirtab[5]);
        squareqb1(-KNIGHT, kdirtab[7]);
    }

    //king moves
    {
        int src = POS(-KING, 0);
        float srcval = b->currMove->move.val - *VALUE6(-KING, src);
        squareqb(-KING, dirtab[0]);
        squareqb(-KING, dirtab[2]);
        squareqb(-KING, dirtab[4]);
        squareqb(-KING, dirtab[6]);
        squareqb(-KING, dirtab[1]);
        squareqb(-KING, dirtab[3]);
        squareqb(-KING, dirtab[5]);
        squareqb(-KING, dirtab[7]);

    }
    //pawn moves
    for (const int *ppp = &POS(-PAWN, CNT(-PAWN) - 1); ppp >= &POS(-PAWN, 0); ppp--) {  //piece position pointer
        int src = *ppp;
        float srcval = b->currMove->move.val - *VALUE6(-PAWN, src);
        //promotions
        if (src < 2 * ROW) {
            int to;
            if (BOARD22(to = src + dirtab[5])->b > 0)
                m = insertMoveb(b, m, src, to, SPEC_PROMOQ, -PAWN, srcval + *VALUE6(-QUEEN, to) - *VALUE6(0, BOARD22(to)->w + to));
            if (BOARD22(to = src + dirtab[7])->b > 0)
                m = insertMoveb(b, m, src, to, SPEC_PROMOQ, -PAWN, srcval + *VALUE6(-QUEEN, to) - *VALUE6(0, BOARD22(to)->w + to));
            if (BOARD22(to = src + dirtab[6])->b == 0)
                m = insertMoveb(b, m, src, to, SPEC_PROMOQ, -PAWN, srcval + *VALUE6(-QUEEN, to));
        } else {
            int to;
            if (BOARD22(to = src + dirtab[5])->b > 0)
                m = insertMoveb(b, m, src, to, 0, -PAWN, srcval + *VALUE6(-PAWN, to) - *VALUE6(0, BOARD22(to)->w + to));
            if (BOARD22(to = src + dirtab[7])->b > 0)
                m = insertMoveb(b, m, src, to, 0, -PAWN, srcval + *VALUE6(-PAWN, to) - *VALUE6(0, BOARD22(to)->w + to));
        }
    }
    return m;
}

static int inMate(Board *b) {
    Move firstMove[MAXMOVES];
    Move *lastMove = movegen((constBoard *)b, firstMove);
    for (Move *currMove=firstMove; currMove<lastMove; currMove++) {
        doMove(b, currMove);
        int check = inCheck((constBoard *)b, b->color);
        undoMove(b);
        if (!check)
            return 1;
    }
    return 0;
}
