/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "board.h"
#include <math.h>
#include <stdlib.h>

//32*2 relative position neurons (x,y) for each piece = 2048 input neurons
//connnections
//o = output layer
//h = hidden
//i = input
#define CNT_IN (2*12*8*8)
#define CNT_HID (12*8*8)
#define CNT_OUT (12*8*8)
//last index: x
//2nd last:   y
//3rd last:   piece type
//4th last:   color
//5th last (if any): 1 for piece, 0 for possible destination square
float Neuroni[2][2][6][8][8];
float Neuronh[2][6][8][8];
float errorh[2][6][8][8];
float Neurono[2][6][8][8];
float erroro[2][6][8][8];

float NNhi[CNT_HID][CNT_IN];
float NNoh[CNT_OUT][CNT_HID];

void initNN() {
    for (int xh=0; xh<CNT_HID; xh++)
        for (int xi=0; xi<CNT_IN; xi++)
                NNhi[xh][xi] = 1.0/CNT_IN;
                
    for (int xo=0; xo<CNT_OUT; xo++)
        for (int xh=0; xh<CNT_HID; xh++)
            NNoh[xo][xh] = 1.0/CNT_HID;
}

inline static void squareNN(int dst, int piece, int color)  __attribute__ ((always_inline));
inline static void squareNN(int dst, int piece, int color)
{
    if (!(dst & 0x220)) 
        *(float *)((char*)Neuroni[0][color][piece] + dst) += 1.0;
}

inline static void rayNN(const Board * b, int src, int piece, int offset) __attribute__ ((always_inline));
inline static void rayNN(const Board * b, int src, int piece, int offset)
{
    int dst = src + offset;
    while (!(dst & 0x220)) {
        *(float *)((char*)Neuroni[piece] + dst) += 1.0;
         if (BOARD22(dst)->all) break;
         dst += offset;
    }
}

void inputNN(const Board *b) 
{
    for (int xi=0; xi<CNT_IN; xi++) 
        Neuroni[0][0][0][0][xi] = 0.0;
                
    for (int col=-1; col<=1; col+=2) {
        int coli = (col+1)/2;
        for (const int *ppp = &POS(col*ROOK, CNT(col*ROOK) - 1); ppp >= &POS(col*ROOK, 0); ppp--) {     //piece position pointer
            int src = *ppp;
            *(float *)( (char*) &Neuroni[1][coli][ROOK] + src ) += 1.0;
            rayNN(b, src, ROOK, dirtab[0]);
            rayNN(b, src, ROOK, dirtab[2]);
            rayNN(b, src, ROOK, dirtab[4]);
            rayNN(b, src, ROOK, dirtab[6]);
        }

        for (const int *ppp = &POS(col*BISHOP, CNT(col*BISHOP) - 1); ppp >= &POS(col*BISHOP, 0); ppp--) {     //piece position pointer
            int src = *ppp;
            *(float *)( (char*) &Neuroni[1][coli][BISHOP] + src ) += 1.0;
            rayNN(b, src, BISHOP, dirtab[1]);
            rayNN(b, src, BISHOP, dirtab[3]);
            rayNN(b, src, BISHOP, dirtab[5]);
            rayNN(b, src, BISHOP, dirtab[7]);
        }

        for (const int *ppp = &POS(col*QUEEN, CNT(col*QUEEN) - 1); ppp >= &POS(col*QUEEN, 0); ppp--) {     //piece position pointer
            int src = *ppp;
            *(float *)( (char*) &Neuroni[1][coli][QUEEN] + src ) += 1.0;
            rayNN(b, src, QUEEN, dirtab[1]);
            rayNN(b, src, QUEEN, dirtab[3]);
            rayNN(b, src, QUEEN, dirtab[5]);
            rayNN(b, src, QUEEN, dirtab[7]);
            rayNN(b, src, QUEEN, dirtab[0]);
            rayNN(b, src, QUEEN, dirtab[2]);
            rayNN(b, src, QUEEN, dirtab[4]);
            rayNN(b, src, QUEEN, dirtab[6]);
        }

        { 
            const int *ppp = &POS(col*KING, 0);
            int src = *ppp;
            *(float *)( (char*) &Neuroni[1][coli][KING] + src ) += 1.0;
            squareNN(src, KING, dirtab[1]);
            squareNN(src, KING, dirtab[3]);
            squareNN(src, KING, dirtab[5]);
            squareNN(src, KING, dirtab[7]);
            squareNN(src, KING, dirtab[0]);
            squareNN(src, KING, dirtab[2]);
            squareNN(src, KING, dirtab[4]);
            squareNN(src, KING, dirtab[6]);
        }

        for (const int *ppp = &POS(col*KNIGHT, CNT(col*KNIGHT) - 1); ppp >= &POS(col*KNIGHT, 0); ppp--) {     //piece position pointer
            int src = *ppp;
            *(float *)( (char*) &Neuroni[1][coli][KNIGHT] + src ) += 1.0;
            squareNN(src, KNIGHT, kdirtab[1]);
            squareNN(src, KNIGHT, kdirtab[3]);
            squareNN(src, KNIGHT, kdirtab[5]);
            squareNN(src, KNIGHT, kdirtab[7]);
            squareNN(src, KNIGHT, kdirtab[0]);
            squareNN(src, KNIGHT, kdirtab[2]);
            squareNN(src, KNIGHT, kdirtab[4]);
            squareNN(src, KNIGHT, kdirtab[6]);
        }

        for (const int *ppp = &POS(col*PAWN, CNT(col*PAWN) - 1); ppp >= &POS(col*PAWN, 0); ppp--) {     //piece position pointer
            int src = *ppp;
            *(float *)( (char*) &Neuroni[1][coli][PAWN] + src ) += 1.0;
            squareNN(src, PAWN, dirtab[1]);
            squareNN(src, PAWN, dirtab[3]);
        }    
    }
}

float sigma(float fx) {
    if (fx > 3.5) return 1;
    if (fx < -3.5) return 0;
    long double x = fx;
    long double x2 = x*x;
    
    long double s = x;              // x    
    long double iter = 1;

    for (;;) {    
        x *= x2/iter;  // - x^3/(1*3) - x^7/(6*7)
        if (fabs(x) < 1e-8) break;
        s -= x/(2*iter+1);
        iter++;

        x *= x2/iter;;  // + x^5/(2*5) + x^9/(24*9)
        if (fabs(x) < 1e-8) break;
        s += x/(2*iter+1);
        iter++;
    }
    return s/sqrt(M_PI) + 0.5;
    printf("\n");
}

void think() {
    
    for (int xh=0; xh<CNT_HID; xh++) {
#ifdef USE_ASM                      
        asm volatile("\
            pxor %%mm0, %%mm0;\
            mov %3, %%eax;\
        1:  prefetch 256(%2);\
            movq (%1), %%mm1;\
            movq (%2), %%mm2;\
            pfmul %%mm1, %%mm2;\
            pfadd %%mm2, %%mm0;\
            movq 8(%1), %%mm3;\
            movq 8(%2), %%mm4;\
            pfmul %%mm3, %%mm4;\
            pfadd %%mm4, %%mm0;\
            movq 16(%1), %%mm1;\
            movq 16(%2), %%mm2;\
            pfmul %%mm1, %%mm2;\
            pfadd %%mm2, %%mm0;\
            movq 24(%1), %%mm3;\
            movq 24(%2), %%mm4;\
            pfmul %%mm3, %%mm4;\
            pfadd %%mm4, %%mm0;\
            movq 32(%1), %%mm1;\
            movq 32(%2), %%mm2;\
            pfmul %%mm1, %%mm2;\
            pfadd %%mm2, %%mm0;\
            movq 40(%1), %%mm3;\
            movq 40(%2), %%mm4;\
            pfmul %%mm3, %%mm4;\
            pfadd %%mm4, %%mm0;\
            movq 48(%1), %%mm1;\
            movq 48(%2), %%mm2;\
            pfmul %%mm1, %%mm2;\
            pfadd %%mm2, %%mm0;\
            movq 56(%1), %%mm3;\
            movq 56(%2), %%mm4;\
            pfmul %%mm3, %%mm4;\
            pfadd %%mm4, %%mm0;\
            add $64, %1;\
            add $64, %2;\
            dec %%eax;\
            jnz 1b;\
            movq %%mm0, (%0)"
            :
            :"r" (&Neuronh[0][0][xh]), "r" (Neuroni), "r" (NNhi), "i" (CNT_IN/16)
            :"%eax");
#else                
        Neuronh[0][0][0][xh] = 0.0;
        for (int xi = 0; xi<CNT_IN; xi++)
            Neuronh[0][0][0][xh] += Neuroni[0][0][0][0][xi] * NNhi[xh][xi];
#endif  
    }              
#ifdef USE_ASM
    asm("emms");                   
#endif
    //transform activation values in hidden layer to output values
    for (int xh=0; xh<CNT_HID; xh++)
        Neuronh[0][0][0][xh] = sigma(Neuronh[0][0][0][xh]);
    
    for (int xo=0; xo<CNT_OUT; xo++) {
        Neurono[0][0][0][xo] = 0.0;
        for (int xh = 0; xh<CNT_HID; xh++)
            Neurono[0][0][0][xo] += Neuronh[0][0][0][xh] * NNoh[xo][xh];
    }      

    //transform act values in output layer into output values
    for (int xh=0; xh<CNT_HID; xh++)
        Neurono[0][0][0][xh] = sigma(Neurono[0][0][0][xh]);

}

        
float backpropagate(const Move *m) {
    float eta = 0.015625;
    //transform output values into error values
    for (int xo = 0; xo<CNT_OUT; xo++) 
        erroro[0][0][0][xo] = -Neurono[0][0][0][xo];
    int col = m->piece>0 ? 1 : 0;
    erroro[col][abs(m->piece)][Y(m->dst)][X(m->dst)] += 1.0;
    
    for (int xh = 0; xh<CNT_HID; xh++) 
         errorh[0][0][0][xh] = 0;
         
    double sum = 0.0;
    for (int xo = 0; xo<CNT_OUT; xo++) { 
        sum += erroro[0][0][0][xo] * erroro[0][0][0][xo];
        erroro[0][0][0][xo] *= exp(-Neurono[0][0][0][xo]*Neurono[0][0][0][xo]);
        
        for (int xh = 0; xh<CNT_HID; xh++) {
            errorh[0][0][0][xo] += NNoh[xo][xh] *  erroro[0][0][0][xo];
            NNoh[xo][xh] +=  erroro[0][0][0][xo] * Neurono[0][0][0][xo] * eta;
        }
    }
    
    for (int xh = 0; xh<CNT_HID; xh++) { 
        sum += errorh[0][0][0][xh] * errorh[0][0][0][xh];
        errorh[0][0][0][xh] *= exp(-Neuronh[0][0][0][xh]*Neuronh[0][0][0][xh]);
        
        for (int xi = 0; xi<CNT_IN; xi++) {
//            errori[0][0][0][xh] += NNhi[xh][xi] *  errorh[0][0][0][xh];
            NNhi[xh][xi] +=  errorh[0][0][0][xh] * Neuronh[0][0][0][xh] * eta;
        }
    }
    return sum;
}

void learn(const Board *b, const Move *m) {
    float epsilon = 1e-3;
    inputNN(b);
    for (;;) {
        think();
        if (backpropagate(m) < epsilon) break;
    }        
}
