/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdbool.h>
#include <malloc.h>
#include "globals.h"
#include "opening.h"
#include "hash.h"
#include "board.h"
#include "main.h"
#include "search.h"
#include "util.h"
#include "movegen_c.h"

static int missBook;
static BookElem *bookBuf = 0;
static int bookNum;
static int bookMinmaxPly = 10;

static int compare(const void *p1, const void *p2) {
    if (*(unsigned long long *) p1 < *(unsigned long long *) p2)
        return -1;
    if (*(unsigned long long *) p1 > *(unsigned long long *) p2)
        return 1;
    return 0;
}

void initBook() {
    FILE *f;
    struct stat st;

    missBook = 0;
    if (bookBuf)
        return;
    f = fopen("openings.bin", "r");
    if (f) {
        stat("openings.bin", &st);
        bookNum = st.st_size / sizeof(BookElem);
        bookBuf = (BookElem *) malloc(st.st_size + 16000);
        fread(bookBuf, st.st_size, 1, f);
        fclose(f);
    } else {
        bookNum = 0;
        bookBuf = (BookElem *) malloc(16000);
    }
}

void addPos() {
    BookElem newpos;
    newpos.key = board.key.main.key ^ board.color;
    newpos.val = 0.0;
    bookBuf[bookNum++] = newpos;
    qsort(bookBuf, bookNum, sizeof(BookElem), compare);
    FILE *f = fopen("newbook.bin", "w");
    fwrite(bookBuf, bookNum, sizeof(BookElem), f);
    fclose(f);
}

void delPos() {
    BookElem *p = bookBuf + bookNum;

    while (--p >= bookBuf)
        if (p->key == (board.key.main.key ^ board.color)) {
            p->key = 0xffffffffffffffffLL;
            qsort(bookBuf, bookNum, sizeof(BookElem), compare);
            bookNum--;
            break;
        }

}

Move *bookMoves(Move * firstMove) {
    Move *currMove, *tempMove;

    //    if (missBook > 10)
    //        return firstMove;
    //    sync();
    static Board *const b = &board;
    Move *lastMove = movegen((constBoard *)b, firstMove);
    for (currMove = firstMove; currMove < lastMove; currMove++) {
        doMove(b, currMove);
        //the list of hashkeys is sorted, so we can use a binary search
        bool found = false;
        int lower = 0;
        int upper = bookNum - 1;
        int median;
        BookElem probe;
        unsigned long long key = board.key.main.key ^ board.color;
        while (upper >= lower) {
            median = (lower + upper) / 2;
            probe = bookBuf[median];
            if (key < probe.key) {
                upper = median - 1;
            } else if (key > probe.key) {
                lower = median + 1;
            } else {
                found = true;
                break;
            }
        }

        if (!found) {
            for (tempMove = currMove + 1; tempMove < lastMove; tempMove++)      //remove move from list
                tempMove[-1] = tempMove[0];
            lastMove--;
            currMove--;
        } else
            currMove->val = probe.val;
        undoMove(b);
    }

    if (lastMove == firstMove)
        missBook++;
    else
        missBook = 0;
    return lastMove;
}

#define mask 0x1fffff
BookElem hashhash[mask + 1];

static float searchBook(BookElem *buf, int *num, int *endpoint) {

    static Board *const b = &board;

    Move *currMove, *lastMove, firstMove[MAXMOVES];
    float bestEval = -INF;

    initBook();
    lastMove = bookMoves(firstMove);

    if (lastMove == firstMove) {
        //            printf("%d\n", glmove-glstart);
        //fflush(stdout);
        mvwprintw(inputWnd, 1, 0, "previous: %d, now %d, endpoints %d\n", bookNum, num, ++endpoint);

        wrefresh(inputWnd);

        searchm(bookMinmaxPly);
        copyBoard();
        printStats(statWnd);
        displayBoard(boardWnd);
        wrefresh(boardWnd);
        wrefresh(statWnd);
        return stats.eval;
    } else {
        for (currMove = firstMove; currMove < lastMove; currMove++) {
            doMove(&board, currMove);
            if (IN_CHECK(board.color)) {
                undoMove(&board);
                continue;
            }
            bool found = false;
            BookElem probe;
            unsigned ix;
            probe.key = (board.key.main.key ^ board.color);
            ix = probe.key & mask;
            while (hashhash[ix].key != 0) {
                if (hashhash[ix].key == probe.key) {
                    if (hashhash[ix].val > bestEval)
                        bestEval = hashhash[ix].val;
                    found = true;
                    break;
                }
                ix = (ix + 1) & mask;
            }
            /*
               BookElem *temp = buf + num;
               while(temp > buf)
               if ((--temp)->key == (key.main.key ^ color)) {
               if (temp->val > bestEval)
               bestEval = temp->val;
               found = true;
               break;
               }
             */
            if (!found) {
                if (!insertRep(&board)) {
                    probe.val = -searchBook(buf, num, endpoint);
                    //probe.key = key.main.key ^ color;
                    buf[(*num)++] = probe;
                    hashhash[ix] = probe;
                    if (*num >= bookNum * 2) {
                        printf("book full\n");
                        exit(4);
                    }
                    if (probe.val > bestEval)
                        bestEval = probe.val;
                    deleteRep(&board);
                } else {
                    if (0.0 > bestEval)
                        bestEval = 0.0;
                }
            }
            undoMove(&board);
        }

        /*
                        sync();
                    lastMove = movegen(firstMove);
                    for (currMove = firstMove; currMove < lastMove; currMove++) {
                        doMove(currMove);
                        if (inCheck()) {
                            undoMove();
                            continue;
                        }
                        currEval = -search(3,bookMinmaxPly, -INF, bestEval, bookMinmaxPly, &bestLine);
                        if (currEval > bestEval) {
                            bestEval = currEval;
                        }
                        undoMove();
                    }
                    */
        return bestEval;
    }
}

void minmaxBook() {
    FILE *of;
    int num = 0;
    int endpoint = 0;
    BookElem *const buf = (BookElem *) malloc(bookNum * sizeof(BookElem) * 2);
//    static Board *const b = &board;


    //    initBoard();
    //   endwin();
    t0 = t1 = 0;
    flags.nobook = true;
    searchBook(buf, &num, &endpoint);
    printf("previous: %d, now %d\n", bookNum, num);
    qsort(buf, num, sizeof(BookElem), compare);
    of = fopen("openings.tmp", "w");
    fwrite(buf, sizeof(BookElem), num, of);
    fclose(of);
    free(buf);
}
