/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <ncurses.h>
#include <math.h>
#include "hash.h"
#include "eval.h"
#include "main.h"
#include "util.h"
#include "search.h"
#include "opening.h"
#include "movegen_c.h"
#include "debug.h"
#include "parse.h"
#include "neuron.h"
#include <signal.h>

#ifdef HAVE_ARGZ_H
#include <argz.h>
#endif

#define ACCEPT_FORCE
#define ACCEPT_COLORS

extern int maxNullMove;

static int points(unsigned long long t) {
    if (t<10 S) return 30;
    if (t<30 S) return 25;
    if (t<90 S) return 20;
    if (t<180 S) return 15;
    if (t<390 S) return 10;
    if (t<600 S) return 5;
    return 0;
}
        
static void bench() {
    static const char *benchPos[] = {
//        "1k6/ppp5/8/3BK3/1Q1NP3/1R6/8/KR6 w - -",
        "rnb1kb1r/1p3ppp/p5q1/4p3/3N4/4BB2/PPPQ1P1P/R3K2R w KQkq -",        // bm O-O-O; id "BS2830.13"; c0 "Nimzowitsch-Alapin, 4/97-40 ChessBase, 1911";
        "r1b1r1k1/ppp2ppp/2nb1q2/8/2B5/1P1Q1N2/P1PP1PPP/R1B2RK1 w - -",     // bm Bb2; id "BS2830.12"; c0 "Aljechin-Rabinowitsch, 4/97-39 Meine besten Partien, 1918";
        "8/k7/3p4/p2P1p2/P2P1P2/8/8/K7 w - - bm Kb1; id VA.01, Study by Dr. Lasker and Reichhel; c0 1. Ka1-b1 Ka7-b7 2. Kb1-c1;        2kr1br1/ppp2p1p/2n1pp2/8/1P1PN1q1/P4N2/2P2PPP/R2QR1K1 w - -",
        "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq -",
//"r1R5/1P6/3nbqrp/PpPPPPP1/8/b1n5/rQq/R3K2R w KQkq b5",
        "rq2r1k1/5pp1/p7/4bNP1/1p2P2P/5Q2/PP4K1/5R1R w - - bm Nxg7; id ""BT2630.01""; c0 ""Short-Ribli Nxg7 gewinnt"";",
        "5r1k/p1q2pp1/1pb4p/n3R1NQ/7P/3B1P2/2P3P1/7K w - - bm Re6; id ""BT2630.03""; c0 ""Kwinones-Mjagmarsuren Skopje 1972 Re6 gewinnt"";",
        "5r1k/1P4pp/3P1p2/4p3/1P5P/3q2P1/Q2b2K1/B3R3 w - - bm Qf7; id ""BT2630.04""; c0 ""BT04 Harding 1978 Qf7 gewinnt"";",
        "6k1/p3q2p/1nr3pB/8/3Q1P2/6P1/PP5P/3R2K1 b - - bm Rd6; id ""BT2630.07""; c0 ""Black forces a draw, Majzeliw ...Rd6 haelt remis"";",
        "2r3k1/1qr1b1p1/p2pPn2/nppPp3/8/1PP1B2P/P1BQ1P2/5KRR w - - bm Rxg7+; id ""BT2630.10""; c0 ""Kasparov-Meph.Dallas Rxg7 gewinnt"";",
        "1br3k1/p4p2/2p1r3/3p1b2/3Bn1p1/1P2P1Pq/P3Q1BP/2R1NRK1 b - - bm Qxh2+; id ""BT2630.11""; c0 ""Karpov-Chandler 1983 ...Qxh2+ gewinnt"";",
        "8/pp3k2/2p1qp2/2P5/5P2/1R2p1rp/PP2R3/4K2Q b - - bm Qe4; id ""BT2630.12""; c0 ""Elite Glasgow-H.Haupt ...Qe4 bester Zug"";",
        "2bq3k/2p4p/p2p4/7P/1nBPPQP1/r1p5/8/1K1R2R1 b - - bm Be6; id ""BT2630.13""; c0 ""Bagirow-Gufeld"";",
        "2b1q3/p7/1p1p2kb/nPpN3p/P1P1P2P/6P1/5R1K/5Q2 w - - bm e5; id ""BT2630.15""; c0 ""Kasparov-Karpov"";",
        "2krr3/pppb1ppp/3b4/3q4/3P3n/2P2N1P/PP2B1P1/R1BQ1RK1 b - - bm Nxg2; id ""BT2630.16""; c0 ""Steinitz-Lasker"";",
        "4r1k1/p1pb1ppp/Qbp1r3/8/1P6/2Pq1B2/R2P1PPP/2B2RK1 b - -",  // bm Qxf3; id "BS2830.01"; c0 "Paulsen-Morphy, 4/97-39 Damenopfer, Neistadt, 1857";
        "7r/2qpkp2/p3p3/6P1/1p2b2r/7P/PPP2QP1/R2N1RK1 b - -",       // bm f5; id "BS2830.02"; c0 "J.Xie-J.Polgar, 4/97-39 Schacholympiade Saloniki, 1988";
        "r1bq1rk1/pp4bp/2np4/2p1p1p1/P1N1P3/1P1P1NP1/1BP1QPKP/1R3R2 b - -", // bm Bh3+; id "BS2830.03"; c0 "Raman-Aaron, 4/97-39 Albin Poetzsch, Spass am Kom, 1977";
        "2r1r3/p3bk1p/1pnqpppB/3n4/3P2Q1/PB3N2/1P3PPP/3RR1K1 w - -",        // bm Rxe6; id "BS2830.05"; c0 "Sturua-Kutirow, 4/97-39 Rochochade Europa, 1976";
        "rn2r1k1/pp3ppp/8/1qNp4/3BnQb1/5N2/PPP2PPP/2KR3R b - -",    // bm Bh5; id "BS2830.08"; c0 "Stoica-Uhlmann, 4/97-39 Franzoesisch-Uhlmann, 1979";
        "r3kb1r/1p1b1p2/p1nppp2/7p/4PP2/qNN5/P1PQB1PP/R4R1K w kq -",        // bm Nb1; id "BS2830.09"; c0 "Spassky,B.-Fischer,R., 4/97-39 WCh Reykjavik, 1972";
        "r3r1k1/pp1bp2p/1n2q1P1/6b1/1B2B3/5Q2/5PPP/1R3RK1 w - -",   // bm Bd2; id "BS2830.10"; c0 "Khalifman-Gawrikow, 4/97-39 Gruenfeld-Iindisch Ador, 1988";
        "r3k2r/pb3pp1/2p1qnnp/1pp1P3/Q1N4B/2PB1P2/P5PP/R4RK1 w kq -",       // bm exf6; id "BS2830.11"; c0 "Tal-Hecht, 4/97-39 Schacholympiade, 1962";
        "r1bqr1k1/pp1n1ppp/5b2/4N1B1/3p3P/8/PPPQ1PP1/2K1RB1R w - -",        // bm Nxf7; id "BS2830.14"; c0 "Timman-Kortschnoi, 4/97-40 CSS, 1991";
        "2r2rk1/1bpR1p2/1pq1pQp1/p3P2p/P1PR3P/5N2/2P2PPK/8 w - -",  // bm Kg3; id "BS2830.15"; c0 "Short-Timman, 4/97-40 CSS, 1991";
        "r1b1qrk1/ppBnppb1/2n4p/1NN1P1p1/3p4/8/PPP1BPPP/R2Q1R1K w - -",     // bm Ne6; id "BS2830.17"; c0 "Bednorz,A.-Spinath P.D., 4/97-40 Fernpartie, 1990";
        "r3k2r/pp1n1ppp/1qpnp3/3bN1PP/3P2Q1/2B1R3/PPP2P2/2KR1B2 w kq -",    // bm Be1; id "BS2830.19"; c0 "Karpov-Rogers, 4/97-40 So schnell siegen WM-K, 1983";
        "r1bqk2r/pppp1Npp/8/2bnP3/8/6K1/PB4PP/RN1Q3R b kq -",       // bm O-O; id "BS2830.20"; c0 "Gofman-Petrow, 4/97-40 Schachmate-Maizeliw,I, 1844";
        "r4r1k/pbnq1ppp/np3b2/3p1N2/5B2/2N3PB/PP3P1P/R2QR1K1 w - -",        // bm Ne4; id "BS2830.21"; c0 "Kasparov-Marjanovic, 4/97-40 So schnell sieg.WM-Kar, 1980";
        "r2qr2k/pbp3pp/1p2Bb2/2p5/2P2P2/3R2P1/PP2Q1NP/5RK1 b - -",  // bm Qxd3; id "BS2830.22"; c0 "Euwe-Keres, 4/97-40 Rochade Europa, 1940";
        "5r2/1p4r1/3kp1b1/1Pp1p2p/2PpP3/q2B1PP1/3Q2K1/1R5R b - -",  // bm Rxf3; id "BS2830.23"; c0 "Palatnik-Krupa, 4/97-40 Spass am Kombinieren, 1984";
        "r5k1/p4n1p/6p1/2qPp3/2p1P1Q1/8/1rB3PP/R4R1K b - -",        // bm Rf8; id "BS2830.25"; c0 "Polugajewsky-Gutman, 4/97-41 Gruenfeld-Indisch, Ador, 1985";
        "1r4k1/1q2pN1p/3pPnp1/8/2pQ4/P5PP/5P2/3R2K1 b - -", // bm Qd5; id "BS2830.26"; c0 "Gyimesi,Z.-Polgar,S., 4/97-41 ChessBase, 1995";
        "2rq1rk1/pb3ppp/1p2pn2/4N3/1b1PPB2/4R1P1/P4PBP/R2Q2K1 w - -"
    };                      // bm d5; id "BS2830.27"; c0 "Kramnik,W.-Anand,V., 4/97-41  Schach Report, 1996";
    endwin();
    stop();

    flags.xboard = true;
    flags.post = false;
    setlinebuf(stdout);
    nodeWnd = 0;

    double t2 = searchTime();
    int benchCnt = sizeof(benchPos)/sizeof(char *);
    int nMoves = 0;
    for (int j = 0; j < benchCnt; j++) {
        importEPD(benchPos[j]);
	eList = movegen((constBoard *)&board, mList);
	for (int i = 0; i < 10000; i++) {
	    for (Move *currMove = mList; currMove < eList; currMove++) {
		doMove(&board, currMove);
		undoMove(&board);
	    }
	}
	nMoves += eList-mList;
    }
    t2 = searchTime() - t2;
    printf("doing/undoing %6.2f M moves per sec\n", 10000.0*nMoves/t2);

    t0 = searchTime();
    for (int j = 0; j < benchCnt; j++) {
        importEPD(benchPos[j]);
        for (int i = 0; i < 100000; i++) 
            eList = movegen((constBoard *)&board, mList);
    }
    t0 = searchTime() - t0;
    printf("generating %6.2f M moves per sec\n", 100000.0*nMoves/t0);
    
    double t3 = searchTime();
    int qMoves = 0;
    for (int j = 0; j < benchCnt; j++) {
        importEPD(benchPos[j]);
	eList = movegenq((constBoard *)&board, mList, 0.0);
	for (int i = 0; i < 30000; i++) {
	    for (Move *currMove = mList; currMove < eList; currMove++) {
		doMove(&board, currMove);
		undoMove(&board);
	    }
	}
	qMoves += eList-mList;
    }
    t3 = searchTime() - t3;    
    printf("doing/undoing %6.2f M cap moves per sec\n", 30000.0*qMoves/t3);
    
    t1 = searchTime();
    for (int j = 0; j < benchCnt; j++) {
        importEPD(benchPos[j]);
	for (int i = 0; i < 300000; i++) {
            eList = movegenq((constBoard *)&board, mList, 0.0);
	}
    }
    t1 = searchTime() - t1;
    printf("generating %6.2f M cap moves per sec\n", 300000.0*qMoves/t1);
    
    exitProfiler();
//    for (int bits = 29; bits < 30; bits+=2) {
    double pos, ttime;
    double avgPos = 0.0;
    double avgTime = 0.0;
    t0 = t1 = 1 S;
    for (int i = 0; i < benchCnt; i++) {
//        allocHash(&mainHash, bits);
        importEPD(benchPos[i]);
//        dynEval(&board);
//        exit(0);
        searchm(MAXDEPTH);
        pos = stats.pos + stats.qPos;
        ttime = searchTime() - stats.bTime;
        avgPos += log(pos);
        avgTime += log(ttime / 1e6);
        printf("position %2d: pos: %10.0f, time: %3.3f\n", i, pos * 0.001, ttime / 1e6);
    }
    ttime = exp(avgTime / benchCnt);
    pos = exp(avgPos / benchCnt);
    printf("%10.0f nps\n",  pos / ttime);
//    }
    exitProfiler();
    exit(0);
}

static void getTag(char *buf, const char *tag, const char *str)
{
    char *bmend;
    char *s = buf;
    char *bmstart = strstr(str, tag) + strlen(tag);
    if (bmstart >= str) {
        bmend = strstr(bmstart, ";");
        if (!bmend)
            bmend = bmstart + strlen(bmstart);
        while (bmstart < bmend)
            *s++ = *bmstart++;
    }
    *s = 0;
}

void sepParse(char *str) 
#ifdef HAVE_ARGZ_H
{
    char *argz;
    size_t argz_len;
    argz_create_sep(str, ' ', &argz, &argz_len);
    char **argv = alloca((argz_count (argz, argz_len) + 1) * sizeof (char *));
    argz_extract (argz, argz_len, argv);
    parse(argv);
    free(argz);
}
#else
{}
#endif

static int loadPos(const char *filename, const char *name) {
    char buf[1024];
    FILE *f = fopen(filename, "r");
    if (f) {
        for (int i=0; fgets(buf, sizeof(buf), f); i++)
            if (strstr(buf, name)) {
                importEPD(buf);
                char *s = strstr(buf, "id");
                if (s) {
                    char *s1 = strchr(s, ';');
                    if (s1)
                        *s1=0;
                    char out[256];
                    sprintf(out, "position %s ", s+3);
                    if (s1) {
                        s1 = strstr(s1+1, "c0");
                        if (s1) {
                            char *s2 = strchr(s1, ';');
                            if (s2)
                                *s2=0;
                            strcat(out, "(");
                            strcat(out, s1+3);
                            strcat(out, ") ");
                        }
                    }
                    strcat(out, "loaded");
                    output("%s", out);
                }
                return i;
            }
        output("position %s not found\n", name);
    } else
        output("file %s not found\n", filename);
    return -1;
}
 
void parse(char **argv)
{
    int inc;
    static Board *const b = &board;

    for (; *argv; argv++) {
//remove leading slashes, in case someone types --help as argument
	while (**argv == '-')
	    (*argv)++;
#ifdef MYDEBUG
    if (flags.debug & DEBUG_IO)
        debug("in:  %s\n", *argv);
#endif
    extern FILE *logfile;
    
    if (!strcmp(*argv, "help") || !strcmp(*argv, "h")) {
	if (!flags.xboard)
	    endwin();
	printf("sliboengine commands:\n");
	printf("\n");
	printf("new                 reset board to start position, set computer to black\n");
	printf("setboard <position> set board to a specific position, set computer to black\n");
	printf("                    position must be given in fen notation\n");
	printf("stat                output some more or less useful statistics in html format\n");
	printf("log                 log the game in a file slibo000.log\n");
	printf("nolog               disable logging\n");
#ifdef MYDEBUG	
	printf("debug <n>           give some debug information. n is the sum of\n");
	printf("                    1 send debug output to file debug.log\n");
	printf("                    2 send debug output to stderr\n");
	printf("                    4 show moves executed\n");
	printf("                    8 check integrity of internal board representation\n");
	printf("                    16 show incoming/outgoing commands\n");
	printf("                    32 show smp info\n");
	printf("                    64 show searchtree\n");
#endif	
	printf("test <match>        load the position from a file ""testsuite.epd"", where\n");
	printf("                    the ""id"" or ""c0"" tag contains <match>\n");
	printf("book                use opening book\n");
	printf("nobook              don't use opening book\n");
	printf("quit                end program\n");
	printf("hash <float>        set main transposition table to <float> MBytes\n");
	printf("hashp <float>       set pawn position cache to <float> MByte\n");
	printf("hasha <float>       set piece position cache to <float> MByte\n");
	printf("level <moves> <time> <increment> set time control to <moves> moves in <time>,\n");
	printf("                    with additional <increment> seconds per move. time in\n");
	printf("                    seconds or in ""minutes:seconds""\n");
	printf("easy                don't use the opponent thinking time\n");
	printf("hard                use it\n");
	printf("post                output main variations during search\n");
	printf("nopost              dont't do so\n");
	printf("analyze             analyze position, don't move\n");
	printf("move/go             set computer on turn and begin thinking\n");
	printf("bench               do a short benchmark\n");
	exit(0);
    }
    if (!strcmp(*argv, "new")) {
        if (flags.log)
            if (logfile) {
                fclose(logfile);
                logfile = 0;
            }
        if (!stopped()) {
            computerTurn = false;
            stop();
        }
        initBoard();
        initTC();
    } else if (!strcmp(*argv, "setboard")) {
        if (flags.log)
            if (logfile) {
                fclose(logfile);
                logfile = 0;
            }
        if (!stopped()) {
            computerTurn = false;
            stop();
        }
	if (!*++argv) {
	    output("usage: setboard <fen-board-position>");
	    return;
	}
	char boardstr[256];
	strcpy(boardstr, *argv);
	strcat(boardstr, " ");
	strcat(boardstr, *++argv);
	strcat(boardstr, " ");
	strcat(boardstr, *++argv);
	strcat(boardstr, " ");
	strcat(boardstr, *++argv);
	
	importEPD(boardstr);
        initTC();
    } else if (!strcmp(*argv, "stat")) {
        output("stat <table>");
        output("<tr><td>ply</td><td>%.2f</td><td></td></tr>", (double)stats.ply/PLY);
        output("<tr><td>move #</td><td>%d/%d</td></tr>", stats.currMoveNum + 1, stats.moveCount);
        output("<tr><td>positions</td><td>%.0f</td><td>%.0f</td></tr>", (stats.pos+stats.qPos) * 0.001, (stats.pos+stats.qPos) * 1000.0 / (searchTime() - stats.startTime));
        output("<tr><td>moves done</td><td>%.0f</td><td>%.0f</td></tr>", (stats.moves+stats.qMoves) * 0.001, (stats.moves+stats.qMoves)*1000.0/(searchTime()-stats.startTime));
	if ((stats.pos+stats.qPos) && (stats.moves+stats.qMoves))
        output("<tr><td>quiescence positions/moves</td><td>%.2f%%</td><td>%.2f%%</td></tr>", stats.qPos/(stats.qPos + stats.pos) * 100.0, stats.qMoves/(stats.moves+stats.qMoves)*100.0);
//    mvwprintw(wnd, 2, 0, "%8.0f %8.0f %8.0f %8.0f", stats.pos[1] * 0.001, stats.pos[2] * 0.001, stats.pos[3] * 0.001, stats.pos[4] * 0.001);
//    mvwprintw(wnd, 3, 0, "%8.0f %8.0f %8.0f %8.0f", stats.qPos[1] * 0.001, stats.qPos[2] * 0.001, stats.qPos[3] * 0.001, stats.qPos[4] * 0.001);
        output("<tr><td>main TT probes</td><td>%.0f</td></tr>", mainHash.read * 0.001);
        output("<tr><td>main TT 1st level stores</td><td>%.0f</td></tr>", mainHash.write * 0.001);
        output("<tr><td>main TT 2nd level stores</td><td>%.0f</td></tr>", stats.write2 * 0.001);
        output("<tr><td>main TT stores & shift</td><td>%.0f</td></tr>", stats.shift2 * 0.001);
        output("<tr><td>main TT fill</td><td>%d/%d</td></tr>", mainHash.fill / 1024, mainHash.count / 1024);
        output("<tr><td>main TT 1st level hits</td><td>%.0f</td></tr>", mainHash.hit * 0.001);
        output("<tr><td>main TT 2nd level hits</td><td>%.0f</td></tr>", stats.hit2 * 0.001);
        if (mainHash.read)
        output("<tr><td>main TT hit ratio</td><td>%.0f%%</td></tr>", 100.0*(mainHash.hit+stats.hit2)/mainHash.read);

        output("<tr><td>main TT return</td><td>%.0f</td></tr>", stats.ret * 0.001);
        output("<tr><td>main TT alpha return</td><td>%.0f</td></tr>", stats.aret * 0.001);
        output("<tr><td>main TT beta return</td><td>%.0f</td></tr>", stats.bret * 0.001);
        output("<tr><td>main TT pre cut</td><td>%.0f</td></tr>", stats.precut * 0.001);
        output("<tr><td>main TT best move</td><td>%.0f</td></tr>", stats.BM * 0.001);

        output("<tr><td>pawn TT probes</td><td>%.0f</td></tr>", pawnHash.read * 0.001);
        output("<tr><td>pawn TT stores</td><td>%.0f</td></tr>", pawnHash.write * 0.001);
        output("<tr><td>pawn TT fill</td><td>%d/%d</td></tr>", pawnHash.fill / 1024, pawnHash.count / 1024);
        output("<tr><td>pawn TT hits</td><td>%.0f</td></tr>", pawnHash.hit * 0.001);
        if (pawnHash.read)
        output("<tr><td>pawn TT miss ratio</td><td>%.0f%%</td></tr>", 100.0*pawnHash.write/pawnHash.read);

        output("<tr><td>piece TT probes</td><td>%.0f</td><td>%.0f</td></tr>", whiteHash.read * 0.001, blackHash.read * 0.001);
        output("<tr><td>piece TT stores</td><td>%.0f</td><td>%.0f</td></tr>", whiteHash.write * 0.001, blackHash.write * 0.001);
        output("<tr><td>piece TT fill</td><td>%d/%d</td><td>%d/%d</td></tr>", whiteHash.fill / 1024, whiteHash.count / 1024, blackHash.fill / 1024, blackHash.count / 1024);
        output("<tr><td>piece TT hits</td><td>%.0f</td><td>%.0f</td></tr>", whiteHash.hit * 0.001, blackHash.hit * 0.001);
        if (whiteHash.read && blackHash.read)
        output("<tr><td>piece TT miss ratio</td><td>%.0f%%</td><td>%.0f%%</td></tr>", 100.0*whiteHash.write/whiteHash.read, 100.0*blackHash.write/blackHash.read);

 	output("<tr><td>pawn extension</td><td>%.0f</td></tr>", stats.extPawn * 0.001);
        output("<tr><td>mate extension</td><td>%.0f</td></tr>", stats.extMate * 0.001);
        output("<tr><td>check extension</td><td>%.0f</td></tr>", stats.extCheck * 0.001);
        output("<tr><td>threat extension</td><td>%.0f</td></tr>", stats.extHorizon * 0.001);
        output("<tr><td>recapture extension</td><td>%.0f</td></tr>", stats.extRecapture * 0.001);

        output("<tr><td>null move cutoff</td><td>%.0f</td></tr>",  stats.null * 0.001);
        output("<tr><td>of</td><td>%.0f</td></tr>",  stats.nulltry * 0.001);
        output("<tr><td>zugzwang detected</td><td>%.0f</td></tr>",  stats.zz * 0.001);
        output("</table>\n");
    } else if (!strcmp(*argv, "log")) {
        flags.log = true;
    } else if (!strcmp(*argv, "nolog")) {
        if (flags.log)
            if (logfile) {
                fclose(logfile);
                logfile = 0;
            }
        flags.log = false;
    } else if (!strcmp(*argv, "debug")) {
        if (!*argv[1]) {
            output("debug takes one parameter");
            return;
        }
        if (!sscanf(*++argv, "%d", &flags.debug))
            output("debug: wrong parameter");
    } else if (!strcmp(*argv, "minmaxbook")) {
        minmaxBook();
    } else if (!strcmp(*argv, "test")) {
        if (!*++argv) {
            output("test takes one parameter");
            return;
        }
        stop();
        loadPos("testsuite.epd", *argv);
    } else if (!strcmp(*argv, "book")) {
        flags.nobook = false;
        output("using opening book");
    } else if (!strcmp(*argv, "nobook")) {
        flags.nobook = true;
        output("not using opening book");
    } else if (!strcmp(*argv, "ping")) {
        output("pong %s\n", *++argv);

    } else if (!strcmp(*argv, "protover")) {
        argv++;
        output("feature stat=1 ping=1 setboard=1 playother=0 time=0 draw=0 sigint=0 sigterm=0 reuse=1 analyze=1 colors=0 name=0 ics=0 myname=\"Slibo 0.4\" done=1\n");

    } else if (!strcmp(*argv, "quit") || !strcmp(*argv, "exit")) {
        if (!flags.xboard)
            endwin();
#ifdef MYDEBUG
        if (flags.debug & DEBUG_TOLOGFILE)
            fclose(debuglog);
#endif
        exitProfiler();
        exit(0);

    } else if (!strcmp(*argv, "hash")) {
        if (!*++argv) {
            output("usage: hash <hashsize in MByte>");
            return;
        }
        stop();
        double newsize;
        sscanf(*argv, "%lf", &newsize);
        allocHash(&mainHash, lrint(newsize*0x100000));
        output("main hashtable set to %d MBytes, %d k Entrys", mainHash.size >> 20, mainHash.count >> 10);
    } else if (!strcmp(*argv, "hashp")) {
        if (!*++argv) {
            output("usage: hashp <hashsize in MByte>");
            return;
        }
        stop();
        double newsize;
        sscanf(*argv, "%lf", &newsize);
        allocHash(&pawnHash, lrint(newsize*0x100000));
        output("pawn hashtable set to %d MBytes, %d k Entrys", pawnHash.size >> 20, pawnHash.count >> 10);
    } else if (!strcmp(*argv, "hasha")) {
        if (!*++argv) {
            output("usage: hasha <hashsize in MByte>");
            return;
        }
        stop();
        double newsize;
        sscanf(*argv, "%lf", &newsize);
        allocHash(&whiteHash, lrint(newsize*0x100000));
        allocHash(&blackHash, lrint(newsize*0x100000));
        output("piece hashtables set to %d MBytes, %d k Entrys each", whiteHash.size >> 20, whiteHash.count >> 10);
    } else if (!strcmp(*argv, "level")) {
        long long sec = -1;
        long long min;
        long long moves;
        if (!argv[0] || !argv[1] || !argv[2]) {
            output("usage: level <moves> <time> <increment>");
            return;
        }
        sscanf(*++argv, "%lld", &moves);
        sscanf(*++argv, "%lld:%lld", &min, &sec);
        sscanf(*++argv, "%d", &inc);
        if (sec != -1)
            initTime[0] = initTime[2] = (min * 3600 + sec * 60) * 1 S;
        else
            initTime[0] = initTime[2] = min * 60 * 1 S;

        initMoves[0] = initMoves[2] = moves;
        initTC();
    } else if (!strcmp(*argv, "white")) {
#ifdef ACCEPT_COLORS
        stop();
        board.color = WHITE;
#endif
        ;
    } else if (!strcmp(*argv, "black")) {
#ifdef ACCEPT_COLORS
        stop();
        board.color = BLACK;
#endif
        ;
    } else if (!strcmp(*argv, "easy")) {
        pondermode = 0;

    } else if (!strcmp(*argv, "hard")) {
        pondermode = 1;

    } else if (!strcmp(*argv, "post")) {
        flags.post = 1;

    } else if (!strcmp(*argv, "nopost")) {
        flags.post = 0;

    } else if (!strcmp(*argv, "xboard")) {
        flags.xboard = 1;
        setbuf(stdout, NULL);
        endwin();
        output("\n");
    } else if (!strcmp(*argv, "analyze")) {
        unPonder();
        mode = Analyze;
        timeRunning = t0 = t1 = 0;
        start();
    } else if (!strcmp(*argv, "benchsearch")) {
        endwin();
        stop();
        flags.xboard = true;
        flags.post = true;
        setbuf(stdout, NULL);
        nodeWnd = 0;
        char buf3[500], buf2[500];
        double lnavgTime = 0.0;
        FILE *f = fopen("all.epd", "r");
        FILE *of = fopen("100.epd", "a");
        double count = 0.0;
        while (!feof(f)) {
            fgets(buf3, 499, f);
            importEPD(buf3);
            eList = movegen((constBoard *)b, mList);
            getTag(buf2, "c0 ", buf3);
            printf("***** %s ", buf2);
            getTag(buf2, "bm ", buf3);
            str2move(buf2, &breakMove, mList, eList);
            printf("***** %s\n", buf2);
            copyBoard();
            printBoard(&board, stdout);
            t0 = 100 S;
            t1 = 100 S;
            solutionTime = 100 S;
            searchm(MAXDEPTH);
            getTag(buf2, "id ", buf3);
            if (solutionTime == 0)
                solutionTime = 100 S;
            printf("%s: %4.3f solution time\n\n", buf2, solutionTime / 1e6);
            lnavgTime += log(solutionTime);
            if (solutionTime > 1 S && solutionTime < 100 S)
                fputs(buf3, of);
            count += 1.0;
        }
        printf("Mean time (geometric) %4.3f s\n", exp(lnavgTime / count));
        fclose(of);
        fclose(f);
        exit(0);
    } else if (!strcmp(*argv, "workload")) {
        endwin();
        flags.xboard = true;
        const char *workload[] = {
"4r1k1/p1pb1ppp/Qbp1r3/8/1P6/2Pq1B2/R2P1PPP/2B2RK1 b - - bm Qxf3; id \"BS2830.01\"; c0 \"Paulsen-Morphy, 4/97-39 Damenopfer, Neistadt, 1857\";",
"7r/2qpkp2/p3p3/6P1/1p2b2r/7P/PPP2QP1/R2N1RK1 b - - bm f5; id \"BS2830.02\"; c0 \"J.Xie-J.Polgar, 4/97-39 Schacholympiade Saloniki, 1988\";",
"r1bq1rk1/pp4bp/2np4/2p1p1p1/P1N1P3/1P1P1NP1/1BP1QPKP/1R3R2 b - - bm Bh3+; id \"BS2830.03\"; c0 \"Raman-Aaron, 4/97-39 Albin Poetzsch, Spass am Kom, 1977\";",
"8/2kPR3/5q2/5N2/8/1p1P4/1p6/1K6 w - - bm Nd4; id \"BS2830.04\"; c0 \"J.de Villeneuve_Esclapon-[+3101.22b1c7], 4/97-39 L Echiquier Francais, 1906\";",
"2r1r3/p3bk1p/1pnqpppB/3n4/3P2Q1/PB3N2/1P3PPP/3RR1K1 w - - bm Rxe6; id \"BS2830.05\"; c0 \"Sturua-Kutirow, 4/97-39 Rochochade Europa, 1976\";",
"8/2p5/7p/pP2k1pP/5pP1/8/1P2PPK1/8 w - - bm f3; id \"BS2830.06\"; c0 \"Szabo-Pirc, 4/97-39 Bauern Endsp. J.Awerbach, 1938\";",
"8/5p1p/1p2pPk1/p1p1P3/P1P1K2b/4B3/1P5P/8 w - - bm b4; id \"BS2830.07\"; c0 \"Smyslow-NN, 4/97-39 Schachmate I.Maizeliw, 1937\";",
"rn2r1k1/pp3ppp/8/1qNp4/3BnQb1/5N2/PPP2PPP/2KR3R b - - bm Bh5; id \"BS2830.08\"; c0 \"Stoica-Uhlmann, 4/97-39 Franzoesisch-Uhlmann, 1979\";",
"r3kb1r/1p1b1p2/p1nppp2/7p/4PP2/qNN5/P1PQB1PP/R4R1K w kq - bm Nb1; id \"BS2830.09\"; c0 \"Spassky,B.-Fischer,R., 4/97-39 WCh Reykjavik, 1972\";",
"r3r1k1/pp1bp2p/1n2q1P1/6b1/1B2B3/5Q2/5PPP/1R3RK1 w - - bm Bd2; id \"BS2830.10\"; c0 \"Khalifman-Gawrikow, 4/97-39 Gruenfeld-Iindisch Ador, 1988\";",
"r3k2r/pb3pp1/2p1qnnp/1pp1P3/Q1N4B/2PB1P2/P5PP/R4RK1 w kq - bm exf6; id \"BS2830.11\"; c0 \"Tal-Hecht, 4/97-39 Schacholympiade, 1962\";",
"r1b1r1k1/ppp2ppp/2nb1q2/8/2B5/1P1Q1N2/P1PP1PPP/R1B2RK1 w - - bm Bb2; id \"BS2830.12\"; c0 \"Aljechin-Rabinowitsch, 4/97-39 Meine besten Partien, 1918\";",
"rnb1kb1r/1p3ppp/p5q1/4p3/3N4/4BB2/PPPQ1P1P/R3K2R w KQkq - bm O-O-O; id \"BS2830.13\"; c0 \"Nimzowitsch-Alapin, 4/97-40 ChessBase, 1911\";",
"r1bqr1k1/pp1n1ppp/5b2/4N1B1/3p3P/8/PPPQ1PP1/2K1RB1R w - - bm Nxf7; id \"BS2830.14\"; c0 \"Timman-Kortschnoi, 4/97-40 CSS, 1991\";",
"2r2rk1/1bpR1p2/1pq1pQp1/p3P2p/P1PR3P/5N2/2P2PPK/8 w - - bm Kg3; id \"BS2830.15\"; c0 \"Short-Timman, 4/97-40 CSS, 1991\";",
"8/pR4pk/1b6/2p5/N1p5/8/PP1r2PP/6K1 b - - bm Rxb2; id \"BS2830.16\"; c0 \"Ortueta,M.-Sanz,J., 4/97-40 Schachendspiele-Gawlik, 1934\";",
"r1b1qrk1/ppBnppb1/2n4p/1NN1P1p1/3p4/8/PPP1BPPP/R2Q1R1K w - - bm Ne6; id \"BS2830.17\"; c0 \"Bednorz,A.-Spinath P.D., 4/97-40 Fernpartie, 1990\";",
"8/8/4b1p1/2Bp3p/5P1P/1pK1Pk2/8/8 b - - bm g5; id \"BS2830.18\"; c0 \"Kotow-Botwinnik, 4/97-40 Schachmate-Majzeliw, 1955\";",
"r3k2r/pp1n1ppp/1qpnp3/3bN1PP/3P2Q1/2B1R3/PPP2P2/2KR1B2 w kq - bm Be1; id \"BS2830.19\"; c0 \"Karpov-Rogers, 4/97-40 So schnell siegen WM-K, 1983\";",
"r1bqk2r/pppp1Npp/8/2bnP3/8/6K1/PB4PP/RN1Q3R b kq - bm O-O; id \"BS2830.20\"; c0 \"Gofman-Petrow, 4/97-40 Schachmate-Maizeliw,I, 1844\";",
"r4r1k/pbnq1ppp/np3b2/3p1N2/5B2/2N3PB/PP3P1P/R2QR1K1 w - - bm Ne4; id \"BS2830.21\"; c0 \"Kasparov-Marjanovic, 4/97-40 So schnell sieg.WM-Kar, 1980\";",
"r2qr2k/pbp3pp/1p2Bb2/2p5/2P2P2/3R2P1/PP2Q1NP/5RK1 b - - bm Qxd3; id \"BS2830.22\"; c0 \"Euwe-Keres, 4/97-40 Rochade Europa, 1940\";",
"5r2/1p4r1/3kp1b1/1Pp1p2p/2PpP3/q2B1PP1/3Q2K1/1R5R b - - bm Rxf3; id \"BS2830.23\"; c0 \"Palatnik-Krupa, 4/97-40 Spass am Kombinieren, 1984\";",
"8/7p/8/7P/1p6/1p5P/1P2Q1pk/1K6 w - - bm Ka1; id \"BS2830.24\"; c0 \"Zakon,E.-[+1000.34b1h2], 4/97-40 The Jerusalem Post, 1953\";",
"r5k1/p4n1p/6p1/2qPp3/2p1P1Q1/8/1rB3PP/R4R1K b - - bm Rf8; id \"BS2830.25\"; c0 \"Polugajewsky-Gutman, 4/97-41 Gruenfeld-Indisch, Ador, 1985\";",
"1r4k1/1q2pN1p/3pPnp1/8/2pQ4/P5PP/5P2/3R2K1 b - - bm Qd5; id \"BS2830.26\"; c0 \"Gyimesi,Z.-Polgar,S., 4/97-41 ChessBase, 1995\";",
"2rq1rk1/pb3ppp/1p2pn2/4N3/1b1PPB2/4R1P1/P4PBP/R2Q2K1 w - - bm d5; id \"BS2830.27\"; c0 \"Kramnik,W.-Anand,V., 4/97-41  Schach Report, 1996\";"
        };
        for (int i=0; i<27; i++) {
            printf(".");
            fflush(stdout);
            importEPD(workload[i]);
            t0 = t1 = 1 S;
            searchm(MAXDEPTH);
        }
        exit(0);
    } else if (!strcmp(*argv, "lct")) {
        endwin();
        setlinebuf(stdout);
        stop();
        flags.xboard = true;
        flags.post = true;
        nodeWnd = 0;
        char buf3[256];
        char buf2[512];
        extern long long solutionMoves;
        FILE *of = fopen("out1.txt", "w");
        setlinebuf(of);
        extern int extThreat;
        extern int extLevel;
//        for (extThreat = (PLY/4); extThreat <= PLY/2; extThreat += PLY/8)
//        for (extCheck = PLY; extCheck <= PLY; extCheck += PLY/4)
//        for (extLevel = PLY*5; extLevel <= 5*PLY; extLevel += 2*PLY)
        for (maxNullMove = 6*PLY/2; maxNullMove <= 6*PLY/2; maxNullMove+=PLY/2) {
            int ppos = 0;
            int pend = 0;
            int pcmb = 0;
            double avgTime = 0.0;
            double avgMoves = 0.0;
            FILE *f = fopen("LCT2.EPD", "r");
            for (int i = 0; i < 35; i++) {
                fgets(buf2, 500, f);
//                if (i<14) continue;
//                if (i>=26) continue;
                importEPD(buf2);
//                printBoard(&board, stdout);
//                printBoard(&board, of);
                eList = movegen((constBoard *)b, mList);
                getTag(buf3, "c0 ", buf2);
                printf("*** %s", buf3);
                getTag(buf3, "bm ", buf2);
                printf(" *** %s\n", buf3);
                str2move(buf3, &breakMove, mList, eList);
                t0 = 600 S;
                t1 = 600 S;
                solutionTime = 0 S;
                searchm(MAXDEPTH);
                getTag(buf3, "id ", buf2);
                if (solutionTime == 0) {
                    solutionTime = 600 S;
                    solutionMoves =  (stats.qMoves + stats.moves);
                }
                if (i<14)
                    ppos += points(solutionTime);
                else if (i<26)
                    pcmb += points(solutionTime);
                else
                    pend += points(solutionTime);    
                fprintf(of, "%s: %4.3f solution time, %6.0f kmoves\n", buf3, solutionTime / 1e6, solutionMoves * 0.001);
                printf("%s: %4.3f solution time, %6.0f kmoves\n", buf3, solutionTime / 1e6, solutionMoves * 0.001);
                avgTime += log(solutionTime / 1e6);
                avgMoves += log(solutionMoves * 0.001);
/*                for (int j=1; j<20; j++)
                if (moveord[j][7] || moveord[j][9]) {
                printf("%5.2f %12lld %5.2f %12lld\n", 100.0*moveord[j][6]/(moveord[j][7]+0.0000001), moveord[j][7], 100.0*moveord[j][8]/(moveord[j][9]+0.00001), moveord[j][9]);
                fprintf(of, "%5.2f %12lld %5.2f %12lld\n", 100.0*moveord[j][6]/(moveord[j][7]+0.0000001), moveord[j][7], 100.0*moveord[j][8]/(moveord[j][9]+0.00001), moveord[j][9]);
}*/
                }
//                fprintf(of, "Ext. H %x Ext. M. %x Ext C. %x Null %x Mean time %4.3f s, %6.0f\n\n", extHorizon, extMate, extCheck, maxNullMove, pow(solutionTime, 1.0 / 20.0), pow(solutionMoves, 1.0 / 20.0));
//                printf("Ext H %x Ext. M. %x Ext C. %x Null %x Mean time %4.3f s, %6.0f\n\n", extHorizon, extMate, extCheck, maxNullMove, pow(solutionTime, 1.0 / 20.0), pow(solutionMoves, 1.0 / 20.0));
            fprintf(of, "***** %x %x %x %x pos %d, cmb %d, end %d, total %d, mean time %4.3f s, mean kmoves %6.0f\n", extThreat, extCheck, extLevel, maxNullMove, ppos, pcmb, pend, 1900+ppos+pcmb+pend, exp(avgTime/(12.0)), exp(avgMoves/(12.0))); 
            printf("***** %x %x %x %x pos %d, cmb %d, end %d, total %d, mean time %4.3f s, mean kmoves %6.0f\n",extThreat, extCheck, extLevel, maxNullMove,  ppos, pcmb, pend, 1900+ppos+pcmb+pend, exp(avgTime/(12.0)), exp(avgMoves/(12.0))); 
                fclose(f);
        }
        exit(0);

    } else if (!strcmp(*argv, "bench")) {
        bench();
    } else if (!strcmp(*argv, "move") || !strcmp(*argv, "go")) {
        mode = Normal;
        if (pondering) {
            unPonder();
        }
        computerTurn = true;
        setTC();
        start();

    } else if (!strcmp(*argv, "auto")) {
        mode = Autoplay;
        if (pondering) {
            unPonder();
        }
        computerTurn = true;
        setTC();
        start();

    } else if (!strcmp(*argv, "force")) {
#ifdef ACCEPT_FORCE
        mode = Force;
        if (pondering) {
            unPonder();

        }
        computerTurn = false;
        timeRunning = 0;
	stop();
        setTC();
#endif
        ;
    } else if (!strcmp(*argv, "random")) {
        ;
    } else if (!strcmp(*argv, "computer")) {
        ;
    } else if (!strcmp(*argv, "accepted")) {
        argv++;
    } else if (strlen(*argv) > 6) {
        output("%s: unknown command", *argv);
    } else if (strlen(*argv) > 0) {
        inputMove(*argv);

    }
    };
}
