/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "search.h"
#include "board.h"
#include "eval.h"
#include "hashfunc.h"
#include "util.h"
#include "opening.h"
#include "movegen_c.h"
#include "debug.h"
#include "smp.h"
#include "squares.h"
#include "neuron.h"
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#define VERIFY_LATER
//#define NEGAINT
//#define ASPIRATION
#define ASPIRATION_WINDOW 8*QUANT
#define USE_QHASH
#define PREHASH
#define PLY_PREHASH (3*PLY)
#define PLY_LEARN (4*PLY)
#define PLY_EVAL (6*PLY)

#define MAXCHILDREN 50
#define MAX_DYNEVAL 1.0     //maximum change in dynEval() in one move

#ifdef MYDEBUG
unsigned long long moveord[64][11];
#endif
float PRUNE3 = 99.0;
float PRUNE2 = 59.0;
float SCORE_TOMOVE = 0.125;
Move killer[MAXDEPTH];
Move killer2[MAXDEPTH];

int innull, inzz;
Move breakMove;
unsigned long long t0, t1, solutionTime, solutionMoves;
unsigned changes[MAXDEPTH][8];
float bookLimit = 0.25;

int maxNullMove = 3*PLY;
int maxPly = 0;

int extHorizon = 0;
int extCheck = PLY;
int extLevel = 5*PLY;
int extRecapture = 0;
int extMate = 0;
int extPawn = 0;
int extThreat = PLY/4;

double cnt, sum;

FILE *logfile = 0;

static float searchq1(Board *, float beta, float alpha, int) __attribute__ ((regparm(1)));
static float searchq(Board *, float beta, float alpha, int) __attribute__ ((regparm(1)));
float search(Board *, int nullMove, int verifyNull, float beta, float alpha, int) __attribute__ ((regparm(1)));

//main search function.
//takes a pointer to the board to search, flags for null move pruning and
//search extensions, the lower and the upper bound, the remaining ply to
//search and a pointer to the best line.
//returns the score of the position and the best line.

float search(Board *b, int nullMove, int verifyNull, float alpha, float beta, int ply) {
    static unsigned int clockCounter = 0;

    BEGIN_PROFILE(TIME_SEARCH);
    // check the clocks, every 64th iteration of search
    // since search is called 100+ times every cpuclock (in MHz), this is fast enough.
   
   if (clockCounter++ > (cpuclock << 6) && t1) {
        clockCounter = 0;
        if (searchTime() > t1 + stats.startTime) {
            stopSearch = true;
            END_PROFILE();
            return drawValue;
        }
    }

    //check for move repetiton and 50 move rule.
    //repTable uses the normal key.main.keys, the main difference is data
    //gets not overwritten. do not insert current position into the reptable,
    //since we still have some return points to check
#ifdef MYDEBUG	
    if (flags.debug & DEBUG_SEARCH) {
        for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
        debug("search: a=%+5.2f, b=%+5.2f, ply=%2d\n", alpha, beta, ply/PLY);
    }
#endif
/*    if (b->currMove[0].move.src == b->currMove[-2].move.dst && b->currMove[0].move.dst == b->currMove[-2].move.src
        && (b->currMove[0].captured | b->currMove[-1].captured | b->currMove[-2].captured) == 0
        && alpha < 0.0) {
            alpha = 0.0;
            if (alpha >= beta) {
                END_PROFILE();
                return 0.0;
            }    
    }*/
        
    int repIdx = REPIDX;
    while (b->repTable[repIdx] && b->repTable[repIdx] != b->key.main.key)
        repIdx = (repIdx + 1) & (REPTABLESIZE - 1);
    if (b->repTable[repIdx] == b->key.main.key) {
        END_PROFILE();
        return drawValue;
    }
    //if 100 moves happend without capturing, return draw. if we are in check,
    //allow 101 moves to see if we are mated.
    if (b->fiftyMoves > 100) {
        END_PROFILE();
        return drawValue;
    }
    //TODO: write inMate()
#define inMate(x) 0    
    if (b->fiftyMoves == 100) {
        if (inMate(b)) {
            END_PROFILE();
            return MATE_VALUE;
        } else {            
            END_PROFILE();
            return drawValue;
        }    
    }

    //ask hashtable if current position is already stored. returns a pointer to
    //a hashEntry if found or NULL.
    //if the ply of the stored pos is sufficient, return if it is
    //a) a exact score
    //b) a lower bound (value of position is larger or equal than the score stored), if
    //   it is higher than the current upper bound of the search window (beta)
    //c) vice versa with alpha.

    MainEntry *p = readHash(b);
    //TODO:
    //check if bestAlpha/Beta are really useful
    float bestBeta = beta;
    float bestAlpha = alpha;
    float bestEval = MATE_VALUE;
    if (p) {
        float v = HASH_VALUE(p);
        if ((p->ply >= ply || p->endofsearch)) {
            switch (p->bound) {
                case EXACT_VALUE:
                    stats.ret++;
                    END_PROFILE();
                    return HASH_VALUE(p);
                case LOWER_BOUND:
                    if (v >= beta) {
                        stats.bret++;
                        END_PROFILE();
                        return v;
                    }
                    if (v > alpha) {
                        nullMove = 0;
//                        bestEval = bestAlpha = v;
                    }
                    stats.alpha++;
                    break;
                case UPPER_BOUND:
                    if (v <= alpha) {
                        stats.aret++;
                        END_PROFILE();
                        return v;
                    }
                    if (v < beta) {
                        nullMove = 0;
                        //                    bestBeta = HASH_VALUE(p);
                    }
                    stats.beta++;
                    break;
            }
        }
#ifdef ASPIRATION
        if (v < bestBeta)
            bestBeta = v + ASPIRATION_WINDOW;
#endif
        }

    {
        float retval;
        if (endgame(b, &retval)) {
            END_PROFILE();
            return retval;
        }
    }

    //null move pruning. do no move and let the other side move again.
    //if the score is still higher than beta, this move is really bad for
    //the other side, cut it off. this doesn't work in zugzwang situations,
    //where by definition not moving is an advantage. to detect this, do a shallow
    //search before, if the score is less than beta, disable nullmoves.

    int check = IN_CHECK(b->color);     //has to be int (not bool), since inCheck() may return somthing > 0x100
    int threat = false;
    if (nullMove && !check && ply > PLY && CNT(b->color * ROOK)*10 + CNT(b->color * BISHOP)*6 + CNT(b->color * QUEEN)*18 + CNT(b->color * KNIGHT)*6 + CNT(b->color * PAWN) > 10) {
//        int newPly = max(ply - nullMove - PLY, PLY);
        int newPly = ply - nullMove - PLY;
        float nullEval, prevEval;
        stats.nulltry++;
        if (newPly > 0) {
#ifndef VERIFY_LATER
            prevEval = search(b, maxNullMove, verifyNull, beta-QUANT, beta, newPly);
#endif
            b->color = -b->color;
            nullEval = -search(b, 0, verifyNull, -beta, -beta+QUANT, newPly);
            b->color = -b->color;
        } else {
#ifndef VERIFY_LATER
            prevEval = searchq(b, beta-QUANT, beta, newPly);
#endif
            b->color = -b->color;
            nullEval = -searchq(b, -beta, -beta+QUANT, newPly);
            b->color = -b->color;
        }
#ifndef VERIFY_LATER 
        if (nullEval >= beta && prevEval >= nullEval) 
#else
        if (nullEval >= beta) 
#endif
        {
#ifdef VERIFY_LATER
	    if (!verifyNull)
		goto nullprune;
	    newPly = ply - PLY;
            if (newPly > 0)
		prevEval = search(b, maxNullMove, 0, beta-QUANT, beta, newPly);
	    else
		prevEval = searchq(b, beta-QUANT, beta, newPly);
	    if (prevEval >= beta) {
nullprune:
#endif
		stats.null++;
		//TODO: why not writing with ply? wasn't this already done by zz?
//		writeHash(b, CURR_HASH_PTR, newPly, nullEval, alpha, beta, 0, 0);
#ifdef MYDEBUG
		if (flags.debug & DEBUG_SEARCH) {
		    for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
		    debug("return %+4.2f (pruned)\n", nullEval);
		}	
#endif
		END_PROFILE();
		return nullEval;
#ifdef VERIFY_LATER
	    }
#endif
        } else 
            stats.zz++;
    }

    if (check) {
        if (!b->currMove->move.attack)
            ply += extCheck;	
        int newPly = ply - 2*PLY - PLY;
        float prevEval;
        if (newPly > 0) 
            prevEval = search(b, maxNullMove, verifyNull, alpha, alpha+QUANT, newPly);
        else 
            prevEval = searchq(b, alpha, alpha+QUANT, newPly);
        if (prevEval <= alpha && ply > PLY)
            ply -= PLY;   
    }
    //generate moves
    //if (ply > PLY_EVAL) eval();
    Move firstMove[MAXMOVES];
        b->color = -b->color;
        inCheckSig(b);
        b->color = -b->color;
    Move *lastMove = movegen((constBoard *)b, firstMove);
    (ply <= 0) ? stats.qPos : stats.pos += lastMove - firstMove;

    Move bestMove;
#ifdef PREHASH
    //preview all moves in hashtable
    //look at every position a move would generate in the hasttable. the moves are
    //not really executed, because this is way slower than the pretty fast
    //generation of haskeys. maybe we get a quick beta cutoff.
    BEGIN_PROFILE(TIME_PREHASH);
    if (ply >= PLY_PREHASH)
        for (Move *currMove = firstMove; currMove < lastMove; currMove++) {
            MainEntry *nextHash = preHash(b, currMove);
            //TODO: include move repetition here
            if (nextHash)
                if ((nextHash->ply >= ply - PLY || nextHash->endofsearch)) {
                    float currEval;
                    switch (nextHash->bound) {
                        case EXACT_VALUE:
                        case UPPER_BOUND:
                            currEval = -HASH_VALUE(nextHash);
                            if (currEval >= beta) {  //make sure to only return the shortest mate, so continue search
                                bestMove = *currMove;
                                if (currEval > INF)
                                    currEval--;
                                stats.precut++;
                                //TODO: check if writing with +1 is really valid
                                writeHash((constBoard *)b, ply, currEval, alpha, beta, &bestMove, 0);
                                END_PROFILE();
                                END_PROFILE();
                                return currEval;
                            }
                        default:
                            ;
                    }
                }
        }
    END_PROFILE();
#endif

    //sort moves, not a full sort, only the first 8 are sorted.
    BEGIN_PROFILE(TIME_SORT);
    sort(firstMove, lastMove - 1, b->color);
    END_PROFILE();
    
    //move ordering: fist take the move from the hashtable (if the ply was sufficient)
    //else do a iterative search with 2 plys less (if the ply > 2)
    BEGIN_PROFILE(TIME_MOVEORD);
    bestMove.src = bestMove.dst = 0;
    float oldEval = -INF;
    if (p && p->ply >= ply - 2*PLY) {
        oldEval = HASH_VALUE(p);
        //we have a hashentry, and its already presearched with a sufficent ply.
        //now check if there is a known best move. 
move_ord_2nd_try:        
        if (p->src | p->dst) {
            for (Move *currMove = firstMove; currMove < lastMove; currMove++)
#ifdef SMALLHASH             
                if (currMove->src == ADR(p->src >> 3, p->src & 7) && currMove->dst == ADR(p->dst >> 3, p->dst & 7)) {
#else                
                if (currMove->src == p->src << 2 && currMove->dst == p->dst << 2 && currMove->special == p->special) {
#endif                    
                    bestMove = *currMove;
                    *currMove = *firstMove;
                    *firstMove = bestMove;
                    break;
                }
        //if not, check if this is position was searched in the quiscense search before,
        //in this case we at least know that a non-capturing move is the best one.
        //if it was searched deeper, its a mate or something, so don't bother further
        //with move ordering.
        } else if (p->ply <= 0) {
            Move *firstncap;
            for (firstncap = firstMove; firstncap < lastMove; firstncap++) 
                if (BOARD22(firstncap->dst) == 0 && firstncap->special != SPEC_PROMOQ) break;
            Move *ncap = firstncap;
            for (Move *curr=firstMove; curr<firstncap && ncap < lastMove; curr++, ncap++) {
                Move temp = *curr;
                *curr = *ncap;
                *ncap = temp;
            }
        }
        stats.BM++;
    } else if (ply > 2*PLY) {
        if (ply > 2*PLY)    
            oldEval = search(b, maxNullMove, verifyNull, bestAlpha, bestBeta, ply - 2*PLY);
        else
            oldEval = searchq(b, bestAlpha, bestBeta, ply - 2*PLY);
    
#ifdef ASPIRATION
        if (bestBeta == beta && v < bestBeta)
            bestBeta = v + ASPIRATION_WINDOW;
#endif
        if (!p) { 
            p = readHash(b);
            if (p) goto move_ord_2nd_try;
        }    
    } /*else {
        for (currMove = firstMove; currMove < lastMove; currMove++)
            if (currMove->src == killer[currDepth].src  && currMove->dst == killer[currDepth].dst) {
                bestMove = *currMove;
                *currMove = *firstMove;
                *firstMove = bestMove;
                break;
            }
    }
        */

    END_PROFILE();
    //store position in repetition table before we
    b->repTable[repIdx] = b->key.main.key;
    //finally check all possible moves.
#ifdef SMP

    struct {
        Line bestLine;
        int thread;
        float val;
        volatile int finished;
    }
    child[MAXCHILDREN];
    int children = 0;
#endif

    int  noLegalMove = true;
    float dyn = dynEval(b);
#ifdef MYDEBUG
    float v = b->currMove->move.val + dyn;
    v=v;
#endif
    int bmcnt = 0;

    if (ply > 0) {
        if (threat && extThreat) {
            ply += min(extThreat, (extLevel*extThreat)/ply);
            stats.extHorizon++;
        }
    }  
    
    Move *currMove;
    for (currMove = firstMove; currMove < lastMove; currMove++) {
        //if the next ply is greater 0, recursivly call search else
        //call searchq (quiescense)
        float currEval;
        int extension=0;
        if (currMove->attack) {
            stats.extCheck++;
	    extension = min(extCheck, (extLevel*extCheck)/ply);
        }
	//endings, especially pawn endings, positively tend to have less move possibilities.
	//so a increase in search deep comes at less cost. instead of increasing the search deep
	//with a static increment, when the game transites into to a pawn ending, we simply use
	//a smaller ply-decrement during the phase with less move possibilites. this seems to be
	//more logical since the movetree is of size O(movePossibilities^(c*ply), with c usually less than 0.5
	//in this case, half a ply means doubling the search deep.
	//TODO: maybe a dynamic decrement throughout the whole tree is useful, for check evasion and so on.
	//something like decPly = ln(movePossibilities)/ln(40)
	//maybe make the incPly at the root search smaller in this cases
        if (lastMove - firstMove <= 10)
	    extension = max(extension, PLY/2);
	
	int newPly = ply - PLY + extension;
 	
	
        if (newPly > 0 && b->currMove < b->lastMove) {
            stats.moves++;
            doMove(b, currMove);
             if (inCheckSig(b)) {
                undoMove(b);
                continue;
            }
#ifdef MYDEBUG	
            if (flags.debug & DEBUG_SEARCH) {
                for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
                debug("doMove %s (%+4.2f)\n", moveToStr(currMove), currMove->val);
            }    
#endif            
            noLegalMove = 0;    //at least one move was possible, so stale or mate is impossible
            b->color = -b->color;
#ifdef SMP
#define SMPPLY (2*PLY)
            int thread;
            if (newPly > SMPPLY && bestEval > MATE_VALUE && currMove < lastMove-1 && children < MAXCHILDREN && (thread = getFreeThread()) >= 0) {
                child[children].move = *currMove;
                child[children].thread = thread;
                threadSearch(thread, &child[children].val, &child[children].finished, b, maxNullMove, maxExtensions - 1, -beta, -bestAlpha, newPly);
                children++;
                b->color = -b->color;
                undoMove(b);
                continue;
            } else {
#endif
//                if (newPly > PLY)
//                    oldEval = -search(b, maxNullMove, -bestBeta, -bestAlpha, newPly-PLY);
                currEval = -search(b, maxNullMove, verifyNull, -bestBeta, -bestAlpha, newPly);
#ifdef MYDEBUG
    //evaluation is the same, but value of previous best move has changed to negative
/*    int p;
    p = b->currMove->captured;
    if (newPly > PLY) {
        if ( currEval > oldEval + 1.0 && currEval > bestEval && oldEval <= bestEval ) {
            moveord[ply >> 8][7]++;
            if (p)
                moveord[ply >> 8][6]++;
        } else
        if ( currEval < oldEval - 1.0 && currEval < beta && oldEval >= beta ) {
            moveord[ply >> 8][5]++;
            if (p)
                moveord[ply >> 8][4]++;
        } else {
            moveord[ply >> 8][9]++;
            if (p)
                moveord[ply >> 8][8]++;
        }        

    }    */
//    if (bmcnt < 6)
//    moveord[ply >> 8][bmcnt]++;
#endif
#ifdef ASPIRATION
                if (currEval >= bestBeta && currEval < beta)
                    currEval = -search(b, maxNullMove, verifyNull, -beta, -currEval, newPly);
#endif
#ifdef SMP
            }
#endif            
            b->color = -b->color;
#ifdef MYDEBUG	
            if (flags.debug & DEBUG_SEARCH) {
                for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
                debug("undoMove %s\n", moveToStr(currMove));
            }    
#endif
            undoMove(b);
        } else {

            //since quiescense search assumes there is at least one move which is at
            //least as good as the current position, it will always return a value
            //which is less or equal the value of the current positon. so there is
            //no need to actually execute the move if its position value is known to be
            //less than the best value so far.
            noLegalMove = 0;    //this is the last node before searchq, so we cant make a correct decision here
            currEval = (currMove->val + dyn) * b->color - SCORE_TOMOVE + MAX_DYNEVAL;
            if (currEval > bestEval) {
                stats.moves++;
                doMove(b, currMove);
                if (IN_CHECK(b->color)) {
                    undoMove(b);
                    continue;
                }
#ifdef MYDEBUG	
                if (flags.debug & DEBUG_SEARCH) {
                    for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
                    debug("doMove %s (%+4.2f)\n", moveToStr(currMove), currMove->val);
                }    
#endif                
                b->color = -b->color;
                currEval = -searchq(b, -bestBeta, -bestAlpha, newPly);
#ifdef ASPIRATION
                if (currEval >= bestBeta && currEval < beta)
                    currEval = -searchq(b, -beta, -currEval);
#endif
                b->color = -b->color;
#ifdef MYDEBUG	
                if (flags.debug & DEBUG_SEARCH) {
                    for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
                    debug("undoMove %s\n", moveToStr(currMove));
                }    
#endif                
                undoMove(b);
            }
        }
        if (stopSearch) break;
        if (currEval > bestEval) {
            bmcnt++;
            //TODO: in case of bmcnt > 1 check if a presorting again without the first mvoe
            //improves move ordering
            //the presorted move, found in a shallower search (oldEval), sucks now (bestEval).
            //if the new move found (currEval) brings us up again, chances are, it is a
            //shift-behind-the-search-horizon kind of move, so consider it for extension
/*            if (ply > 2 && bmcnt > 1 && bestEval < oldEval - 0.5 && currEval >= beta)
                if (preextension < PLY && extHorizon && ply > 0) {
                    stats.extHorizon++;
                    preextension += extHorizon;
                    if (preextension > PLY)
                        preextension = PLY;
                    goto again;
                }*/    
            bestMove = *currMove;
            bestEval = currEval;
            if (currEval >= beta) break;   //exit move loop
            if (currEval > bestAlpha) 
                bestAlpha = currEval;
                
#ifdef NEGAINT

            bestBeta = bestAlpha + QUANT;
#endif

        }
    }                           //move loop end
#ifdef SMP
    //we need to wait for every little child before we can continue
    //if we got a beta curoff, signal to all childs they can stop.
    while (children) {
        --children;
        while (!child[children].finished)
            ;
        if (child[children].val > bestEval) {
            int i;
            bestMove = child[children].move;
            bestEval = child[children].val;
        }
    }
#endif

    if (noLegalMove) {
        alpha = MATE_VALUE;
        beta = -MATE_VALUE;
        if (check)
            bestEval = MATE_VALUE;      // + currDepth; //matt
        else
            bestEval = drawValue;       //patt
    }
    if (bestEval > INF)
        bestEval -= 1.0;
#ifdef TACTIC_CHECK

    if (lastEval != 4711) {
        float diff;

        if (lastEval > bestBeta)
            lastEval = bestBeta;

        if (bestEval > bestBeta)
            diff = bestBeta - lastEval;
        else
            diff = bestEval - lastEval;

        if (diff > 75) {
            char s[256];
            convert2EPD(s);
        }
    }
#endif

    if (!stopSearch)
        writeHash((constBoard *)b, ply, bestEval, alpha, beta, &bestMove, noLegalMove);
    b->repTable[repIdx] = 0;    //delete position form repetition table
    //   if (ply > PLY_EVAL) eval();
#ifdef MYDEBUG	
    if (flags.debug & DEBUG_SEARCH) {
        for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
        debug("return %+4.2f\n", bestEval);
    }    
#endif    
    END_PROFILE();
    return bestEval;
}

//quiescence search.
//searches only non-quiet moves like captures and promotions, until a quiet pos
//is reached (hence the name). it is assumed there is always a move which at
//least scores equal or better than the current position, so from the view
//of a caller of opposite color it always returns a value less or equal
//the current postion, which allows some optimizations.
float searchq(Board * b, float alpha, float beta, int ply) {
    BEGIN_PROFILE(TIME_SEARCHQ);
    float bestEval;
    float dyn = dynEval(b);
    {
        float v = b->color * (b->currMove->move.val + dyn) + SCORE_TOMOVE;
#ifdef MYDEBUG        
        if (flags.debug & DEBUG_SEARCH) {
            for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
            debug("searchq:a=%+5.2f, b=%+5.2f, c=%+5.2f\n", alpha, beta, v);
        }
#endif     
        if (v >= beta) {
#ifdef MYDEBUG	
                if (flags.debug & DEBUG_SEARCH) {
                    for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
                    debug("return (stand pat pruned) %+5.2f\n", v);
                }    
#endif                
            END_PROFILE();
            return v;
        }                       
        bestEval = alpha;
	if (v > alpha) {
            bestEval = v; 
//	    alpha = v;
	}
    }
    float bestBeta = beta;

#ifdef USE_QHASH
    //TODO
    //maybe shift this block and the write block behind the movegen, since the movegen is pretty fast
    MainEntry *p = readHash(b);
    if (p) {
        switch (p->bound) {
            case EXACT_VALUE:
                END_PROFILE();
                return HASH_VALUE(p);
            case LOWER_BOUND:
                if (HASH_VALUE(p) >= bestBeta) {
                    END_PROFILE();
                    return HASH_VALUE(p);
                }
                if (HASH_VALUE(p) > bestEval)
                    bestEval = HASH_VALUE(p);
                stats.alpha++;
                break;
            case UPPER_BOUND:
                if (HASH_VALUE(p) <= bestEval) {
                    END_PROFILE();
                    return HASH_VALUE(p);
                }
                if (HASH_VALUE(p) < bestBeta)
                    bestBeta = HASH_VALUE(p);
                stats.beta++;
                break;
        }
    }
#endif
    Move firstMove[MAXMOVES];
    Move *lastMove = movegenq((constBoard *)b, firstMove, alpha);
    Move bestMove;
    bestMove.src = bestMove.dst = 0;
    
#ifdef MYDEBUG
    int bmcnt=0;
#endif
    if (lastMove != firstMove && b->currMove < b->lastMove) {
        stats.qPos += lastMove - firstMove;
#ifdef USE_QHASH
        BEGIN_PROFILE(TIME_QSORT);
        sort(firstMove, lastMove-1, b->color);
        END_PROFILE();
#if 0        
        if (lastMove - firstMove > 1 && p && p->src | p->dst)
            for (Move *currMove = firstMove + 1; currMove < lastMove; currMove++)
                if (currMove->src == p->src << 2 && currMove->dst == p->dst << 2 && currMove->special = p->special) {
                    bestMove = *currMove;
                    *currMove = *firstMove;
                    *firstMove = bestMove;
                    break;
                }
#endif
#endif
        for (Move *currMove = firstMove; currMove < lastMove; currMove++) {
        //TODO: if the move to be done is known not to change dynEval, limit
        //maxdyneval to 0.
            float currEval = b->color * (currMove->val + dyn) - SCORE_TOMOVE + MAX_DYNEVAL;
            if (currEval <= bestEval/*&& (!currMove->attack || ply < -2*PLY)*/) {
#ifdef MYDEBUG	
                if (flags.debug & DEBUG_SEARCH) {
                   for (FullMove *i = b->startMove-1; i<b->currMove; i++) debug("    ");
                    debug("Move %s (%+4.2f) futile\n", moveToStr(currMove), currEval);
                }    
#endif                
                continue;
            }
#ifdef MYDEBUG	
            if (flags.debug & DEBUG_SEARCH) {
                for (FullMove *i = b->startMove-1; i<b->currMove; i++) debug("    ");
                debug("doMove %s (%+4.2f > %+4.2f, %d)", moveToStr(currMove), currEval, bestEval, currMove->attack);
            }    
#endif
            doMove(b, currMove);
            stats.qMoves++;
            if (IN_CHECK(b->color)) {
#ifdef MYDEBUG	
		if (flags.debug & DEBUG_SEARCH) 
		    debug(" -> illegal\n");
#endif
                undoMove(b);
                continue;
            }
#ifdef MYDEBUG	
	    debug("\n");
#endif
            b->color = -b->color;
/*            if (currMove->attack && ply >= -2*PLY)
                currEval = -search(b, 0, -bestBeta, -bestEval, ply-PLY);
            else    
*/                currEval = -searchq(b, -bestBeta, -bestEval, ply-PLY);
            b->color = -b->color;
#ifdef MYDEBUG	
            if (flags.debug & DEBUG_SEARCH) {
                for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
                debug("undoMove %s\n", moveToStr(currMove));
            }    
#endif
            undoMove(b);

            if (currEval > bestEval) {
//#define MYDEBUG            
#ifdef MYDEBUG
                bmcnt++;
#endif

                bestEval = currEval;
                bestMove = *currMove;
                if (bestEval >= bestBeta) {
                    if (bestEval > INF)
                        bestEval -= 1;
                    break;
                }
            }
        }
    }    
                          //if lastmove != firstmove
#ifdef USE_QHASH
    writeHash((constBoard *)b, 0, bestEval, alpha, beta, &bestMove, 0);
#endif    
#ifdef MYDEBUG	
    moveord[0][bmcnt]++;
    if (flags.debug & DEBUG_SEARCH) {
        for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
        debug("return %+4.2f\n", bestEval);
    }    
#endif	
    END_PROFILE();
    return bestEval;
}

//quiescence search.
//searches only non-quiet moves like captures and promotions, until a quiet pos
//is reached (hence the name). it is assumed there is always a move which at
//least scores equal or better than the current position, so from the view
//of a caller of opposite color it always returns a value less or equal
//the current postion, which allows some optimizations.
float searchq1(Board * b, float alpha, float beta, int ply) {
    BEGIN_PROFILE(TIME_SEARCHQ);
    float bestEval;
    float dyn = dynEval(b);
    {
        float v = b->color * (b->currMove->move.val + dyn) + SCORE_TOMOVE;
#ifdef MYDEBUG        
        if (flags.debug & DEBUG_SEARCH) {
            for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
            debug("searchq1:a=%+5.2f, b=%+5.2f, c=%+5.2f\n", alpha, beta, v);
        }
#endif     
        if (v >= beta) {
#ifdef MYDEBUG	
                if (flags.debug & DEBUG_SEARCH) {
                    for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
                    debug("return (stand pat pruned) %+5.2f\n", v);
                }    
#endif                
            END_PROFILE();
            return v;
        }                       
        bestEval = alpha;
        if (v > alpha) 
            bestEval = v; 
    }
    float bestBeta = beta;

#ifdef USE_QHASH
    //TODO
    //maybe shift this block and the write block behind the movegen, since the movegen is pretty fast
    MainEntry *p = readHash(b);
    if (p) {
        switch (p->bound) {
            case EXACT_VALUE:
                END_PROFILE();
                return HASH_VALUE(p);
            case LOWER_BOUND:
                if (HASH_VALUE(p) >= bestBeta) {
                    END_PROFILE();
                    return HASH_VALUE(p);
                }
                if (HASH_VALUE(p) > bestEval)
                    bestEval = HASH_VALUE(p);
                stats.alpha++;
                break;
            case UPPER_BOUND:
                if (HASH_VALUE(p) <= bestEval) {
                    END_PROFILE();
                    return HASH_VALUE(p);
                }
                if (HASH_VALUE(p) < bestBeta)
                    bestBeta = HASH_VALUE(p);
                stats.beta++;
                break;
        }
    }
#endif
    Move firstMove[MAXMOVES];
    Move *lastMove = movegenq1((constBoard *)b, firstMove, alpha);
    Move bestMove;
    bestMove.src = bestMove.dst = 0;
    
#ifdef MYDEBUG
    int bmcnt=0;
#endif
    if (lastMove != firstMove && b->currMove < b->lastMove) {
        stats.qPos += lastMove - firstMove;
#ifdef USE_QHASH
        BEGIN_PROFILE(TIME_QSORT);
        sort(firstMove, lastMove-1, b->color);
        END_PROFILE();
#if 0        
        if (lastMove - firstMove > 1 && p && p->src | p->dst)
            for (Move *currMove = firstMove + 1; currMove < lastMove; currMove++)
#ifdef SMALLHASH                
                if (currMove->src == ADR(p->src >> 3, p->src & 7) && currMove->dst == ADR(p->dst >> 3, p->dst & 7)) {
#else                
                if (currMove->src == p->src << 2 && currMove->dst == p->dst << 2) {
#endif                    
                    bestMove = *currMove;
                    *currMove = *firstMove;
                    *firstMove = bestMove;
                    break;
                }
#endif
#endif
        for (Move *currMove = firstMove; currMove < lastMove; currMove++) {
        //TODO: if the move to be done is known not to change dynEval, limit
        //maxdyneval to 0.
            float currEval = b->color * (currMove->val + dyn) - SCORE_TOMOVE + MAX_DYNEVAL;
            if (currEval <= bestEval && !currMove->attack) {
#ifdef MYDEBUG	
                if (flags.debug & DEBUG_SEARCH) {
                   for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
                    debug("Move %s (%+4.2f) futile\n", moveToStr(currMove), currEval);
                }    
#endif                
                continue;
            }
            doMove(b, currMove);
            stats.qMoves++;
            if (IN_CHECK(b->color)) {
                undoMove(b);
                continue;
            }
#ifdef MYDEBUG
            if (flags.debug & DEBUG_SEARCH) {
                for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
                debug("doMove %s (%+4.2f > %+4.2f, %d)\n", moveToStr(currMove), currEval, bestEval, currMove->attack);
            }    
#endif
            b->color = -b->color;
            if (currMove->attack && inMate(b)) {
                bestEval = -MATE_VALUE-1;
                writeHash((constBoard *)b, 0, bestEval, alpha, beta, currMove, true);
                END_PROFILE();
                return bestEval;
            } else        
                currEval = -searchq1(b, -bestBeta, -bestEval, ply-PLY);
            b->color = -b->color;
#ifdef MYDEBUG	
            if (flags.debug & DEBUG_SEARCH) {
                for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
                debug("undoMove %s\n", moveToStr(currMove));
            }    
#endif
            undoMove(b);

            if (currEval > bestEval) {
//#define MYDEBUG            
#ifdef MYDEBUG
                bmcnt++;
#endif

                bestEval = currEval;
                bestMove = *currMove;
                if (bestEval >= bestBeta) {
                    if (bestEval > INF)
                        bestEval -= 1;
                    break;
                }
            }
        }
    }    
                          //if lastmove != firstmove
#ifdef USE_QHASH
    writeHash((constBoard *)b, 0, bestEval, alpha, beta, &bestMove, 0);
#endif    
#ifdef MYDEBUG	
    moveord[0][bmcnt]++;
    if (flags.debug & DEBUG_SEARCH) {
        for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
        debug("return %+4.2f\n", bestEval);
    }    
#endif	
    END_PROFILE();
    return bestEval;
}

//root search

void searchm(int endply) {
    Move firstMove[MAXMOVES];
    Move *lastMove, *currMove, *tempMove, bestMove;
    
    float currEval, alpha, beta, bestAlpha, bestBeta, bestEval;
    static Board *const b = &board;

    if (flags.log) {
        if (!logfile) {
            char logfilename[256];
            int logfilenum = -1;
            do {
                logfilenum++;
                sprintf(logfilename, "slibo%04d.log", logfilenum);
                logfile = fopen(logfilename, "r");
                if (logfile)
                    fclose(logfile);
            } while (logfile);
            logfile = fopen(logfilename, "w");
        }
        printBoard(b, logfile);
    }

    int plyhold = 0;
    solutionTime = 0;
    stopSearch = false;
    returnMove.src = ponderMove.src = returnMove.dst = ponderMove.dst = 0;
    stats.alpha = stats.beta = stats.BM = stats.pos = stats.qPos = stats.moves = stats.qMoves = 0;
    board.startMove = board.currMove;

    if (IN_CHECK(-b->color))
        return;
    if (!flags.nobook) {
        lastMove = bookMoves(firstMove);
        if (lastMove > firstMove) {
            bestEval = -INF;
            for (currMove = firstMove; currMove < lastMove; currMove++)
                if (currMove->val > bestEval)
                    bestEval = currMove->val;
            do {
                returnMove = firstMove[rand() % (lastMove - firstMove)];
            } while (returnMove.val < bestEval - bookLimit);
            return;
        }
    }

    //TODO: generate smaller REPTABLE (50 moves + MAXDEPTH plys) here
    eval(b);
    board.bsig = 1;
    board.wsig = 1;
    lastMove = movegen((constBoard *)b, firstMove);
    stats.pos += lastMove - firstMove;

    sort(firstMove, lastMove - 1, b->color);

    beta = INF;
    alpha = -INF;
    stats.startTime = searchTime();

    purgeHash();       //this may take a while, for benchmarking we take the time
    preloadHash();
    stats.bTime = searchTime();    //here

    bestAlpha = -INF;
    bestBeta = INF;
#ifdef MYDEBUG

    memset(moveord, 0, sizeof(moveord));
#endif
    
    for (stats.ply = 2*PLY; stats.ply <= endply; stats.ply+=PLY) {
//search depth limit.
//don't limit to low values, the short quiescense searches may poison the cache!
        b->lastMove = b->currMove + MAXDEPTH / PLY;
        stats.eval = MATE_VALUE;
#ifdef MYDEBUG
        if (flags.debug & DEBUG_SEARCH)
            debug("search: a=%+5.2f, b=%+5.2f, ply=%2d\n", bestAlpha, bestBeta, stats.ply/PLY);
#endif
        // main move loop

        for (currMove = firstMove; currMove < lastMove; currMove++) {
            stats.currMoveNum = currMove - firstMove;
            doMove(b, currMove);
            if (IN_CHECK(b->color)) {
                undoMove(b);
                for (tempMove = currMove + 1; tempMove < lastMove; tempMove++)  //remove illegal move from list
                    tempMove[-1] = tempMove[0];
                lastMove--;
                currMove--;
                continue;
            }
#ifdef MYDEBUG
            if (flags.debug & DEBUG_SEARCH) {
                for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
                debug("doMove %s (%+4.2f)\n", moveToStr(currMove), currMove->val);
            }    
#endif            
            b->color = -b->color;
            do {
                currEval = -search(&board, maxNullMove, 1, -bestBeta, -bestAlpha, stats.ply - PLY);
#ifdef ASPIRATION

                if (currEval >= bestBeta && currEval < beta) {
                    bestBeta = beta;
                    continue;
                }
                if (currEval <= bestAlpha && currEval > alpha && currMove == firstMove) {
                    bestAlpha = alpha;
                    continue;
                }
#endif

            } while (0);
            b->color = -b->color;
#ifdef MYDEBUG
            if (flags.debug & DEBUG_SEARCH) {
                for (FullMove *i = b->startMove; i<b->currMove; i++) debug("    ");
                debug("undoMove %s\n", moveToStr(currMove));
            }    
#endif            
            undoMove(b);
            if (currEval > stats.eval && !stopSearch) {
                returnMove = *currMove;
                stats.bestMove = *currMove;
                stats.eval = currEval;
                if (stats.pos > flags.minPos )
                    postLines();
                bestMove = *currMove;
                for (Move *i = currMove-1; i>=firstMove; i--)
                    i[1] = i[0];
                *firstMove = bestMove;

                if (breakMove.src == stats.bestMove.src && breakMove.dst == stats.bestMove.dst) {
                    if (solutionTime == 0) {
                        solutionMoves = stats.qMoves + stats.moves;
                        solutionTime = searchTime() - stats.startTime;
                    } else if (++plyhold > 0 && searchTime() - stats.startTime > 10 S)
                        stopSearch = true;
                    if (searchTime() - stats.startTime > t1 / 2)
                        stopSearch = true;
                } else {
                    solutionTime = 0;
                    solutionMoves = 0;
                    plyhold = 0;
                }
                if (currEval >= beta && stats.ply > 2*PLY)
                    break;

                if (currEval > bestAlpha)
                    bestAlpha = currEval;
#ifdef ASPIRATION

                bestBeta = currEval + 8*QUANT;
#endif

            }
	    flushHash();
            //          list->setText(0, SliboInfo::move2qstring(*currMove) + QString().setNum(currEval, 'f', 2));
            if (stats.eval > stats.lastEval - 0.25 && searchTime() > stats.startTime + t0 && currMove > firstMove && t0) {
                stopSearch = true;
                break;
            }
        }                       // end move loop

        stats.moveCount = lastMove - firstMove;

        if (abs(stats.eval) > INF || stopSearch || stats.moveCount <= 1)
            break;

        stats.lastEval = stats.eval;
#ifdef ASPIRATION
        bestAlpha = stats.eval[stats.ply] - 8*QUANT;
        bestBeta = stats.eval[stats.ply] + 8*QUANT;
#else
        bestAlpha = -INF;
        bestBeta = INF;
#endif

#ifdef COMBOS

        printf("                -400      -200       -50        50       200       400\n");
        for (i = 2; i <= 20; i++) {
            printf("%2d-%2d ply ", i - 1, i);
            for (j = 0; j < 7; j++)
                printf("%10d", changes[i][j]);
            printf("\n");
        }
#endif

    }                           // end iteration loop
    if (stats.ply > endply)
        stats.ply--;
    postLines();
}
