/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "smp.h"
#include "search.h"
#include <pthread.h>

#ifdef SMP
struct Param {
    Board board __attribute__ ((aligned (64)));
    volatile float *ret;
    volatile int *finished;
    int nullMove;
    int maxExtensions;
    float alpha;
    float beta;
    int ply;
    Line *prevLine;
} param[THREADS];

int idleList[THREADS];
int idleCnt = 0;

pthread_cond_t waitCond[THREADS] = { [0 ... THREADS-1] = PTHREAD_COND_INITIALIZER};
pthread_mutex_t waitMutex[THREADS] = { [0 ... THREADS-1] = PTHREAD_MUTEX_INITIALIZER};
pthread_mutex_t idleListMutex = PTHREAD_MUTEX_INITIALIZER;

static void threadWaitLoop(int thread) {
    pthread_mutex_lock(&waitMutex[thread]);
    for (;;) {
        pthread_cond_wait(&waitCond[thread], &waitMutex[thread]);
        *param[thread].ret = -search(&param[thread].board, param[thread].nullMove, param[thread].maxExtensions, param[thread].alpha, param[thread].beta, param[thread].ply, param[thread].prevLine);
        *param[thread].finished = 1;    
        pthread_mutex_lock(&idleListMutex);
        idleList[idleCnt++] = thread;
        pthread_mutex_unlock(&idleListMutex);
    }
}

void initThreads() {
    pthread_t th;
    for (idleCnt=0; idleCnt<THREADS; idleCnt++) {
        idleList[idleCnt] = idleCnt;
        pthread_create(&th, 0, (void *(*)(void *))threadWaitLoop, (void *)(idleCnt+1));  
    }
}

int getFreeThread() {
    pthread_mutex_lock(&idleListMutex);
    int ret = -1;
    if (idleCnt > 0)
        ret = idleList[--idleCnt];
    pthread_mutex_unlock(&idleListMutex);
    return ret;    
}

void threadSearch(int thread, float *ret, volatile int *finished, Board *b, int nullMove, int maxExtensions, float alpha, float beta, int ply, Line * prevLine) {
    param[thread].ret = ret;
    param[thread].finished = finished;
    //we need a deep copy of the board for the new child.
    //TODO: copy only the really used values, initialize the static values beforehand
    param[thread].board = *b;
    param[thread].nullMove = nullMove;
    param[thread].maxExtensions = maxExtensions;
    param[thread].alpha = alpha;
    param[thread].beta = beta;
    param[thread].ply = ply;
    param[thread].prevLine = prevLine;
    *param[thread].finished = 0;
    pthread_mutex_lock(&waitMutex[thread]);
    pthread_mutex_unlock(&waitMutex[thread]);
    pthread_cond_broadcast(&waitCond[thread]);
}
 
void threadWait(int thread) {
    //TODO:
    //need semaphore here
    while (!*param[thread].finished) ; 
}

static inline fastcpy(int *d, const int *s) {
    for (int y=0; y<8; y++) {
        d[0] = s[0];
        d[1] = s[1];
        d[2] = s[2];
        d[3] = s[3];
        d[4] = s[4];
        d[5] = s[5];
        d[6] = s[6];
        d[7] = s[7];
        s += 16;
        d += 16;
    }    
}    

void fastCopyBoard(Board *a, const Board *b) {
    fastcpy(&a->board[2][2], &b->board[2][2]);
    fastcpy(&a->attack[2][2], &b->attack[2][2]);
    for (int p=0; p<13; p++) {
        if (p==6) continue;
    //TODO: build new attack sigs instead of copying, proably cheaper
        fastcpy(&a->pext[p][2][2], &b->pext[p][2][2]);
        d = &a->pieces[p].cnt;
        s = &b->pieces[p].cnt;
        int cnt = *s;
        for (int i=0; i<=cnt; i++)
            *d++ = *s++;
    }
    memcpy(a->repTable, b->repTable, REPTABLESIZE * sizeof(unsigned long long));
    Move *d = a->moveList;
    for (Move *s = b->moveList; s <= b->currMove; s++, d++)
        *d = *s;
    d--;
    a->currMove = d;
    a->startMove = d;
    a->lastMove = a->moveList + (b->lastMove - b->moveList);

    a->epstatus = b->epstatus;         
    a->o_o = b->o_o;
    a->color = b->color;
    a->fiftyMoves = b->fiftyMoves;

    a->key = b->key;

    a->kingSafety = b->kingSafety;
    a->wsig = b->wsig;
    a->bsig = b->bsig;
}
#endif
