/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <ncurses.h>
#include <ctype.h>
#include "hashfunc.h"
#include "util.h"
#include "search.h"
#include "opening.h"
#include "debug.h"
#include "movegen_c.h"
#include "eval.h"

#define ILLEGAL -47110.0
#define UNKNOWN -12000.0

Move moveNode = { 0.0, 0, 0, 0, 0, -1, 0 };
int cntNode;
int selNode = 0;

struct Stats stats;

piece_t visibleBoard[12][16];

void printBoard(const Board *b, FILE *f) {
    fprintf(f,"\n       +---+---+---+---+---+---+---+---+\n");
    for(int i=7;i>=0;i--) {
        fprintf(f,"   %2d  ",i+1);
        for(int j=0;j<8;j++)
            fprintf(f,"| %c ","kpnqbr RBQNPK"[TYPE(i,j)+6]);
        fprintf(f,"|\n");
        fprintf(f,"       +---+---+---+---+---+---+---+---+\n");
    }
    fprintf(f,"         a   b   c   d   e   f   g   h\n\n");
    char buf[256];
    convert2EPD(b, buf);
    fprintf(f, "%s", buf);
    fprintf(f, "\n\n");
}

void copyBoard()
{
    static Board *const b = &board;
    eval(b);
    memcpy(visibleBoard, b->board, sizeof(b->board));
    eList = movegen((constBoard *)b, mList);
#ifdef MYDEBUG
    if (flags.debug & DEBUG_INFO)
        debug("inf: copyboard, generating moves for %d\n", b->color);
#endif    
	b->startMove = b->currMove;
    selNode = 0;
    if (nodeWnd && !flags.xboard)
        printNode(nodeWnd);
}

static char *value2str(float v)
{
    static char s[256];
    if (v < -INF)
        sprintf(s, " -M%02.0f", v - MATE_VALUE);
    else if (v > INF)
        sprintf(s, " +M%02.0f", 1 - v - MATE_VALUE);
    else if (v == UNKNOWN)
        sprintf(s, "unknwn");
    else
        sprintf(s, "%+2.2f", v);
    return s;
}

extern int collision;
void printStats(WINDOW * wnd)
{
    mvwprintw(wnd, 0, 0, "Pos:   %12.0f", stats.pos * 0.001);
    mvwprintw(wnd, 1, 0, "qPos:  %12.0f", stats.qPos * 0.001);
    mvwprintw(wnd, 2,0, "moves: %12.0f", stats.moves * 0.001);
    mvwprintw(wnd, 3,0, "qMoves:%12.0f", stats.qMoves * 0.001);
//    mvwprintw(wnd, 2, 0, "%8.0f %8.0f %8.0f %8.0f", stats.pos[1] * 0.001, stats.pos[2] * 0.001, stats.pos[3] * 0.001, stats.pos[4] * 0.001);
//    mvwprintw(wnd, 3, 0, "%8.0f %8.0f %8.0f %8.0f", stats.qPos[1] * 0.001, stats.qPos[2] * 0.001, stats.qPos[3] * 0.001, stats.qPos[4] * 0.001);
    mvwprintw(wnd, 4, 0, "mRead: %12.0f", mainHash.read * 0.001);
    mvwprintw(wnd, 5, 0, "mHit:  %12.0f", mainHash.hit * 0.001);
    mvwprintw(wnd, 6, 0, "mWrite:%12.0f", mainHash.write * 0.001);
    mvwprintw(wnd, 7, 0, "mFill:%6d/%6d", mainHash.fill / 1024, mainHash.count / 1024);
    mvwprintw(wnd, 8, 0, "alpha: %12.0f", stats.alpha * 0.001);
    mvwprintw(wnd, 9, 0, "beta:  %12.0f", stats.beta * 0.001);
    mvwprintw(wnd, 10, 0, "aret:  %12.0f", stats.aret * 0.001);
    mvwprintw(wnd, 11, 0, "bret:  %12.0f", stats.bret * 0.001);
    mvwprintw(wnd, 12, 0, "ret:   %12.0f", stats.ret * 0.001);
    mvwprintw(wnd, 13, 0, "preCut:%12.0f", stats.precut * 0.001);
    mvwprintw(wnd, 14, 0, "bmove: %12.0f", stats.BM * 0.001);
    mvwprintw(wnd, 15, 0, "hit2:  %12.0f", stats.hit2 * 0.001);
    mvwprintw(wnd, 16, 0, "shift2:%12.0f", stats.shift2 * 0.001);
    mvwprintw(wnd, 17, 0, "write2:%12.0f", stats.write2 * 0.001);

    mvwprintw(wnd, 0, 20, "/s:    %12.0f", stats.pos * 1000.0 / (searchTime() - stats.bTime));
    mvwprintw(wnd, 1, 20, "/s:    %12.0f", stats.qPos * 1000.0 / (searchTime() - stats.bTime));
    mvwprintw(wnd, 2,20, "/s:    %12.0f", stats.moves*1000.0/(searchTime()-stats.startTime));
    mvwprintw(wnd, 3,20, "/s:    %12.0f", stats.qMoves*1000.0/(searchTime()-stats.startTime));
//extern unsigned long long c[];    
//    mvwprintw(wnd, 0, 20, "c0:    %12.0f", c[0]*0.001);
//    mvwprintw(wnd, 1, 20, "c1:    %12.0f", c[1]*0.001);
//    mvwprintw(wnd, 2,20, "c2:    %12.0f", c[2]*0.001);
//    mvwprintw(wnd, 3,20, "c3:    %12.0f", c[3]*0.001);
    mvwprintw(wnd, 4, 20, "pRead: %12.0f", pawnHash.read * 0.001);
    if (pawnHash.read) 
    mvwprintw(wnd, 5, 20, "pWrite %11.3f\%%", 100.0*pawnHash.write/pawnHash.read);
    mvwprintw(wnd, 6, 20, "fRead: %12.0f", (blackHash.read + whiteHash.read) * 0.001);
    if (whiteHash.read + blackHash.read)
    mvwprintw(wnd, 7, 20, "fWrite %11.3f\%%", 100.0*(blackHash.write + whiteHash.write)/(blackHash.read + whiteHash.read));
    mvwprintw(wnd, 8, 20, "extM:  %12.0f", stats.extMate * 0.001);
    mvwprintw(wnd, 9, 20, "extC:  %12.0f", stats.extCheck * 0.001);
    mvwprintw(wnd, 10, 20, "extH:  %12.0f", stats.extHorizon * 0.001);
    mvwprintw(wnd, 11, 20, "extP:  %12.0f", stats.extPawn * 0.001);
    mvwprintw(wnd, 12, 20, "extR:  %12.0f", stats.extRecapture * 0.001);
    mvwprintw(wnd, 13, 20, "zz:    %12.0f", stats.zz * 0.001);
    mvwprintw(wnd, 14, 20, "null:  %12.0f", stats.null * 0.001);
    mvwprintw(wnd, 15, 20, "nultry:%12.0f", stats.nulltry * 0.001);
    mvwprintw(wnd, 16, 20, "ply:          %5.2f", (double)stats.ply/PLY);
    mvwprintw(wnd, 17, 20, "move:       %3d/%3d", stats.currMoveNum + 1, stats.moveCount);
 
#ifdef MYDEBUG
    for (int j=0; j<10; j++) {
//        unsigned long long total = 0;
//        for (int i=1; i<6; i++) 
//            total += moveord[j][i];
//        if (total) {
            for (int i=0; i<5; i+=2)
                if (moveord[j][i+5])
                    mvwprintw(wnd, 18+j, i*6, "%5.2f %5.2f ",  
            100.0*moveord[j][i+4]/moveord[j][i+5], log10(moveord[j][i+5]));
////                if (1000*moveord[j][i]/total)
//                    mvwprintw(wnd, 18+j, i*5, "%4d ", 1000*moveord[j][i]/total);
                else
                    mvwprintw(wnd, 18+j, i*12, "            ");
//            mvwprintw(wnd, 18+j, 49, "%10d ", total);
        }
//        mvwprintw(wnd, 18+j, 20, "%5.2f %5.2f %5.2f %5.2f %5.2f %5.2f",
//            100.0*moveord[j][6]/(moveord[j][7]+0.000001), log10(moveord[j][7]),
//            100.0*moveord[j][8]/(moveord[j][9]+0.000001), log10(moveord[j][9])
//            );
#endif
    mvwprintw(wnd, 0, 40, "line:   ");
    int i=0;
    for (FullMove *m = board.startMove+1; i<30; i++) 
        if ( m <= board.currMove ) {
            mvwprintw(wnd, i/2+1, (i&1)*10+40, "%8s", moveToStr(&m->move));
            m++;
            if (m[0].move.piece * m[-1].move.piece > 0) {
                i++;
                mvwprintw(wnd, i/2+1, (i&1)*10+40, "        ");                   
            }    
        } else
            mvwprintw(wnd, i/2+1, (i&1)*10+40, "        ");
        
}

void postLines()
{
    char buf[MAXDEPTH * 8];

    line2str(&stats.bestMove, buf);
    if (flags.xboard && flags.post) 
        printf("%2d %5.0f %8.0f %12lld %s\n", stats.ply/PLY, stats.eval * 100.0, (searchTime() - stats.startTime) * 0.000001, stats.moves + stats.qMoves, buf);
    if (!flags.xboard) 
        wprintw(lineWnd, "%4.1f %s %8.1f %s\n", (double)stats.ply/PLY, value2str(stats.eval), (searchTime() - stats.startTime) * 0.000001, buf);
    extern FILE *logfile;
    if (flags.log && logfile) 
        fprintf(logfile,  "%4.1f %s %8.1f %s\n", (double)stats.ply/PLY, value2str(stats.eval), (searchTime() - stats.startTime) * 0.000001, buf); 
}

void printMoveList()
{
    char s[256];
    int linecnt = min((board.startMove - board.moveList + 3) / 2, LINES - 20);
    FullMove *m = max(board.startMove - 2 * linecnt + ((board.startMove-board.moveList) & 1), (FullMove *)board.moveList);
    for (int l = 0; l < linecnt; l++) {
        if (m < board.startMove) 
            sprintf(s, "%3d. %-9s", 1 + (m-board.moveList) / 2, moveToStr(&m->move));
        else
            sprintf(s, "                    ");
        mvwaddstr(moveWnd, l, 0, s);
        m++;
        
        if (m < board.startMove) 
            sprintf(s, "%3d. %-9s", 1 + (m-board.moveList) / 2, moveToStr(&m->move));
        mvwaddstr(moveWnd, l, 14, s);
        m++;
    }
}

void printAttackBy()
{
/*
    int x, y, col;
    char buf[256], *s;
    for (y = 9; y >= 2; y--) {
        for (x = 2; x < 10; x++) {
            s = buf;
            for (col = 0; col < 6; col++)
                *s++ = ' ';

            s = buf;
            for (col = 0; col < 2; col++) {
                if (attackBy[0][y][x] & (KING_ATTACK << (16 * col)))
                    *s++ = 'K' + 32 * col;
                if (attackBy[0][y][x] & 7 * (QUEEN_ATTACK << (16 * col)))
                    *s++ = 'Q' + 32 * col;
                if (attackBy[0][y][x] & 7 * (ROOK_ATTACK << (16 * col)))
                    *s++ = 'R' + 32 * col;
                if (attackBy[0][y][x] & 7 * (BISHOP_ATTACK << (16 * col)))
                    *s++ = 'B' + 32 * col;
                if (attackBy[0][y][x] & 7 * (KNIGHT_ATTACK << (16 * col)))
                    *s++ = 'N' + 32 * col;
                if (attackBy[0][y][x] & 3 * (PAWN_ATTACK << (16 * col)))
                    *s++ = 'P' + 32 * col;
            }
            buf[6] = 0;

            printf("%s|", buf);
        }
        printf("\n");
    }
*/
}

void upNode()
{
    unPonder();
    if (!stopped()) {
        computerTurn = false;
        stop();
    }
    if (selNode) {
        selNode--;
        printNode(nodeWnd);
    }
}

void downNode()
{
    unPonder();
    if (!stopped()) {
        computerTurn = false;
        stop();
    }
    if (selNode < cntNode - 1) {
        selNode++;
        printNode(nodeWnd);
    }
}

void leftNode()
{
    unPonder();
    if (board.currMove > board.moveList+1) {
        if (!stopped()) {
            computerTurn = false;
            stop();
        }
        deleteRep(&board);
        board.color = -board.color;
        undoMove(&board);
        copyBoard();
    }
}

void rightNode()
{
    if (moveNode.src < 0)
        return;
    unPonder();
    if (!stopped()) {
        computerTurn = false;
        stop();
    }
    doMove(&board, &moveNode);
    board.color = -board.color;
    insertRep(&board);
    copyBoard();
}

void printNode(WINDOW * wnd)
{
    Move firstMove[MAXMOVES], *lastMove, *currMove, m;
    Move firstMove2[MAXMOVES], *lastMove2, *currMove2;
    MainEntry *p;
    int y, i, j;
    char s[256];
    float val[MAXMOVES], v;
    static Board *const b = &board;

    werase(wnd);
    b->color = -b->color;
    inCheckSig(b);
    b->color = -b->color;
    lastMove = movegen((constBoard *)b, firstMove);
    for (currMove = firstMove, y = 0; currMove < lastMove; currMove++, y++) {
        doMove(b, currMove);
        b->color = -b->color;
        if (IN_CHECK(-b->color)) {
            val[y] = ILLEGAL;
        } else if ((p = readHash(b))) {
            val[y] = -HASH_VALUE(p);
            if (p->bound == UPPER_BOUND)
                val[y] += 0x0.0001P0;
            if (p->bound == LOWER_BOUND)
                val[y] -= 0x0.0001P0;
        } else
            val[y] = UNKNOWN;
        b->color = -b->color;
        undoMove(b);
    }

    lastMove2 = bookMoves(firstMove2);
    for (currMove = firstMove, y = 0; currMove < lastMove; currMove++, y++)
        for (currMove2 = firstMove2; currMove2 < lastMove2; currMove2++)
            if (currMove2->src == currMove->src && currMove2->dst == currMove->dst && currMove2->special == currMove->special)
                val[y] = currMove2->val;

    for (i = 1; i < y; i++)
        for (j = 1; j < y; j++)
            if (val[j - 1] < val[j]) {
                m = firstMove[j - 1];
                v = val[j - 1];
                firstMove[j - 1] = firstMove[j];
                val[j - 1] = val[j];
                firstMove[j] = m;
                val[j] = v;
            }

    moveNode.src = -1;
    for (currMove = firstMove, y = 0; currMove < lastMove; currMove++, y++) {
        if (val[y] <= ILLEGAL)
            break;

        doMove(&board, currMove);
        board.color = -board.color;

        strcpy(s, moveToStr(currMove));

        if (y == selNode) {
            wattrset(wnd, A_REVERSE);
            moveNode = *currMove;
        }

        mvwprintw(wnd, y % 17, (y / 17) * 24, "%7s ", s);

        p = readHash(b);
        if (p) {
            wprintw(wnd, "%c%s (%2d) ", HASH_BOUND_STR[p->bound], value2str(val[y]), p->ply/PLY);
            if (p->dst | p->src)
                wprintw(wnd, "%c%c%c%c", X(p->src) + 'a', Y(p->src) + '1', X(p->dst) + 'a', Y(p->dst) + '1');
        } else if (val[y] != UNKNOWN) {
            wprintw(wnd, "book: %s", value2str(val[y]));
        }
        waddstr(wnd, "\n");
        if (y == selNode)
            wattrset(wnd, A_NORMAL);

        board.color = -board.color;
        undoMove(&board);

    }
    cntNode = y;
    wrefresh(wnd);
}

void displayBoard(WINDOW * wnd)
{
    int x, y, col;
    for (y = 7; y >= 0; y--)
        for (x = 0; x < 8; x++) {
            col = visibleBoard[y + 2][x + 2].w >= 0 ? 1 : -1;
            if (wnd) {
                wattrset(wnd, COLOR_PAIR(((x + y) & 1) + col + 4) | (col > 0 ? A_BOLD : 0));
                mvwprintw(wnd, 7 - y, 2 * x, " %c", "kpnqbr RBQNPK"[visibleBoard[y + 2][x + 2].w / 512 + 6]);
            } else {
                printf(" %c", "kpnqbr RBQNPK"[visibleBoard[y + 2][x + 2].w / 512 + 6]);
                if (x == 7)
                    printf("\n");
            }
        }
    if (wnd)
        wattrset(wnd, COLOR_PAIR(1));
}

static int moveToStrX(Board *b, Move *m, char *s) {
    
    Move firstMove[MAXMOVES];
    Move *lastMove = movegen((constBoard *)b, firstMove);
    
    int found = 0;
    for (Move *currMove = firstMove; currMove < lastMove; currMove++)
        if (currMove->src == m->src && currMove->dst == m->dst && currMove->special == m->special) {
            *m = *currMove;
            found = 1;
            break;
        }

    if (!found) 
        return 0;
    
    doMove(b, m);
    if (m->special == SPEC_CASTLEQ) {
        s += sprintf(s, "0-0-0");
    } else if (m->special == SPEC_CASTLEK) {
        s += sprintf(s, "0-0");
    } else {

        int sourceRow = 1;

        if (abs(m->piece) != PAWN)
            *s++ = "K NQBR RBQN K"[m->piece+6];
        else if (b->currMove->captured) {
            *s++ = X(m->src) + 'a';
            sourceRow = 0;
        }

        for (;;) {
            int matches = 0;
            for (Move *currMove = firstMove; currMove < lastMove; currMove++)
                if (currMove->dst == m->dst && currMove->piece == m->piece && currMove->special == m->special)
                    matches++;
            if (matches == 1)
                break;

            matches = 0;
            for (Move *currMove = firstMove; currMove < lastMove; currMove++)
                if (currMove->dst == m->dst && currMove->piece == m->piece && X(currMove->src) == X(m->src) && currMove->special == m->special)
                    matches++;
            if (matches == 1) {
                if (sourceRow)
                    *s++ = X(m->src) + 'a';
                break;
            }

            matches = 0;
            for (Move *currMove = firstMove; currMove < lastMove; currMove++)
                if (currMove->dst == m->dst && currMove->piece == m->piece && Y(currMove->src) == Y(m->src) && currMove->special == m->special)
                    matches++;
            if (matches == 1) {
                *s++ = Y(m->src) + '1';
                break;
            }

            for (Move *currMove = firstMove; currMove < lastMove; currMove++)
                if (currMove->dst == m->dst && currMove->piece == m->piece && currMove->src == m->src && currMove->special == m->special) {
                    *s++ = X(m->src) + 'a';
                    *s++ = Y(m->src) + '1';
                    break;
                }
            break;
        }
        if (b->currMove->captured)
            *s++ = 'x';

        *s++ = X(m->dst) + 'a';
        *s++ = Y(m->dst) + '1';
    }
    switch (m->special) {
    case SPEC_PROMOQ:
        *s++ = 'Q';
        break;
    case SPEC_PROMON:
        *s++ = 'N';
        break;
    case SPEC_PROMOB:
        *s++ = 'B';
        break;
    case SPEC_PROMOR:
        *s++ = 'R';
        break;
    }

    if (IN_CHECK(-b->color))
        *s++ = '+';
    *s++ = ' ';
    *s=0;
    undoMove(b);
    return 1;
}

void line2str(const Move *firstMove, char *s) {
    int i;
    MainEntry *p;
    char *sstart = s;
    static Board *const b = &board;
//    b->currMove->move.val = slowEval(b); //to make b->score have a defined value

    if (b->color == -1) {
        *s++ = '.';
        *s++ = '.';
        *s++ = '.';
    }

    unsigned long long reptable[MAXDEPTH];
    Move m = *firstMove;
    for (i=0; ;) {
        char mstr[10];
        if (!moveToStrX(b, &m, mstr)) break;
        if (b->color == 1)
            s += sprintf(s, "%d. ", (b->currMove - b->moveList) / 2 + 1);
        s += sprintf(s, "%s", mstr);
        
        doMove(b, &m);
        b->color = -b->color;
        i++;
        
        for (int j=0; j<i-1; j++)
            if (reptable[j] == b->key.main.key)
                goto lpend;
        if (i==MAXDEPTH)
            break;
                    
        reptable[i-1] = b->key.main.key;
                
        p = readHash(b);
        if (p) {
#ifdef SMALLHASH        
            m.src = ADR(p->src >> 3, p->src & 7);
            m.dst = ADR(p->dst >> 3, p->dst & 7);
#else
            m.src = p->src << 2;
            m.dst = p->dst << 2;
	    m.special = p->special;
#endif            
        } else    
            break;
    }
lpend:

    if (s > sstart)
        s--;
    *s = 0;
    for (; i > 0; --i) {
        b->color = -b->color;
        undoMove(b);
    }
}

const char *moveToStr(const Move * m)
{
    static char str[256];
    char *s = str;
    
    if (m->special == SPEC_CASTLEQ)
        return "0-0-0";
    else if (m->special == SPEC_CASTLEK)
        return "0-0";

    switch (abs(m->piece)) {
    case ROOK:
        *s++ = 'R';
        break;
    case BISHOP:
        *s++ = 'B';
        break;
    case QUEEN:
        *s++ = 'Q';
        break;
    case KING:
        *s++ = 'K';
        break;
    case KNIGHT:
        *s++ = 'N';
        break;
    }

    *s++ = 'a' + X(m->src);
    *s++ = '1' + Y(m->src);
    if (m->capture)
        *s++ = 'x';
    else
        *s++ = '-';

    *s++ = 'a' + X(m->dst);
    *s++ = '1' + Y(m->dst);

    switch (m->special) {
    case SPEC_PROMOQ:
        *s++ = 'Q';
        break;
    case SPEC_PROMON:
        *s++ = 'N';
        break;
    case SPEC_PROMOB:
        *s++ = 'B';
        break;
    case SPEC_PROMOR:
        *s++ = 'R';
        break;
    }

    if (m->attack)
        *s++ = '+';
    *s = 0;
    return str;
}

int str2move(const char *s, Move * m, Move * firstMove, Move * lastMove)
{
    Move *currMove;
    int l;
    int spec = 0, type = 0;
    int x1 = 0, x2 = 0, _y1 = 0, y2 = 0;
    int dundef = 1;
    int xundef = 1;
    int yundef = 1;
    int tundef = 1;
    int sundef = 1;
    char *sstart;
    char buf[256];
    int matches = 0;

    sstart = buf;

    do {
        if ((*s >= 'a' && *s <= 'r') || (*s >= 'A' && *s <= 'R') || (*s >= '0' && *s <= '9'))
            *sstart++ = *s;
    } while (*s++);
    *sstart = 0;

    l = strlen(buf) - 1;
    if (l < 1 || l > 4)
        return 0;

    do {
        if (!strcmp(buf, "000") || !strcmp(buf, "ooo") || !strcmp(buf, "OOO")) {
            sundef = 0;
            spec = SPEC_CASTLEQ;
            break;
        }
        if (!strcmp(buf, "00") || !strcmp(buf, "oo") || !strcmp(buf, "OO")) {
            sundef = 0;
            spec = SPEC_CASTLEK;
            break;
        }

        if (buf[l] == 'R' || buf[l] == 'r') {
            sundef = 0;
            spec = SPEC_PROMOR;
            --l;
        } else if (buf[l] == 'Q' || buf[l] == 'q') {
            sundef = 0;
            spec = SPEC_PROMOQ;
            --l;
        } else if (buf[l] == 'B' || buf[l] == 'b') {
            sundef = 0;
            spec = SPEC_PROMOB;
            --l;
        } else if (buf[l] == 'N' || buf[l] == 'n') {
            sundef = 0;
            spec = SPEC_PROMON;
            --l;
        }

        if (l < 1)
            return 0;

        y2 = buf[l] - '1';
        x2 = buf[l - 1] - 'a';
        dundef = 0;
        l -= 2;

        if (l < 0) {
            type = PAWN;
            tundef = 0;
            break;
        }

        if (buf[l] >= '1' && buf[l] <= '8') {
            yundef = 0;
            _y1 = buf[l] - '1';
            l--;
        }
        if (l < 0)
            break;

        if (buf[l] >= 'a' && buf[l] <= 'h') {
            xundef = 0;
            x1 = buf[l] - 'a';
            if (l == 0 && buf[l] == 'b' && yundef) {    //this is something like 'bc3'... could be a pawn or a bishop move. check for pawn move
                for (currMove = firstMove; currMove < lastMove; currMove++)
                    if ((currMove->dst == ADR(y2, x2))
                        && (abs(currMove->piece) == PAWN)
                        && (X(currMove->src) == x1)
                        && (x1 != x2)
                        && (currMove->special == spec || sundef)) {
                        *m = *currMove;
                        matches++;
                    }
                if (matches == 1)
                    return 1;
                xundef = 1;     //no pawn move found, so the letter supplied must be 'b' for bishop.
                l++;
            }
            l--;
        }
        if (l < 0)
            break;
            
        switch (buf[l]) {
        case 'N':
        case 'n':
            type = KNIGHT;
            tundef = 0;
            l--;
            break;
        case 'B':
        case 'b':
            type = BISHOP;
            tundef = 0;
            l--;
            break;
        case 'Q':
        case 'q':
            type = QUEEN;
            tundef = 0;
            l--;
            break;
        case 'R':
        case 'r':
            type = ROOK;
            tundef = 0;
            l--;
            break;
        case 'K':
        case 'k':
            type = KING;
            tundef = 0;
            l--;
            break;
        }
        if (xundef && yundef && tundef && sundef && !dundef)
            type = PAWN;
        if (l >= 0)
            return 0;
    } while (0);

    for (currMove = firstMove; currMove < lastMove; currMove++)
        if ((currMove->dst == ADR(y2, x2) || dundef)
            && (X(currMove->src) == x1 || xundef)
            && (Y(currMove->src) == _y1 || yundef)
            && (abs(currMove->piece) == type || tundef)
            && (currMove->special == spec || sundef)) {
            *m = *currMove;
            matches++;
        }
    if (matches > 1 || !matches)
        return 0;
    return 1;
}

int importEPD(const char *s)
{
    static Board *const b = &board;

    clearBoard(b);

    int x = 0;
    int y = 7;

    while (isspace(*s))
        s++;

    while (!isspace(*s) && *s) {
        switch (*s) {
        case '1'...'8':
            x += *s - '1';
            break;
        case 'k':
            putPiece(b, ADR(y, x), -KING);
            break;
        case 'p':
            putPiece(b, ADR(y, x), -PAWN);
            break;
        case 'b':
            putPiece(b, ADR(y, x), -BISHOP);
            break;
        case 'n':
            putPiece(b, ADR(y, x), -KNIGHT);
            break;
        case 'r':
            putPiece(b, ADR(y, x), -ROOK);
            break;
        case 'q':
            putPiece(b, ADR(y, x), -QUEEN);
            break;
        case 'K':
            putPiece(b, ADR(y, x), KING);
            break;
        case 'P':
            putPiece(b, ADR(y, x), PAWN);
            break;
        case 'B':
            putPiece(b, ADR(y, x), BISHOP);
            break;
        case 'N':
            putPiece(b, ADR(y, x), KNIGHT);
            break;
        case 'R':
            putPiece(b, ADR(y, x), ROOK);
            break;
        case 'Q':
            putPiece(b, ADR(y, x), QUEEN);
            break;
        case '/':
            x = -1;
            y--;
            break;
        default:
            initBoard();
            return 1;
        }
        x++;
        s++;
    }

    while (isspace(*s))
        s++;

    switch (*s++) {
    case 'w':
    case 'W':
        board.color = 1;
        break;
    case 'b':
    case 'B':
        board.color = -1;
    }

    while (isspace(*s))
        s++;

    while (*s && !isspace(*s))
        switch (*s++) {
        case 'Q':
            board.o_o |= CAST_WQ;
            break;
        case 'K':
            board.o_o |= CAST_WK;
            break;
        case 'q':
            board.o_o |= CAST_BQ;
            break;
        case 'k':
            board.o_o |= CAST_BK;
        }

    while (isspace(*s))
        s++;

    if (s[0] >= 'a' && s[0] <= 'h' && s[1] >= '1' && s[2] <= '2')
        board.epstatus = ADR(s[1] - '1', s[0] - 'a');
    copyBoard();
    return 0;
}
