/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#ifndef VIEW_H
#define VIEW_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// include files for Qt
#include <qvaluelist.h>
#include <qpoint.h>

/* defines */
#define MAXPIECES 64

class Game;

//simple base class for coordinate representation
//of a piece.
template <class T> struct Coor {
    T x;
    T y;
    Coor() {};
    Coor(T a, T b) {
        x=a;
        y=b;
    }
    bool operator == (const Coor a) const {
        return (a.x == x && a.y == y);
    }
    Coor operator + (const Coor a) const {
        return Coor(a.x + x, a.y + y);
    }
    Coor operator * (T v) const {
        return Coor(x * v, y * v);
    }
};

typedef Coor<int> CoorInt;
typedef Coor<float> CoorFloat;

//base class for visible pieces
struct ViewPiece {
    CoorInt pos;
    CoorFloat curr;		//current postion
    CoorFloat src;	//source position for sliding pieces
    CoorFloat dst;	//destination for sliding pieces
    CoorFloat vec;	//moving vector
    float d;
    float alpha, stime;
    int moving, fading, picked;
    int type, idx;
};

/** The View class provides the view widget for the SliboApp instance.
 * The View instance inherits QWidget as a base class and represents the view object of a KTMainWindow. As View is part of the
 * docuement-view model, it needs a reference to the document object connected with it by the SliboApp class to manipulate and display
 * the document structure provided by the Game class.
 *
 * @author Source Framework Automatically Generated by KDevelop, (c) The KDevelop Team.
 * @version KDevelop version 0.4 code generation
 */

class View {
public:
    View();
    virtual ~View();
    bool repaintNeeded;
    int picking;
    Game *game;

protected:
    virtual CoorFloat screen2board(QPoint)=0;
    virtual QPoint board2screen(CoorFloat)=0;
    virtual int selection(QPoint)=0;

    bool leftClick(QPoint);
    float edges[4][3];
    CoorInt src, dst;
    QPoint pickPos, oldpickPos;
    QPoint oldPos;
};


#endif				// View_H
